#ifndef DNAMODEL_H
#define DNAMODEL_H

#include "Models/MatrixModel.h"
class Perturbator;
class ParametersSet;
class SequenceTable;


class DnaModel : public MatrixModel {
  
protected:
    DnaModel( const string & registrationName );

    DnaModel( ParametersSet & parameter );
    
    array2D<int> matrixIndex;
    
public:
    virtual ~DnaModel();

    
    /** ************************************************************************
     * getSymbolNumber
     * @input   base, a string representing a symbol of the model
     * @return  The basenumber of the symbol
     *           e.g. "A"=0 , "C"=1 , "G"=2 and "U"=3 for the DNA models
     ************************************************************************ */
    int getSymbolNumber( const string & base, int symbolCategory = 0 ) const;


    /** ************************************************************************
     * getSymbol
     * @input   symbolNumber, the index of a symbol of the model
     * @return  The string which represents the state
     ************************************************************************ */
    string getSymbol( int symbolNumber, int symbolCategory = 0 ) const;

    /** ************************************************************************
     * getState
     * @input   stateNumber, the index of a state of the model
     * @return  The string which represents the state
     ************************************************************************ */
    virtual string getState( int stateNumber, int symbolCategory = 0 ) const;
    
    /** ************************************************************************
     * getNumberSymbols
     * @return  The number of distinct symbols recognized by the model
     ************************************************************************ */
    inline int getNumberSymbols( int symbolCategory = 0 ) const {
        assert( symbolCategory == 0 );
        return 8;
    }

    /** ************************************************************************
     * getNumberStates
     * @return  The number of states a model has
     ************************************************************************ */
    int getNumberStates( int symbolCategory = 0 ) const {
        assert( symbolCategory == 0 );
        return 4;
    }
 
    /** ************************************************************************
     * retrieveEmpiricalFrequencies
     * @input      sequenceTable, the SequenceTable instance containing the
     *             species' sequences used to perform the inference.
     * @semantics  initialises the frequencies vector from the sequence
     ************************************************************************ */
    vector < double > retrieveEmpiricalFrequencies( SequenceTable * sequenceTable, int modelId );


    /** ************************************************************************
     * retrieveEmpiricalRates
     * @input      sequenceTable, the SequenceTable instance containing the
     *             species' sequences used to perform the inference.
     * @input      a pointer to a double to store the proportion of invariant
     *             site (not compulsory)
     * @return     a rate ratio vector of length 5 initialised according to the
     *             sequences in the sequenceTable
     ************************************************************************ */
    vector < double > retrieveEmpiricalRates( SequenceTable * sequenceTable,
    int modelId, double * propInvariant = NULL );

        
    /** ************************************************************************
     * setAverageMutationRate
     * @input  newAverageRate the new average rate of mutation
     ************************************************************************ */
    virtual void setAverageMutationRate( double newAverageRate, int symbolCategory = 0 );

    
    /** ************************************************************************
     * probability
     * @input   oldState, a state of the model
     * @input   newState, a state of the model
     * @input   time, a delay value
     * @input   category, a category
     * @return  The probability of ....
     ************************************************************************ */
    double probability( int oldState, int newState, double time,
    int category, int symbolCategory = 0) const;

    /** ************************************************************************
     * diffProbability
     * @input   oldState, a state of the model
     * @input   newState, a state of the model
     * @input   time, a delay value
     * @input   category, a category
     * @return  The probability of ....
     ************************************************************************ */
    double diffProbability( int oldState, int newState, double time,
    int category, int symbolCategory = 0 ) const;

    /** ************************************************************************
     * secondDiffProbability
     * @input   oldState, a state of the model
     * @input   newState, a state of the model
     * @input   time, a delay value
     * @input   category, a category
     * @return  The probability of ....
     ************************************************************************ */
    double secondDiffProbability( int oldState, int newState, double time,
    int category, int symbolCategory = 0 ) const;


    /** ************************************************************************
     * printParameters
     * @input      outputStream, the stream used to output the parameters
     * @semantics  Output the parameters names and values of the model in a
     *             readable format.
     *             -frequencies of the state
     *             -gamma shape parameter (if relevant)
     *             -proportion of invariant sites (if relevant)
     *             -mutation rate for each category
     ************************************************************************ */
    virtual void printParameters( ostream & outputStream ) const;


    void initEquivalencyTable();
   
protected:
    virtual void setEigenMatrix() = 0;

};

#endif //DNAMODEL_H
