#ifndef MATRIXMODEL_H
#define MATRIXMODEL_H

#include <vector>
using namespace std;

#include "Models/Model.h"

class Perturbator;
class ParametersSet;
class SequenceTable;

class MatrixModel : public Model {

protected :
    MatrixModel( const string & registrationName );

    MatrixModel( ParametersSet & parameter );
    
public :
    virtual ~MatrixModel();

    void initialisation( SequenceTable * sequenceTable = NULL, int modelId = 0 );
    
    /** ************************************************************************
     * getNumberSymbolCategory
     * @return  The number of categories of symbol the model deals with
     ************************************************************************ */
    inline int getNumberSymbolCategory() const {
        return 1;
    }
    /** ************************************************************************
     * getSymbol
     * @input   symbolNumber, the index of a symbol of the model
     * @return  The string which represents the state
     ************************************************************************ */
    virtual string getSymbol( int symbolNumber, int symbolCategory = 0 ) const = 0;

    /** ************************************************************************
     * getState
     * @input   stateNumber, the index of a state of the model
     * @return  The string which represents the state
     ************************************************************************ */
    virtual string getState( int stateNumber, int symbolCategory = 0 ) const = 0;

    
    /** ************************************************************************
     * getModelParameters
     * @semantics  create a ParametersSet and fill it with the parameters of
     *             the model
     ************************************************************************ */
    ParametersSet getModelParameters();

    /** ************************************************************************
     * setModelParameters
     * @semantics      set the parameters of the model according to the content
     *                 of the parameters set
     * @preconditions  the parameters in the set are consistent
     ************************************************************************ */
    void setModelParameters( ParametersSet & parameters );

    
    /** ************************************************************************
     * getFrequency
     * @intput  residue, the index of a state of the model
     * @intput  rateCategory, the discrete gamma category
     * @intput  symbolCategory, must be 0
     * @return  The frequencies of the given state of the model
     ************************************************************************ */
    inline virtual double getFrequency( int residue,
    int rateCategory, int symbolCategory = 0 ) const {
        assert( symbolCategory == 0 );
        assert ( residue < (int)frequencies[frequencyCategory[rateCategory]].size() );
        return frequencies[frequencyCategory[rateCategory]] [residue];
    }

    /** ************************************************************************
     * getFrequency
     * @intput  sbase, a state of the model
     * @intput  rateCategory, the discrete gamma category
     * @intput  symbolCategory, must be 0
     * @return  The frequencies of the given state of the model
     ************************************************************************ */
    inline double getFrequency( const string & sbase,
    int rateCategory, int symbolCategory = 0 ) const {
        return getFrequency( getSymbolNumber( sbase, 0 ),
        rateCategory, symbolCategory );
    }


    /** ************************************************************************
     * getProportionInvariantSites
     * @output  prop, the proportion
     ************************************************************************ */
    double getProportionInvariantSites( int symbolCategory = 0) const;

    /** ************************************************************************
     * setProportionInvariantSites
     * @input  prop, the new proportions
     ************************************************************************ */
    void setProportionInvariantSites( double newProp, int symbolCategory = 0 );
    
    /** ************************************************************************
     * getGammaShapeParameter
     * @return  the gamma shape parameter alpha
     ************************************************************************ */
    double getGammaShapeParameter( int symbolCategory = 0 ) const;

    /** ************************************************************************
     * setGammaShapeParameter
     * @input  newAlpha, the new gamma shape parameter
     ************************************************************************ */
    void setGammaShapeParameter( double newAlpha, int symbolCategory = 0 );


    /** ************************************************************************
     * getNumberRateCategories
     * @return  The number of rate categories a model has
     ************************************************************************ */
    inline int getNumberRateCategories( int symbolCategory = 0 ) const {
        assert( symbolCategory == 0 );
        return numberRateCategories;
    }
    
    /** ************************************************************************
     * getRateCategoryProbability
     * @input   category, the index of a category
     * @return  The probability of a particular rate category
     ************************************************************************ */
    double getRateCategoryProbability( int category, int symbolCategory = 0 ) const;

    /** ************************************************************************
     * getAverageMutationRate
     * @return  The average rate of mutation
     ************************************************************************ */
    virtual double getAverageMutationRate( int symbolCategory = 0 ) const;

    
    /** ************************************************************************
     * setAverageMutationRate
     * @input  newAverageRate the new average rate of mutation
     ************************************************************************ */
    virtual void setAverageMutationRate( double newAverageRate, int symbolCategory = 0) = 0;

    /** ************************************************************************
     * initialisePerturbator
     * @input      perturbation parameters
     * @semantics  Warn the model that it will have to perturb itself and
     *             provide some useful perturbation parameters
     ************************************************************************ */
    virtual void initialisePerturbator( ParametersSet & parameters );

    /** ************************************************************************
     * perturb
     * @semantics  ask the model to perturb some of its parameters
     ************************************************************************ */
    virtual double perturb();

    /** ************************************************************************
     * validatePerturbation
     * @input      validation, true if the perturbation is accepted
     * @semantics  restore the model to the state before the perturbation
     *             if validation==false
     ************************************************************************ */
    virtual bool validatePerturbation( bool validation );

    /** ************************************************************************
     * stopBurn
     * @input      none
     * @semantics  this function is called to inform the model that the burnin
     *             period is finished while performing a MCMC run
     ************************************************************************ */
    virtual void stopBurn();

    
    /** ************************************************************************
     * printPerturbationParameters
     * @input      outputStream, the stream used to output the parameters
     * @semantics  Output the perturbation parameters names and values of the
     *             model in a readable format
     ************************************************************************ */
    virtual void printPerturbationParameters( ostream & outputStream );

    
    /** ************************************************************************
     * getAllParameters
     * @input   pars, an instance of vector <double>
     * @output  pars, an instance of vector <double> filled with all the
     *          parameters of the model reparametrized so as to be independant
     ************************************************************************ */
    virtual void getAllParameters(vector < double > & pars ) const;

    /** ************************************************************************
     * setAllParameters
     * @input  gamma, an instance of vector <double>, filled with all the new
     *         parameters of the model
     ************************************************************************ */
    virtual void setAllParameters(const vector < double > & pars );

    /** ************************************************************************
     * getNumberFreeParameters
     * @return   The number of free parameters in the model (and
     *          the size of the pars vector above
     ************************************************************************ */
    virtual unsigned int getNumberFreeParameters();
    
    /** ************************************************************************
     * @input outputStream  The stream used to output the parameters
     * @semantics output the parameters values of the model in a single line
     ************************************************************************ */
    virtual void printLine( ostream & outputStream );

    /** ************************************************************************
     * fromLine
     * @input the new parameters of the model as they are printed with
     *        printLine (ie, same order)
     ************************************************************************ */
    virtual void fromLine( const vector<double>& parameters );
            
    /** ************************************************************************
     * getNumberLineParameters
     * @input the number of parameters printed/read with printLine and fromLine
     ************************************************************************ */
    virtual unsigned int getNumberLineParameters();
    
    /** ************************************************************************
     * printParameters
     * @input      outputStream, the stream used to output the parameters
     * @semantics  Output the parameters names and values of the model in a
     *             readable format. (the frequencies and matrices cannot be
     *             handled at that level
     *             -gamma shape parameter (if relevant)
     *             -proportion of invariant sites (if relevant)
     *             -mutation rate for each category
     ************************************************************************ */
    virtual void printParameters( ostream & outputStream ) const;

    
    /** ************************************************************************
     * getEquivalencyTable
     * @input   oldState, a state of the model
     * @input   newState, a state of the model
     * @input   time, a delay value
     * @input   category, a category
     * @return  The equivalency table (symbol->state)
     ************************************************************************ */
    inline const array2D< double > &
    getEquivalencyTable( int symbolCategory = 0 ) const {
        assert( symbolCategory == 0 );
        return equivalencyTable;
    }
    
protected :

    // average substitution rate
    double averageMutationRate;

    // each model allows a different number of variable frequencies
    int numberFrequencies;
    // each model use a matrix with a given number of different rates
    // for the JukesCantor model, numberRatesRatiosPerMatrix = 0
    // for the GTR model, numberRatesRatiosPerMatrix = 5
    int numberRatesRatiosPerMatrix;


    Perturbator * perturbator;

    int numberRateCategories;

    //invariant category flag
    int invariantCategory;
    //proportion of invariant sites
    double proportionInvariantSites;    
    // discrete gamma model flag
    bool discreteGamma;
    // gamma shape parameter
    double alpha;
    // number of gamma categories
    int numberGammaCategories;
        
    // each gamma category can be assigned its own set of states frequencies
    // Usually numberRatesMatrices == 1 (the usual case) or
    // numberGammaCategories + (invariantCategory ? 1 : 0).
    int numberFrequenciesCategories;
    // vector used to make the correspondance between a gamme rate category and
    // the states frequencies.
    vector < int > frequencyCategory;
    
    // each gamma category can be assigned its own private rate matrix.
    // Usually numberRatesMatrices == 1 (the usual case) or
    // numberGammaCategories.
    // each different rate matrix will be assigned its own rate ratios
    int numberRatesMatrices;
    // vector used to make the correspondance between a rate category and its
    // rate matrix.
    vector < int > matrixCategory;

    // vector used to store the frequencies of each state
    vector < double > * frequencies;
    vector < double > * rateRatios;
    
    array2D< double > equivalencyTable;

    // multiplication factor used in front of each transition matrice
    vector < double > mutationRate;
    
    // eigen system
    vector < double > * eigenValues;
    array2D< double > * eigenMatrix;
    array2D< double > * ieigenMatrix;
    //transition rate matrix
    array2D< double > * rateMatrix;
    //array2D< double > equivalencyMatrix;



};

#endif //MATRIXMODEL_H
  
