#ifndef MIXED_H
#define MIXED_H

#include <assert.h>

#include <iostream>
#include <iomanip>

#include <vector>


#include "Models/Model.h"

#include "Util/array2D.h"

class ParametersSet;
class SequenceTable;
class MixedPerturbator;

using namespace std;

class Mixed : public Model {
private:

    array2D< double > equivalencyTable;

    vector< Model* > model;

    static Mixed prototype;

protected:
    MixedPerturbator * perturbator;


private:
    /** .**********************************************************************
     * Mixed
     * @semantics  default constructor of the mixed mutation
     *             model object, this construtor must be used only once
     *             to register the prototype to the ModelFactory
     ************************************************************************ */
    Mixed( const string & registrationName );

public:
    /** .**********************************************************************
     * clone
     * @semantics  function used by the ModelFactory to clone the prototype and
     *             create a real model.
     ************************************************************************ */
    virtual Model * clone( ParametersSet & parameters ) const {
        return new Mixed( parameters );
    }

    /** ************************************************************************
     * getModelParameters
     * @semantics  create a ParametersSet and fill it with the parameters of
     *             the model
     ************************************************************************ */
    ParametersSet getModelParameters();


    /** ************************************************************************
     * setModelParameters
     * @semantics      set the parameters of the model according to the content
     *                 of the parameters set
     * @preconditions  the parameters in the set are consistent
     ************************************************************************ */
    void setModelParameters( ParametersSet & parameters );

protected:
    /** .***********************************************************************
     * Mixed
     * @pre        parameters coherent
     * @semantics  Constructor of the Mixed mutation model object
     ************************************************************************ */
    Mixed( ParametersSet & parameters );


public :
    void initialisation( SequenceTable * SequenceTable, int modelId = 0 );

    /** ************************************************************************
     * getNumberSymbolCategory
     * @return  The number of categories of symbol the model deals with
     ************************************************************************ */
    virtual int getNumberSymbolCategory() const {
        return (int)model.size();
    }


    virtual ~Mixed();

    /** ************************************************************************
     * getNumberStates
     * @return  The number of states a model has
     ************************************************************************ */
    inline virtual int getNumberStates( int symbolCategory ) const {
        assert( ( symbolCategory >=0  ) && ( symbolCategory < (int)model.size()  ) );
        return model[symbolCategory]->getNumberStates( 0 );
    }

    /** ************************************************************************
     * getNumberSymbols
     * @return  The number of distinct symbols recognized by the model
     ************************************************************************ */
    inline virtual int getNumberSymbols( int symbolCategory ) const {
        assert( ( symbolCategory >=0  ) && ( symbolCategory < (int)model.size() ) );
        return model[symbolCategory]->getNumberSymbols( 0 );
    }

    /** ************************************************************************
     * getSymbolNumber
     * @input   base, a string representing a state of the model
     * @return  The basenumber of the state
     *           e.g. "A" = 1 , "C"="2" , "G"=3 and U="4" for the DNA models
     ************************************************************************ */
    inline virtual int getSymbolNumber( const string & base, int symbolCategory ) const {
        assert( ( symbolCategory >=0  ) && ( symbolCategory < (int)model.size() ) );
        return model[symbolCategory]->getSymbolNumber( base, 0 );
    }



    /** ************************************************************************
     * getSymbol
     * @input   basenumber, the index of a state of the model
     * @return  The string which represents the state
     ************************************************************************ */
    inline virtual string getSymbol( int baseNumber, int symbolCategory ) const {
        assert( ( symbolCategory >=0  ) && ( symbolCategory < (int)model.size() ) );
        return model[symbolCategory]->getSymbol( baseNumber, 0 );
    }

    /** ************************************************************************
     * getName
     * @return  The name of the model
     ************************************************************************ */
    inline virtual string getName( void ) const {
        string name( "Mixed model : " + model[0]->getName() );
        for (unsigned int i = 1; i < model.size(); ++i ){
             name = name + " & " + model[i]->getName();
        }
        return name;
    }

    /** ************************************************************************
     * getFrequency
     * @intput  residue, the index of a state of the model
     * @return  The frequencies of the given state of the model
     ************************************************************************ */
    inline virtual double getFrequency( int residue, int rateCategory,
                                        int symbolCategory ) const {
        assert( ( symbolCategory >=0  ) && ( symbolCategory < (int)model.size() ) );
        return model[symbolCategory]->getFrequency( residue, rateCategory, 0 );
    }

    /** ************************************************************************
     * getFrequency
     * @intput  sbase, a state of the model
     * @return  The frequencies of the given state of the model
     ************************************************************************ */
    inline virtual double getFrequency( const string & sbase, int rateCategory,
                                        int symbolCategory ) const {
        return getFrequency( getSymbolNumber( sbase, symbolCategory ),
                             rateCategory, symbolCategory );
    }

    /** ************************************************************************
     * getNumberRateCategories
     * @return  The number of rate categories a model has
     ************************************************************************ */
    inline virtual int getNumberRateCategories( int symbolCategory ) const {
        assert( ( symbolCategory >=0  ) && ( symbolCategory < (int)model.size() ) );
        return model[symbolCategory]->getNumberRateCategories( 0 );
    }

    /** ************************************************************************
     * getRateCategoryProbability
     * @input   category, the index of a category
     * @return  The probability of a particular rate category
     ************************************************************************ */
    inline virtual double getRateCategoryProbability( int category,
                                                   int symbolCategory ) const {
        assert( ( symbolCategory >=0  ) && ( symbolCategory < (int)model.size() ) );
        return model[symbolCategory]->getRateCategoryProbability( category, 0 );
    }

    /** ************************************************************************
     * getAverageMutationRate
     * @return  The average rate of mutation
     ************************************************************************ */
    inline virtual double getAverageMutationRate( int symbolCategory ) const {
        assert( ( symbolCategory >=0  ) && ( symbolCategory < (int)model.size() ) );
        return model[symbolCategory]->getAverageMutationRate( 0 );
    }

    /** ************************************************************************
     * setAverageMutationRate
     * @semantics  Provided to comply to the model interface. Must not be used
     ************************************************************************ */
    inline virtual void setAverageMutationRate( double newAverageRate,
                                                int symbolCategory ){
        assert(0);
        symbolCategory = symbolCategory;
        newAverageRate = newAverageRate;
    }
    
    
    /** ************************************************************************
     * getGammaShapeParameter
     * @return  the gamma shape parameter alpha
     ************************************************************************ */
    inline virtual double getGammaShapeParameter( int symbolCategory ) const {
        assert( ( symbolCategory >=0  ) && ( symbolCategory < (int)model.size() ) );
        return model[symbolCategory]->getGammaShapeParameter( 0 );
    }

    /** ************************************************************************
     * getAllParameters
     * @input   pars, an instance of vector <double>
     * @output  pars, an instance of vector <double> filled with all the
     *          parameters of the model reparametrized so as to be independant
     ************************************************************************ */
    virtual void getAllParameters( vector < double > & pars ) const;

    /** ************************************************************************
     * setAllParameters
     * @input  pars, an instance of vector <double>, filled with all the new
     *         parameters of the model
     ************************************************************************ */
    virtual void setAllParameters( const vector < double > & pars );

    /** ************************************************************************
     * getNumberFreeParameters
     * @return   The number of free parameters in the model (and
     *          the size of the pars vector above
     ************************************************************************ */
    virtual unsigned int getNumberFreeParameters();
    
    
    /** ************************************************************************
     * probability
     * @input   oldState, a state of the model
     * @input   newState, a state of the model
     * @input   time, a delay value
     * @input   category, a category
     * @return  The probability of ....
     ************************************************************************ */
    inline virtual double probability( int oldState, int newState,
                       double time, int category,int symbolCategory ) const {
        assert( ( symbolCategory >=0  ) &&
                ( symbolCategory < (int)model.size() ) );
        return model[symbolCategory]->probability( oldState, newState,
                                                           time, category, 0 );
    }

    /** ************************************************************************
     * diffProbability
     * @input   oldState, a state of the model
     * @input   newState, a state of the model
     * @input   time, a delay value
     * @input   category, a category
     * @return  The probability of ....
     ************************************************************************ */
    virtual double diffProbability( int oldState, int newState,
                       double time, int category, int symbolCategory ) const {
        assert( ( symbolCategory >=0  ) && ( symbolCategory < (int)model.size() ) );
        return model[symbolCategory]->diffProbability( oldState, newState,
                                                           time, category, 0 );
    }

    /** ************************************************************************
     * secondDiffProbability
     * @input   oldState, a state of the model
     * @input   newState, a state of the model
     * @input   time, a delay value
     * @input   category, a category
     * @return  The probability of ....
     ************************************************************************ */
    virtual double secondDiffProbability( int oldState, int newState,
                      double time, int category, int symbolCategory ) const {
        assert( ( symbolCategory >=0  ) && ( symbolCategory < (int)model.size() ) );
        return model[symbolCategory]->secondDiffProbability( oldState, newState,
                                                           time, category, 0 );
    }


    inline virtual const array2D< double > &
    getEquivalencyTable( int symbolCategory ) const {
        assert( ( symbolCategory >=0  ) && ( symbolCategory < (int)model.size() ) );
        return model[symbolCategory]->getEquivalencyTable( 0 );
    }

    /** ************************************************************************
     * printParameters
     * @input      outputStream, the stream used to output the parameters
     * @semantics  Output the parameters names and values of the model in a
     *             readable format
     ************************************************************************ */
    inline virtual void printParameters( ostream & outputStream ) const {
        for (unsigned int i = 0; i < model.size(); ++i ){
            outputStream << "MODEL" << (i+1) << endl;
            model[i]->printParameters( outputStream );
        }
    }

    /** ************************************************************************
     * printLine
     * @input      outputStream  The stream used to output the parameters
     * @semantics  output the parameters values of the model in a single line
     ************************************************************************ */
    inline virtual void printLine( ostream & outputStream );
    
    /** ************************************************************************
     * fromLine
     * @input the new parameters of the model as they are printed with
     *          printLine (ie, same order)
     ************************************************************************ */
    virtual void fromLine( const vector<double>& parameters );
            
    /** ************************************************************************
     * getNumberLineParameters
     * @return the number of parameters printed/read with printLine and
     *         fromLine
     ************************************************************************ */
    inline virtual unsigned int getNumberLineParameters();

    virtual void initialisePerturbator( ParametersSet & parameters );

    /** ************************************************************************
     * printPerturbationParameters
     * @input      outputStream, the stream used to output the parameters
     * @semantics  Output the perturbation parameters names and values of the
     *             model in a readable format
     ************************************************************************ */
    virtual void printPerturbationParameters( ostream & outputStream );

    /** ************************************************************************
     * perturb
     * @semantics  ask the model to perturb some of its parameters
     ************************************************************************ */
    virtual double perturb();

    /** ************************************************************************
     * validatePerturbation
     * @input      validation status : true to validate the change from the
     *             last perturbation, false otherwise
     * @semantics  according to the validation flag, restore the model
     *             parameters as they were before the last call to perturb
     *             or validate the changes.
     ************************************************************************ */
    virtual bool validatePerturbation( bool validation );

    /** ************************************************************************
     * stopBurn
     * @input      none
     * @semantics  this function is called to inform the model that the burnin
     *             period is finished while performing a MCMC run
     ************************************************************************ */
    virtual void stopBurn();

};

#endif //MIXED_H




