#ifndef MODEL_H
#define MODEL_H

#include <string>
#include <fstream>
#include <vector>

#include "Util/array2D.h"
#include "Util/ParametersSet.h"

class Perturbator;
class SequenceTable;

#define MIN(a,b) ((a) < (b) ? (a) : (b))
#define MAX(a,b) ((a) > (b) ? (a) : (b))

class Model {
public:

    /** .***********************************************************************
     * ~Model
     * @semantics A virtual destructor needed for inheritance
     ************************************************************************ */
    virtual ~Model() {
    };

    /** ************************************************************************
     * initialisation
     * @semantics  initialise the model and its parameters according to
     *             default value or the content of the sequences
     ************************************************************************ */
    virtual void initialisation( SequenceTable * sequenceTable = NULL, int modelId = 0 ) = 0;


    /* .**********************************************************************
     * clone
     * @semantics  function used by the ModelFactory to clone the prototype
     *             and create a real model.
    ************************************************************************ */
    virtual Model * clone( ParametersSet & parameters ) const = 0;

    /** ************************************************************************
     * getModelParameters
     * @semantics  create a ParametersSet and fill it with the parameters of
     *             the model
     ************************************************************************ */
    virtual ParametersSet getModelParameters() = 0;

    /** ************************************************************************
     * setModelParameters
     * @semantics      set the parameters of the model according to the content
     *                 of the parameters set
     * @preconditions  the parameters in the set are consistent
     ************************************************************************ */
    virtual void setModelParameters( ParametersSet & parameters ) = 0;

    /** ************************************************************************
     * getNumberStates
     * @return  The number of states a model has
     ************************************************************************ */
    virtual int getNumberStates( int symbolCategory ) const = 0;

    /** ************************************************************************
     * getNumberSymbols
     * @return  The number of distinct symbols recognized by the model
     ************************************************************************ */
    virtual int getNumberSymbols( int symbolCategory ) const = 0;

    /** ************************************************************************
     * getSymbolNumber
     * @input   base, a string representing a state of the model
     * @return  The basenumber of the state
     *           e.g. "A" = 1 , "C"="2" , "G"=3 and U="4" for the DNA models
     ************************************************************************ */
    virtual int getSymbolNumber( const string & base, int symbolCategory ) const = 0;

    /** ************************************************************************
     * getSymbol
     * @input   basenumber, the index of a state of the model
     * @return  The string which represents the state
     ************************************************************************ */
    virtual string getSymbol( int basenumber, int symbolCategory ) const = 0;

    /** ************************************************************************
     * getName
     * @return  The name of the model
     ************************************************************************ */
    virtual string getName() const = 0;

    /** ************************************************************************
     * getFrequency
     * @intput  residue, the index of a state of the model
     * @return  The frequencies of the given state of the model
     ************************************************************************ */
    virtual double getFrequency( int residue, int rateCategory,
                                 int symbolCategory ) const = 0;

    /** ************************************************************************
     * getFrequency
     * @intput  sbase, a STATE of the model
     * @return  The frequencies of the given state of the model
     ************************************************************************ */
    //    virtual double getFrequency( const string & sbase, int rateCategory ) const = 0;

    /** ************************************************************************
     * getFrequencies
     * @input   freq, an instance of vector <double>
     * @output  freq, an instance of vector <double> filled with the frequencies
     ************************************************************************ */
    //  virtual void getFrequencies( vector < double > & freq ) const = 0;

    /** ************************************************************************
     * getNumberRateCategories
     * @return  The number of rate categories a model has
     ************************************************************************ */
    virtual int getNumberRateCategories( int symbolCategory ) const = 0;

    /** ************************************************************************
     * getRateCategoryProbability
     * @input   category, the index of a category
     * @return  The probability of a particular rate category
     ************************************************************************ */
    virtual double getRateCategoryProbability( int category, int symbolCategory ) const = 0;

    /** ************************************************************************
     * getAverageMutationRate
     * @return  The average rate of mutation
     ************************************************************************ */
    virtual double getAverageMutationRate( int symbolCategory ) const = 0;

    /** ************************************************************************
     * SetAverageMutationRate
     * @semantics  Change the average rate of mutation for the given category
     ************************************************************************ */
    virtual void setAverageMutationRate( double newAverageRate, int symbolCategory ) = 0;


    /** ************************************************************************
     * getProportionInvariantSites
     * @output  prop, an instance of vector <double> filled with the proportions
     ************************************************************************ */
    // virtual void getProportionInvariantSites( vector < double > & prop ) const = 0;

    /** ************************************************************************
     * getGammaShapeParameter
     * @return  the gamma shape parameter alpha
     ************************************************************************ */
    virtual double getGammaShapeParameter( int symbolCategory ) const = 0;

    /** ************************************************************************
     * getAllParameters
     * @input   pars, an instance of vector <double>
     * @output  pars, an instance of vector <double> filled with all the
     *          parameters of the model reparametrized so as to be independant
     ************************************************************************ */
    virtual void getAllParameters( vector < double > & pars ) const = 0;

    /** ************************************************************************
     * setAllParameters
     * @input  pars, an instance of vector <double>, filled with all the new
     *         parameters of the model
     ************************************************************************ */
    virtual void setAllParameters( const vector < double > & pars ) = 0;

    /** ************************************************************************
     * getNumberFreeParameters
     * @input   The number of free parameters in the model (and
     *          the size of the pars vector above
     ************************************************************************ */
    virtual unsigned int getNumberFreeParameters() = 0;
    
    /** ************************************************************************
     * probability
     * @input   oldState, a state of the model
     * @input   newState, a state of the model
     * @input   time, a delay value
     * @input   category, a category
     * @return  The probability of ....
     ************************************************************************ */
    virtual double probability( int oldState, int newState, double time,
    int category, int symbolCategory ) const = 0 ;

    /** ************************************************************************
     * diffProbability
     * @input   oldState, a state of the model
     * @input   newState, a state of the model
     * @input   time, a delay value
     * @input   category, a category
     * @return  The probability of ....
     ************************************************************************ */
    virtual double diffProbability( int oldState, int newState,
    double time, int category, int symbolCategory ) const = 0;

    /** ************************************************************************
     * secondDiffProbability
     * @input   oldState, a state of the model
     * @input   newState, a state of the model
     * @input   time, a delay value
     * @input   category, a category
     * @return  The probability of ....
     ************************************************************************ */
    virtual double secondDiffProbability( int oldState,
    int newState, double time, int category, int symbolCategory ) const = 0;

    virtual const array2D< double > &
    getEquivalencyTable( int symbolCategory ) const = 0;

    /** ************************************************************************
     * printParameters
     * @input      outputStream, the stream used to output the parameters
     * @semantics  Output the parameters names and values of the model in a
     *             readable format
     ************************************************************************ */
    virtual void printParameters( ostream & outputStream ) const = 0;

    /** ************************************************************************
     * printLine
     * @input      outputStream  The stream used to output the parameters
     * @semantics  output the parameters values of the model in a single line
     ************************************************************************ */
    virtual void printLine( ostream & outputStream ) = 0;

    /** ************************************************************************
     * fromLine
     * @input the new parameters of the model as they are printed with
     *        printLine (ie, same order)
     ************************************************************************ */
    virtual void fromLine( const vector<double>& parameters ) = 0;
            
    /** ************************************************************************
     * getNumberLineParameters
     * @input the number of parameters printed/read with printLine and fromLine
     ************************************************************************ */
    virtual unsigned int getNumberLineParameters() = 0;
        
    /** ************************************************************************
     * initialisePerturbator
     * @input      perturbation parameters
     * @semantics  Warn the model that it will have to perturb itself and
     *             provide some useful perturbation parameters
     ************************************************************************ */
    virtual void initialisePerturbator( ParametersSet & parameters ) = 0;

    /** ************************************************************************
     * printPerturbationParameters
     * @input      outputStream, the stream used to output the parameters
     * @semantics  Output the perturbation parameters names and values of the
     *             model in a readable format
     ************************************************************************ */
    virtual void printPerturbationParameters( ostream & outputStream ) = 0;

    /** ************************************************************************
     * perturb
     * @semantics  ask the model to perturb some of its parameters
     ************************************************************************ */
    virtual double perturb() = 0;

    /** ************************************************************************
     * validatePerturbation
     * @input      validation status : true to validate the change from the
     *             last perturbation, false otherwise
     * @semantics  according to the validation flag, restore the model
     *             parameters as they were before the last call to perturb
     *             or validate the changes.
     ************************************************************************ */
    virtual bool validatePerturbation( bool validation ) = 0;
    
    /** ************************************************************************
     * stopBurn
     * @input      none
     * @semantics  this function is called to inform the model that the burnin
     *             period is finished while performing a MCMC run
     ************************************************************************ */
    virtual void stopBurn() = 0;
        

protected:

};
#endif




