#ifndef RNAMODEL_H
#define RNAMODEL_H

#include <vector>
using namespace std;

#include "Models/MatrixModel.h"
template < class Q > class array2D;
class Perturbator;
class ParametersSet;
class SequenceTable;

class RnaModel : public MatrixModel {

protected :
    RnaModel( const string & registrationName );

    RnaModel( ParametersSet & parameter, int matrixSize );

    array2D<int> matrixIndex;
    
     virtual void setEigenMatrix() = 0;
   
public :
    virtual ~RnaModel();

    /** ************************************************************************
     * getSymbolNumber
     * @input   base, a string representing a symbol of the model
     * @return  The basenumber of the symbol
     *           e.g. "AU"=0 , "CG"=5 , "UU"=15 and "Y-"=62
     ************************************************************************ */
    int getSymbolNumber( const string & pairedBase, int symbolCategory = 0 ) const;

    
    /** ************************************************************************
     * getSymbol
     * @input   symbolNumber, the index of a symbol of the model
     * @return  The string which represents the state
     ************************************************************************ */
    string getSymbol( int symbolNumber, int symbolCategory = 0 ) const;

    /** ************************************************************************
     * getState
     * @input   stateNumber, the index of a state of the model
     * @return  The string which represents the state
     ************************************************************************ */
    virtual string getState( int stateNumber, int symbolCategory = 0 ) const = 0;

    /** ************************************************************************
     * getNumberSymbols
     * @return  The number of distinct symbols recognized by the model
     ************************************************************************ */
    inline int getNumberSymbols( int symbolCategory = 0 ) const {
        assert( symbolCategory == 0 );
        return 64;
    }

    /** ************************************************************************
     * getNumberStates
     * @return  The number of states a model has
     ************************************************************************ */
    inline int getNumberStates( int symbolCategory = 0) const {
        assert( symbolCategory == 0 );
        return matrixSize;
    }
      

    /** ************************************************************************
     * retrieveEmpiricalRates
     * @input      sequenceTable, the SequenceTable instance containing the
     *             species' sequences used to perform the inference.
     * @input      a pointer to a double to store the proportion of invariant
     *             site (not compulsory)
     * @return     a rate ratio vector of length 5 initialised according to the
     *             sequences in the sequenceTable
     ************************************************************************ */
    array2D< double > retrieveEmpiricalRates( SequenceTable * sequenceTable,
    int modelId, double * propInvariant = NULL );

    /** ************************************************************************
     * retrieveEmpiricalFrequencies
     * @input      sequenceTable, the SequenceTable instance containing the
     *             species' sequences used to perform the inference.
     * @semantics  initialises the frequencies vector from the sequence
     ************************************************************************ */
    vector < double > retrieveEmpiricalFrequencies( SequenceTable * sequenceTable, int modelId );

    /** ************************************************************************
     * probability
     * @input   oldState, a state of the model
     * @input   newState, a state of the model
     * @input   time, a delay value
     * @input   category, a category
     * @return  The probability of ....
     ************************************************************************ */
    double probability( int oldState, int newState, double time,
    int category, int symbolCategory = 0) const;

    /** ************************************************************************
     * diffProbability
     * @input   oldState, a state of the model
     * @input   newState, a state of the model
     * @input   time, a delay value
     * @input   category, a category
     * @return  The probability of ....
     ************************************************************************ */
    double diffProbability( int oldState, int newState, double time,
    int category, int symbolCategory = 0 ) const;

    /** ************************************************************************
     * secondDiffProbability
     * @input   oldState, a state of the model
     * @input   newState, a state of the model
     * @input   time, a delay value
     * @input   category, a category
     * @return  The probability of ....
     ************************************************************************ */
    double secondDiffProbability( int oldState, int newState, double time,
    int category, int symbolCategory = 0 ) const;
    
    /** ************************************************************************
     * printParameters
     * @input      outputStream, the stream used to output the parameters
     * @semantics  Output the parameters names and values of the model in a
     *             readable format.
     *             -frequencies of the state
     *             -gamma shape parameter (if relevant)
     *             -proportion of invariant sites (if relevant)
     *             -mutation rate for each category
     ************************************************************************ */
    virtual void printParameters( ostream & outputStream ) const;

    /** ************************************************************************
     * setAverageMutationRate
     * @input  newAverageRate the new average rate of mutation
     ************************************************************************ */
    virtual void setAverageMutationRate( double newAverageRate, int symbolCategory = 0 );
    
private:
    int matrixSize;

};

#endif //RNAMODEL_H
