#ifndef NODE_H
#define NODE_H

#include <string>
#include <vector>

#include "SequenceTable.h"

#include "Util/array3D.h"

class Tree;

template < class Q >
class array2D;

class randombox;

class depthmap;

class Model;

class InferenceTree;


class Node {

public:
    /** .***********************************************************************
     * Node
     * @semantic  constructor of an internal node
     ************************************************************************ */
    Node( Tree * ptree );

    /** ************************************************************************
     * Node
     * @semantic  constructor of a leaf node
     ************************************************************************ */
    Node( string label, Tree * ptree  );

    /** ************************************************************************
     * ~Node
     * @semantic  destructor
     ************************************************************************ */
    virtual ~Node();

    /** ************************************************************************
     * initialisation
     * @input     leafNode, a boolean to indicate if the node is a leaf
     * @semantic  initialisation of a node once the tree has got a sequence
     *            table and a model
     ************************************************************************ */
    void initialisation( bool leafNode );


    // Display Methods. These methods proceed recursively from a given
    // node (usually the root) to create a string which symbolises the tree
    // topology e.g. ((1,2),3) including different kind of information

    /** ************************************************************************
     * toStringTopology
     * @semantics  This method outputs topology and leafs' labels
     ************************************************************************ */
    string toStringTopology();

    /** ************************************************************************
     * toStringNumbered
     * @semantics  This method outputs topology and leafs' number
     ************************************************************************ */
    string toStringNumbered();

    /** ************************************************************************
     * toStringClade
     * @semantics  This method outputs topology and leafs' label but stops
     *             its descent once a clade in cladeRep is reached
     ************************************************************************ */
    string toStringClade( vector < int > * claderep, int numclades );

    /** ************************************************************************
     * toString
     * @ semantics This method outputs topology, distances and leafs' labels
     ************************************************************************ */
    string toString();


    // Node Query Methods

    inline bool isLeaf(){
        return ( ( child[0] == NULL ) && ( child[1] == NULL ) && ( child[2] == NULL ) );
    }

    bool simulate( const vector< vector<int> > & categories,
                   const vector< vector<int> > & parent,
                   vector< vector < vector<int> > > & result );

    // Partial Likelihood calculation methods
    void likelihood();
    void difflikelihood( int branch );
    void diff2likelihood( int branch );
    void codiff_likelihood( int diffbranch1, int diffbranch2 );


    bool matches_clade( vector < int > &, string & );
    void build_clade( int, SequenceTable & );


    // Topology manipulation methods
    bool NNI( int branch, int abranch, int cbranch );
    void duplicate( Node & );

    // Node Members
    Node & operator = ( Node & );

    // The label of the node
    string label;

    // A link to the parent
    Node * parent;

    // The distance from the parent
    double pdistance;


    // Node "children"
    Node * child[3];

    // Distances to "children"
    double cdistance[3];

    // Partial likelihood storage for each sequences
    // (ie. each model in this case)
    // dim array : site * nucleotide * rate
    vector < array3D< double > > partialLikelihood;

    vector < int > clade;

    //Keeps track of whether the node likelihoods should be recalculated
    bool calculated;

    //Necessary for cluster implementation
    bool clustered;

    //Pointer to the sequence of the node
    string * sequence;

    // to speed up the likelihood computation :
    // since for a given ancestor lots of sites have not changed among its
    // children,
    // for each site, we store the common state in identical
    // (and we store -1 if there is a variation)
    // size(identical[i]) == t_model->getSequencesLength()
    // invariantSites are identical and are treated separatly
    // in lastCalc we store the site where the computation has been done for the
    // first and the last time, with a given identical state for all descendants
    // size(lastCalc[i]) == t_model->getNumberStates()
    // There is a vector for each model (size(identical)==size(lastCalc)==
    // numberModelsUsed
    vector< vector<int> > identical;
    vector< vector<int> > lastCalc;

    //Position in vector indexed structure
    unsigned int index;
    int cnumber;

    //Pointer to the tree
    Tree * ptree;

    static Node * create_node( depthmap & map, int start, Tree * ptree   );
    static Node * makenode( string & nodeString, Tree * ptree  );

    /** ************************************************************************
     * makeRoot
     * @preconditions  The node is not a leaf node
     * @semantics      The internal node becomes the new root of the tree
     ************************************************************************ */
    inline bool makeRoot() {
        assert( !isLeaf() );
        return makeParent( NULL, 0.0 );
    }

    /** ************************************************************************
     * computeIdentical
     * @preconditions  the tree that contains the node has a sequence table
     *                 loaded and THE CHILDREN IDENTICAL IS UP TO DATE
     * @semantics      The internal node becomes the new root of the tree
     ************************************************************************ */
    void computeIdentical();

    void recursiveCopy( Node * srcNode, Node * parent );


protected:
    Node();
    void loadleafdata();
    void destroy();
    void optimizedLikelihoodLeafParent();

private:

    /** .***********************************************************************
     * makeparent
     * @semantics  Recursive function used to change the root of the tree in
     *             makeRoot. While the old root is not reached, the current
     *             parent become a child and the new parent is the one given in
     *             parameter. This function is then called for the old parent
     *             so as to :
     *             1) let it know its child (newParent) is its parent now.
     *             2) make its own parent its child
     ************************************************************************ */
    bool makeParent( Node * newParent, double dist  );

private:
// variable to speed-up the likelihood computation ( to avoid
// memory allocation/desallocation at each call )

    //look-up table : lookup[child][symbolCat][rate]
    array3D< double > * lookup[3];
};


#endif




