#ifndef OPTIMISE_H
#define OPTIMISE_H

#include <math.h>
#include <vector>

class InferenceTree;
class Model;


using namespace std;


class Optimise {
private:
    static double lineSearch( InferenceTree & tree, Model & smodel, vector < double > & point,
    vector < double > & direction, vector < double > & gradient, bool domodel );

    static vector < double > diffLoglikelihoodSqrt( InferenceTree & ptree ,
    vector < double > & branchesLengthSqrt );

public:
    /** .***********************************************************************
     * optimiseQuasiNewton
     * @input      ptree, the phylogenetic tree
     * @input      start, an instance of vector that contains the actual length
     *             of the branches of the tree AND the model parameters
     * @input      direction, an instance of vector that contains the direction
     *             used to modify the actual branch length AND the model params
     * @input      magnitude, the factor used to modify start with direction
     * @return     the loglikelyhood of the tree with the new branches length
     *             and new model parameters
     * @semantics  Optimise the branches (and optionally the model parameters)
     *             of a phylogenetic tree.
     ************************************************************************ */
    static vector < double > optimiseQuasiNewton( InferenceTree & ptree, Model & smodel,
    vector < double > & , bool , double tolerance, bool display );

    /** ************************************************************************
     * loglikelihood
     * @input      ptree, the phylogenetic tree
     * @input      smodel, the model of evolution used
     * @input      start, an instance of vector that contains the actual length
     *             of the branches of the tree AND the model parameters
     * @input      direction, an instance of vector that contains the direction
     *             used to modify the actual branch length AND the model params
     * @input      magnitude, the factor used to modify start with direction
     * @return     the loglikelyhood of the tree with the new branches length
     *             and new model parameters
     * @semantics  The parameters of the current model are modified during
     *             execution but actual parameters are restored on exit.
     ************************************************************************ */
    static double loglikelihood( InferenceTree & ptree, Model & smodel,
    vector < double > & start, vector < double > & direction,
    double magnitude );

    /** ************************************************************************
     * loglikelihood
     * @input      ptree, the phylogenetic tree
     * @input      start, an instance of vector that contains the actual length
     *             of the branches of the tree
     * @input      direction, an instance of vector that contains the direction
     *             used to modify the actual branch length
     * @input      magnitude, the factor used to modify start with direction
     * @return     the loglikelyhood of the tree with the new branches length
     ************************************************************************ */
    static double loglikelihood( InferenceTree & ptree , vector < double > & start,
    vector < double > & direction, double magnitude );

    /** ************************************************************************
     * modelPartialGradients
     * @input      ptree, the phylogenetic tree
     * @input      branchesLength, an instance of vector that contains the
     *             length of the branches of the tree
     * @input      smodel, the model of evolution
     * @input      modelParameters, the point were we look for the gradiant
     *             vector
     * @return     The vector of partial gradients of the model parameters
     *             at the point modelParameters.
     * @semantics  The parameters of the current model are modified during
     *             execution but actual parameters are restored on exit.
     ************************************************************************ */
    static vector < double > modelPartialGradients( InferenceTree & ptree,
         vector < double > & branchesLength,
         Model & smodel, vector < double > & modelParameters, double step );
};

#endif //OPTIMISE_H




