#ifndef TOPOLOGYOPTIMISE_H
#define TOPOLOGYOPTIMISE_H

#include <iostream>
#include <fstream>

#include "Tree/InferenceTree.h"

#include "Models/Model.h"

#include "Optimise.h"

#include "statlib.h"

class ClustersSet;

class SequenceTable;

class TreesSet;

/** ****************************************************************************
 * This class is used by mlphase, it contains algorithm to search the ML
 * trees in the topologies space
 **************************************************************************** */

class TopologyOptimise {
public:

/** ****************************************************************************
 * exhaustiveSearch
 * @input         sequenceTable, the DNA/RNA/... sequences
 * @input         model, the initialised substitution model used
 * @input         optimiseModel, a boolean used to indicate if the model
 *                parameters should be optimised
 * @input         treesSet, a TreesSet to store the trees
 * @input         out, an output stream to send the outputs
 * @input         clusters, facultative parameter to constrain the search
 *                in the tree space with compulsory monophyletic clades
 *                (see ClustersSet for details). All the species must 
 *                appear in the cluster even if they are represented as leaves
 * @precondition  model must be initialised, if optimiseModel is set to false
 *                the model should have been initialised with a file
 * @semantics     This function generates and optimises every possible tree,
 *                using the supplied substitution model, in order to find the
 *                best one (with corresponding branch lenghts and model
 *                parameters if optimiseModel == true)
 **************************************************************************** */
    static void exhaustiveSearch( SequenceTable * sequenceTable, Model * model,
    bool optimiseModel, TreesSet* treesSet, ostream & out );
    static void exhaustiveSearch( SequenceTable * sequenceTable, Model * model,
    bool optimiseModel, TreesSet* treesSet, ostream & out,
    ClustersSet& clusters );
    
/** ****************************************************************************
 * branchAndBoundSearch
 * @input         sequenceTable, the DNA/RNA/... sequences
 * @input         model, the initialised substitution model used
 * @input         optimiseModel, a boolean used to indicate if the model
 *                parameters should be optimised
 * @input         treesSet, a TreesSet to store the trees
 * @input         out, an output stream to send the outputs
 * @input         clusters, facultative parameter to constrain the search
 *                in the tree space with compulsory monophyletic clades
 *                (see ClustersSet for details). All the species must 
 *                appear in the cluster even if they are represented as leaves
 * @precondition  model must be initialised, if optimiseModel is set to false
 *                the model should have been initialised with a file
 * @semantics     This function is aiming at finding the best tree
 *                (topology+branch length [+model parameters] ). Some trees are
 *                skipped thanks to the branch and bound method (if the best
 *                possible likelihood of a subtree is lower than the best
 *                likelihood find so far there is no point in adding more
 *                species to it). This function would always find the best tree
 *                if it was impossible for the optimiser to be trapped
 *                in local minima.
 **************************************************************************** */
    static void branchAndBoundSearch( SequenceTable * sequenceTable, Model * model,
    bool optimiseModel, TreesSet* treesSet, ostream & out );
    static void branchAndBoundSearch( SequenceTable * sequenceTable, Model * model,
    bool optimiseModel, TreesSet* treesSet, ostream & out,
    ClustersSet& clusters );

/** ****************************************************************************
 * buildTreeStepwise
 * @input         sequenceTable, the DNA/RNA/... sequences
 * @input         model, the initialised substitution model used
 * @input         optimiseModel, a boolean used to indicate if the model
 *                parameters should be optimised
 * @input         treesSet, a TreesSet to store the trees
 * @input         out, an output stream to send the outputs
 * @input         clusters, facultative parameter to constrain the search
 *                in the tree space with compulsory monophyletic clades
 *                (see ClustersSet for details). All the species must 
 *                appear in the cluster even if they are represented as leaves
 * @precondition  model must be initialised, if optimiseModel is set to false
 *                the model should have been initialised with a file
 * @semantics     This function aims at finding the ML tree. I will generate one
 *                tree. Branches are added incrementally at the best position
 *                according to the best likelihood found. This is an heuristic
 *                method: there is no guarantee that the best tree is found
 *                which would be quite surprising.
 **************************************************************************** */
    static double buildTreeStepwise( SequenceTable * sequenceTable, Model * model,
    bool optimiseModel, TreesSet* treesSet, ostream & out );
    static double buildTreeStepwise( SequenceTable * sequenceTable, Model * model,
    bool optimiseModel, TreesSet* treesSet, ostream & out,
    ClustersSet& clusters );
};

#endif //TOPOLOGYOPTIMISE_H




