#include "Consensus.h"

#include "Tree/InferenceTree.h"
#include "Tree/Node.h"
#include "Sequence/SequenceTable.h"
#include "Models/Model.h"
#include "Util/FileParser.h"
#include "Util/ParametersSet.h"

#include "PatternDesign/Singleton.h"
#include "Util/ModelFactory.h"

#include <iomanip>

#include <vector>
#include <map>

using namespace std;

Consensus::Consensus(){
    consensusTree = NULL;
    consensusModel1 = NULL;
    consensusModel2 = NULL;
}
        
int Consensus::run( const string& mcmcControlFileName,
                    const string& consensusTopologyFileName ){
    FileParser fileParser( mcmcControlFileName );
    ParametersSet controlMCMC = fileParser.retrieveParametersSet();
    
    expectedNumberSamples = controlMCMC.intParameter("Sampling iterations");
    expectedNumberSamples = expectedNumberSamples /
                    controlMCMC.intParameter("Sampling period");
            
    ifstream consensusTopologyFile( consensusTopologyFileName.c_str() );
    //read the consensus tree
    string stringConsensusTopology;
    FileParser::readTree(consensusTopologyFile, stringConsensusTopology);
    consensusTopologyFile >> ws;
    if ( !consensusTopologyFile.eof() ||
         (stringConsensusTopology.length() == 0) ) {
        cerr << "error while reading the consensus tree file " << consensusTopologyFileName << endl;
        exit( 1 );
    }

    consensusTree = new Tree( stringConsensusTopology );
    consensusTree->unroot();
    consensusTree->createIndex();
    string outgroup = controlMCMC( "DATAFILE" ).stringParameter( "Outgroup" );
    SequenceTable seq(controlMCMC( "DATAFILE" ));
    //The consensus topology must be given in the numbered format,
    //transform the specie's IDs into the specie's names
    for( deque < Node * >::iterator iter = consensusTree->leafRefVector.begin();
         iter != consensusTree->leafRefVector.end();
         ++iter ){
        (*iter)->cnumber = atoi( ((*iter)->label).c_str() );
        (*iter)->label = seq.species[(*iter)->cnumber-1];
    }
    
    string basename = controlMCMC.stringParameter( "Output file" );
    string topologyFileName = basename + string(".samples");
    string branchLengthsFileName = basename + string(".bl");
    ifstream topologyFile( topologyFileName.c_str() );
    ifstream branchLengthsFile( branchLengthsFileName.c_str() );
    
    string consensusTreeFileName = basename + string(".consensus-tree");
    ofstream consensusTreeFile( consensusTreeFileName.c_str() );
     
    Singleton < ModelFactory > & modelFactory = Singleton < ModelFactory >::instance();
    consensusModel1 = modelFactory.create( controlMCMC("MODEL").
                                                     stringParameter( "Model" ),
                                           controlMCMC( "MODEL" ) );
    consensusModel2 = modelFactory.create( controlMCMC("MODEL").
                                                     stringParameter( "Model" ),
                                           controlMCMC( "MODEL" ) );
    consensusModel1->initialisation( NULL );
    consensusModel2->initialisation( NULL );

    consensusTree->standardizeTree( outgroup );

    
    string modelParametersFileName = basename + string(".mp");
    ifstream modelParametersFile( modelParametersFileName.c_str() );
        
    string consensusModelFileName1 = basename + string(".consensus-model-1");
    string consensusModelFileName2 = basename + string(".consensus-model-2");
    string plotBaseName = basename + string(".pdf-");

    ofstream consensusModelFile1;
    ofstream consensusModelFile2;
    FileParser::confirmOpenFile( consensusModelFile1, consensusModelFileName1 );
    FileParser::confirmOpenFile( consensusModelFile2, consensusModelFileName2 );
             
    modelParametersConsensus( modelParametersFile,
                              consensusModel1, consensusModel2,
                              plotBaseName );

    branchLengthsConsensus( topologyFile, branchLengthsFile, consensusTree );
    
    ParametersSet consensusParameters( "consensusParameters", false );
    
    cout << "Consensus 1 for the model" << endl;
    consensusModel1->printParameters( cout );
    consensusParameters = consensusModel1->getModelParameters();
    consensusParameters.saveToFile(consensusModelFile1);
    consensusModelFile1.close();
    cout << "Consensus 2 for the model" << endl;
    consensusModel2->printParameters( cout );
    consensusParameters = consensusModel2->getModelParameters();
    consensusParameters.saveToFile(consensusModelFile2);
    consensusModelFile1.close();
    
    cout << "Consensus for the tree" << endl;
    cout << consensusTree->toString( true );
    if (controlMCMC.stringParameter("Output format") != "bambe" ){
        cout << ';' << endl;
    }
    else{
        cout << endl;
    }  
    consensusTreeFile << consensusTree->toString( true ) << endl;
    consensusTreeFile.close();
    
    return(0);
}
        
int Consensus::modelParametersConsensus( ifstream& modelParametersFile,
                              Model* modelConsensus1, Model* modelConsensus2,
                              const string& plotBaseName ){
    unsigned int numberParameters = modelConsensus1->getNumberLineParameters();
    vector<double> consensusParameters1( numberParameters );
    vector<double> consensusParameters2 = consensusParameters1;
    ofstream outputPDFFile[numberParameters];
    
    char outputPDFFileName[100];
    char* numberPos;
    sprintf( outputPDFFileName, "%s", plotBaseName.c_str() );
    numberPos = outputPDFFileName;
    while(*(++numberPos));
    for ( unsigned int i = 0; i < numberParameters; ++i ) {
        sprintf( numberPos, "%d", i+1 );
        outputPDFFile[i].open(outputPDFFileName);
    }
    
    vector< map< double , int > > sortedValues(numberParameters);
    
    double value;
    int numberInstances = 0;
    
    modelParametersFile >> ws;
    while ( !modelParametersFile.eof() ){
        for ( unsigned int i = 0; i < numberParameters; ++i ) {
            modelParametersFile >> value;
            if ( modelParametersFile.eof() ) {
                cerr << "error while reading the \".mp\" file, sample no:"
                << numberInstances + 1 << ", param no :" << i + 1 << endl;
                exit( 1 );
            }
            consensusParameters1[i] += value;
            ++((sortedValues[i])[value]);
        }
        ++numberInstances;
        if (numberInstances%10000==0){
            cout << numberInstances << endl;
        }
        modelParametersFile >> ws;
    }
    if (numberInstances != expectedNumberSamples){
        cerr << "Warning: the sample size (" << numberInstances
             << ") and the expected size (" << expectedNumberSamples
             << ") are different" << endl;
    }
    
    map< double, int >::iterator iter;
    
    // consensus for each parameter
    for ( unsigned int i = 0; i < numberParameters; ++i ) {
        //mean for the 1st consensus
        consensusParameters1[i] /= numberInstances;
        
        //median for the 2nd consensus
        double medianValue = ((double)numberInstances)/2.0;
        iter = sortedValues[i].begin();
        int sampleCounter = (*iter).second;
        while ( (double)sampleCounter < medianValue ){
            ++iter;
            sampleCounter += (*iter).second;
        }
        consensusParameters2[i] = (*iter).first;
       
        //max for the 2nd consensus
        //double maxFreq = 0.0;
        
        //PDF of the parameter        
        double minValue = (*(sortedValues[i].begin())).first;
        double maxValue = (*(--sortedValues[i].end())).first;
        double halfBinSize = (maxValue-minValue)*100.0/(double)numberInstances;
        outputPDFFile[i].setf(ios::fixed);
        outputPDFFile[i] << setprecision(4);
        //first empty bin
        outputPDFFile[i] << setw(13) << minValue-halfBinSize << ' '
                         << setw(13) << 0.0 << endl;
        double binMaxValue = minValue + (2 * halfBinSize);
        int counter = 0;
        for ( iter = sortedValues[i].begin();
              iter != sortedValues[i].end(); ++iter){
              //add the number of item with that value to the actual bin
              if( (*iter).first <= binMaxValue ){
                  counter += (*iter).second;
              }
              //or finish the bin and start the next one if we pass
              // the upper boundary
              else{
                  //output result for this bin
                  outputPDFFile[i] << setw(13) << binMaxValue-halfBinSize
                                   << ' ' << setw(13)
                                   << (double)counter/(double)numberInstances
                                   << endl;
                  
//                  if ( ((double)counter/(double)numberInstances) > maxFreq ){
//                      consensusParameters2[i] = binMaxValue-halfBinSize;
//                      maxFreq = (double)counter/(double)numberInstances;
//                  }
                  //prepare the next bin
                  counter = (*iter).second;
                  binMaxValue += 2*halfBinSize;
              }
        }
        //last empty bin
        outputPDFFile[i] << setw(13) << binMaxValue-halfBinSize << ' '
                         << setw(13) << 0.0 << endl;
        outputPDFFile[i].close();
    }
    modelConsensus1->fromLine( consensusParameters1 );
    modelConsensus2->fromLine( consensusParameters2 );

    return 0;
}
        
int Consensus::branchLengthsConsensus( ifstream& topologyFile,
                                       ifstream& branchLengthsFile,
                                       Tree* consensusTopology ){

    string stree;
    int numberSimilarTopology = 0;
    string output;

    int numberInstances = 0;

    string stringConsensusTopology = consensusTopology->toStringNumbered( true );
    cout << "Consensus topology : " << endl;
    cout << stringConsensusTopology << endl;
    
    vector< double > branchesLength( consensusTopology->getNumberBranches() );
    vector< double > consensusBranchesLength( consensusTopology->getNumberBranches() );
    for ( int i = 0; i < consensusTopology->getNumberBranches(); ++i ) {
        consensusBranchesLength[i] = 0;
    }
    

    while ( !topologyFile.eof() ) {
        FileParser::readTree( topologyFile, stree );
        if (stree.length() == 0 ){
            cerr << "error while reading the \".sample\" file, sample no:"
                  << numberInstances + 1 << endl;
                exit( 1 );
        }

        for ( unsigned int i = 0; i < branchesLength.size(); ++i ) {
            branchLengthsFile >> branchesLength[i];
            if ( branchLengthsFile.eof() ) {
                cerr << "error while reading the \".bl\" file, sample no:"
                << numberInstances + 1 << ",branchnb:" << i + 1 << endl;
                exit( 1 );
            }
        }
        if ( stringConsensusTopology == stree ) {
            ++numberSimilarTopology;
            for ( unsigned int i = 0; i < branchesLength.size(); ++i ) {
                consensusBranchesLength[i] += branchesLength[i];
            }
        }
        topologyFile >> ws;
        ++numberInstances;
        if (numberInstances%10000==0){
            cout << numberInstances << endl;
        }
    }
    branchLengthsFile >> ws;
    if (numberSimilarTopology){
        cout << numberSimilarTopology << " trees in the sample have the same "
             << "topology than the consensus tree" << endl;
        for ( unsigned int i = 0; i < consensusBranchesLength.size(); ++i ) {
            consensusBranchesLength[i] = consensusBranchesLength[i]/
                                         ( (double)numberSimilarTopology );
        }
    }
    else{
        cout << "Sorry, cannot find a tree in the sample with the same "
             << "topology than your consensus tree" << endl;
        for ( unsigned int i = 0; i < consensusBranchesLength.size(); ++i ) {
            consensusBranchesLength[i] = 0;
        }
    }      
    consensusTopology->setBranchVector( consensusBranchesLength );
    if (numberInstances != expectedNumberSamples){
        cerr << "Warning: the sample size and the expected size are different" << endl;
    }
    assert( branchLengthsFile.eof() );
    return 0;
}


int main( int argc, char * * argv ) {
    if ( argc != 3 ) {
        cerr << "usage: " << argv[0]
        << " mcmc_control_file consensus_tree_file" << endl;
        exit( 1 );
    }
    Consensus consensus;
    return consensus.run( string( argv[1]), string( argv[2] ) );
}
