#include "MCMCphase.h"

#include <algorithm>

/*
#include <dlfcn.h>
*/

#include <iostream>
#include <ctype.h>
#include <functional>


#include "PatternDesign/Singleton.h"
#include "Util/randombox.h"

#include "Util/FileParser.h"
#include "Util/ParametersSet.h"

#include "Models/Model.h"
#include "Util/ModelFactory.h"

using namespace std;

class SequenceTable;

class InferenceTree;


ofstream plotfile;
ofstream samplefile;
ofstream output;
ofstream branchfile;
ofstream bestTreeFile;
ofstream bestModelParametersFile;
ofstream mpfile;

double MCMCphase::runMCMC( InferenceTree * startingTree, Model * startingModel,
double initialBranchStep,
int burninCycles, int samplingCycles, int samplingPeriod,
const string & outgroup, double branchLengthUpperBound, string terminator ) {

    Singleton < randombox > & randBox = Singleton < randombox >::instance();

    int numberBranches = startingTree->getNumberBranches();
    double prob;
    int globalSPRs = 0, globalAcceptedSPRs = 0;
    int globalNNIs = 0, globalAcceptedNNIs = 0;
    int localNNIs = 0, localAcceptedNNIs = 0;
    int totalLengthProposals = 0, totalAcceptedLengthProposals = 0;
    int modelAccepted = 0;

    int dynLengthProposals = 0;
    int dynAcceptedLengthProposals = 0;
    
    vector < double > branch;

    InferenceTree * proposedTree = NULL;
    InferenceTree * currentTree = startingTree;
    Model * model = startingModel;
    double branchStep = initialBranchStep;
    
    //Set outgroup
    currentTree->standardizeTree(outgroup) ; 
    //Invalidate partial likelihood calculations(to force fresh recalculations)
    currentTree->invalidateNodes();
    currentTree->invalidateLeaves();
    
    //initialise the best values
    vector < double > bestBranchVector(numberBranches);
    vector < double > bestModelParameters;
    model->getAllParameters( bestModelParameters );
    currentTree->getBranchVector( bestBranchVector );
    string bestStringTree = currentTree->toString(false);
    double bestLikelihood = currentTree->loglikelihood();
    
    double oldLikelihood = bestLikelihood;
    double newLikelihood = bestLikelihood;

    // Create proposed tree, a copy of the current/initial tree
    proposedTree = new InferenceTree( *currentTree );

    // Simple check just in case...
    assert( bestStringTree == proposedTree->toString(false) );
    assert( oldLikelihood == proposedTree->loglikelihood() );
    //invalidate and check again
    proposedTree->invalidateNodes();
    proposedTree->invalidateLeaves();
    assert( oldLikelihood == proposedTree->loglikelihood() );

    //precision 4 for the output of the likelihood during the run
    cout.setf(ios::fixed);
    cout << setprecision(4);

    //launch the chain for burninCycles + samplingCycles
    for ( int cycle = 0; cycle < ( burninCycles + samplingCycles ); ++cycle ){
        assert (oldLikelihood == newLikelihood);
        
        // Global topology changes
        if ( ( ( ( cycle + 1 ) % 10 ) == 0 ) ) {
            // subtree pruning and regrafting
            if ( randBox.ran() < 0.5 ) {
                ++globalSPRs;
                double frac = randBox.ran();
                double numer;
                int branchId;
                int branchDest;
                do {
                    //choose the branch to move
                    branchId = ( int )( numberBranches * randBox.ran() );
                    //choose the insertion point (can be any edge except
                    //branchId itself
                    branchDest = ( (int)((numberBranches-1)*randBox.ran()) )%
                                  numberBranches;
                    if ( branchDest>=branchId ){
                        ++branchDest;
                    }
                    //try the move
                    numer = proposedTree->insertBranch( branchId, branchDest,
                                                        frac );
                }
                while ( numer == -1.0 );
                proposedTree->standardizeTree( outgroup );
                // Find the "energy" of the new proposal tree
                newLikelihood = proposedTree->loglikelihood();
#ifdef DEBUG
                proposedTree->invalidateNodes();
	            assert(newLikelihood == proposedTree->loglikelihood());
#endif

                // calculate acceptance probablity
                prob = MIN( 1.0, exp( newLikelihood - oldLikelihood ) * numer );

                // Accept the new state
                if ( randBox.ran() < prob ) { 
                    ++globalAcceptedSPRs;
                    if ( newLikelihood > bestLikelihood ) {
                        bestStringTree = proposedTree->toString( false );
                        proposedTree->getBranchVector( bestBranchVector );
                        model->getAllParameters( bestModelParameters );
                        bestLikelihood = newLikelihood;
                    }
                    //change the actual state
                    *currentTree = *proposedTree;
                    oldLikelihood = newLikelihood;
                }
                // Reject new state
                else {
                    //cancel the change in proposed tree for the next
                    //perturbation
                    *proposedTree = *currentTree;
                    newLikelihood = oldLikelihood;
                }
            }

            /* Nearest Neighbor Interchange
            // proposal 1, prob = 0.5
            //   A           C               A           B
            //    \_________/                 \_________/
            //    /         \          =>     /         \
            //   B           D               C           D
            // proposal 2, prob = 0.5
            //   A           C               A           C
            //    \_________/                 \_________/
            //    /         \          =>     /         \
            //   B           D               D           B                 */

            else {
                ++globalNNIs;
                int branchId;
                do {
                    branchId = ( int )( numberBranches * randBox.ran() );
                }
                while ( !currentTree->is_internal_branch( branchId ) );
                double branchLength =
                    proposedTree->getBranchLength( branchId );
                proposedTree->setBranchLength( branchId, -branchLength );
                proposedTree->normalize( branchId,
                ( int )( 2.0 * randBox.ran() ), ( int )( 2.0 * randBox.ran() ) );


                // Find the "energy" of the new proposal tree
                proposedTree->standardizeTree( outgroup );
                newLikelihood = proposedTree->loglikelihood();
#ifdef DEBUG
                proposedTree->invalidateNodes();
	            assert(newLikelihood == proposedTree->loglikelihood());
#endif
                // calculate acceptance probablity
                prob = MIN( 1.0, exp( newLikelihood - oldLikelihood ) );
                // Accept new state
                if ( randBox.ran() < prob ) {
                    ++globalAcceptedNNIs;
                    if ( newLikelihood > bestLikelihood ) {
                        bestStringTree = proposedTree->toString( false );
                        proposedTree->getBranchVector( bestBranchVector );
                        model->getAllParameters( bestModelParameters );
                        bestLikelihood = newLikelihood;
                    }
                    *currentTree = *proposedTree;
                    oldLikelihood = newLikelihood;
                }
                // Reject new state
                else {
                    *proposedTree = *currentTree;
                    newLikelihood = oldLikelihood;
                }
            }
        } //end of global topology change

        // Select a random branch
        int branchId = ( int )( numberBranches * randBox.ran() );
        double oldLength = proposedTree->getBranchLength( branchId );
        double length = oldLength;

        length += ( branchStep * randBox.gasdev() ); // Normal dist. branch proposal

        if ( length > 2*branchLengthUpperBound ) {
            cout << "Out of bounds " << length << endl;
            int n = (int) (length/(2.0*branchLengthUpperBound));
            length -= 2.0*branchLengthUpperBound*(double)n;
            assert (length < 2.0 * branchLengthUpperBound);
            if ( length > branchLengthUpperBound ) {
                length = 2.0*branchLengthUpperBound-length;
            }
            cout << "New value = " << length << endl;
            assert( length <= branchLengthUpperBound );
        }
        proposedTree->setBranchLength( branchId, length );

        // If the proposed branch length is -ive change the topology using NNI
        // or rebound if it is a terminal edge
        int isNNI = 0;
        if ( length < 0.0 ) {
            isNNI = proposedTree->normalize( branchId,
            ( int )( 2.0 * randBox.ran() ), ( int )( 2.0 * randBox.ran() ) );
            if ( isNNI ) {
                ++localNNIs;
                proposedTree->standardizeTree( outgroup );
            }
        }
        if (!isNNI){
            ++totalLengthProposals;
            // increment the number of proposals in the current window
            ++dynLengthProposals;
       }

       
        // Find the "energy" of the new proposal tree
        newLikelihood = proposedTree->loglikelihood();
#ifdef DEBUG
        proposedTree->invalidateNodes();
        assert(newLikelihood == proposedTree->loglikelihood());
#endif
        prob = MIN( 1.0, exp( newLikelihood - oldLikelihood ) );
        // Accept new state
        if ( randBox.ran() < prob ) {
            if (isNNI){
                ++localAcceptedNNIs;
            }
            else{
                ++totalAcceptedLengthProposals;
                // increment number of accepted proposals in the current window
                ++dynAcceptedLengthProposals;
            }
            if ( newLikelihood > bestLikelihood ) {
                bestStringTree = proposedTree->toString( false );
                proposedTree->getBranchVector( bestBranchVector );
                model->getAllParameters( bestModelParameters );
                bestLikelihood = newLikelihood;
            }
            // Make the proposed tree the current tree in the Markov
            // Chain ( by the most efficient means)
            if ( isNNI ) {
                *currentTree = *proposedTree;
            }
            else{
                currentTree->setBranchLength( branchId, fabs(length) );
            }
            oldLikelihood = newLikelihood;
        }
        // Reject new state
        else {
            if ( isNNI ){
                *proposedTree = *currentTree;
            }
            else{
                proposedTree->setBranchLength( branchId, oldLength );
            }
            newLikelihood = oldLikelihood;
        }

        // Dynamically Modify branch step DURING BURN IN
        if ( cycle < burninCycles ) {
            if ( dynLengthProposals == numberBranches*10 ) {
                // If acceptance ratio is too low
                if ( ( (double)dynAcceptedLengthProposals / (double)dynLengthProposals ) < 0.2 ) {
                    branchStep /= 1.1; // If acceptance ratio is not
                    // acceptable lower the STD of the branch proposal distribution
                }
                // If acceptance ratio is too high
                if ( ( (double)dynAcceptedLengthProposals / (double)dynLengthProposals ) > 0.25 ) {
                    if ( (branchStep*10) < branchLengthUpperBound ){
                        branchStep *= 1.1;
                    }
                }
                dynAcceptedLengthProposals = 0;
                dynLengthProposals = 0;
            }
        }
        
        //if the burnin period is over, warn the model (to stop the
        //dynamical modification of the perturbation steps)
        if ( cycle == burninCycles){
            model->stopBurn();
        }
        
        

        double lnHastingRatio = model->perturb();
        // Force fresh partial likelihood calculations
        proposedTree->invalidateNodes();
	
        // Acceptance probability is exp(oenergy-nenergy) * MH_ratio
        newLikelihood = proposedTree->loglikelihood();
#ifdef DEBUG
        proposedTree->invalidateNodes();
        assert(newLikelihood == proposedTree->loglikelihood());
#endif
        prob = MIN( 1.0, exp( newLikelihood - oldLikelihood + lnHastingRatio ) );
        // Accept new state
        if ( randBox.ran() < prob ) {
            ++modelAccepted;
            model->validatePerturbation( true );
            if ( newLikelihood > bestLikelihood ) {
                bestStringTree = proposedTree->toString( false );
                proposedTree->getBranchVector( bestBranchVector );
                model->getAllParameters( bestModelParameters );
                bestLikelihood = newLikelihood;
            }
            // Save new partial likelihood calculation
            *currentTree = *proposedTree;
            oldLikelihood = newLikelihood;
        }
        // Reject new state
        else {
            model->validatePerturbation( false );
            // Restore old partial likelihood calculation
            *proposedTree = *currentTree;
            newLikelihood = oldLikelihood;
        }
#ifdef DEBUG
        proposedTree->invalidateNodes();
        currentTree->invalidateNodes();
#endif

        // output the plot
        if ( !((cycle+1)%samplingPeriod) ) {
            plotfile << -newLikelihood << endl;
        }
        if ( !((cycle+1)%100) ) {
            cout << (cycle+1) << ".LnLikelihood = "
                 << newLikelihood << endl;
            cout << currentTree->toStringNumbered( false ) << endl;
        }
        //sampling
        if ( (cycle>burninCycles) &&
             !((cycle+1)%samplingPeriod) ) {
            samplefile << currentTree->toStringNumbered( true ) <<
            terminator << endl;
            model->printLine( mpfile );
            mpfile << endl;
            currentTree->getBranchVector( branch );
            for ( unsigned int i = 0; i < branch.size(); ++i ){
                branchfile << setw(8) << branch[i] << ' ';
            }
            branchfile << endl;
        }
    }

    // Frequency Proposal Phase (End)
    cout.setf(ios::fixed);
    output.setf(ios::fixed);
    cout << setprecision(2);
    output << setprecision(2);
    cout << "Branch length proposal acceptance rate  =  "
         << (double)totalAcceptedLengthProposals*100.0 / (double)totalLengthProposals
         << "%" << endl;
    cout << "Model acceptance rate                   =  "
         << (double)modelAccepted*100.0 / (double)(burninCycles+samplingCycles)
         << "%" << endl;
    output << "Branch length proposal acceptance rate  =  "
           << (double)totalAcceptedLengthProposals*100.0 / (double)totalLengthProposals
           << "%" << endl;
    output << "Model acceptance rate                   =  "
           << (double)modelAccepted*100.0 / (double)(burninCycles+samplingCycles)
           << "%" << endl;
    
    cout << setprecision(6);
    output << setprecision(6);
    cout << "SPR acceptance rate                    =  "
         << (double)globalAcceptedSPRs*100.0 / (double)globalSPRs
         << "%" << endl;
    cout << "Number of SPR proposals                =  "
         << globalSPRs << endl;
    cout << "Global NNI acceptance rate             =  "
         << (double)globalAcceptedNNIs*100.0 / (double)globalNNIs
         << "%" << endl;
    cout << "Number of global NNI proposals         =  "
         << globalNNIs << endl;
    
    output << "SPR acceptance rate                    =  "
           << (double)globalAcceptedSPRs*100.0 / (double)globalSPRs
           << "%" << endl;
    output << "Number of SPR proposals                =  "
           << globalSPRs << endl;
    output << "NNI acceptance rate                    =  "
           << (double)globalAcceptedNNIs*100.0 / (double)globalNNIs
           << "%" << endl;
    output << "Number of NNI proposals                =  "
           << globalNNIs << endl;
    
    cout << setprecision(2);
    output << setprecision(2);
    cout << "Local NNI (induced by a branch length proposal) acceptance rate =  "
         << (double)localAcceptedNNIs*100.0 / (double)localNNIs
         << "%" << endl;
    cout << "Number of local NNI proposals                                   =  "
         << localNNIs << endl;
    output << "Local NNI (induced by a branch length proposal) acceptance rate =  " 
           << (double)localAcceptedNNIs*100.0 / (double)localNNIs
           << "%" << endl;
    output << "Number of local NNI proposals                                   =  "
           << localNNIs << endl;
    
    cout << setprecision(6);
    output << setprecision(6);
    cout << "Final Branch Step Parameter = " << branchStep << endl;
    output << "Final Branch Step Parameter = " << branchStep << endl;
    
    cout << setprecision(4);
    output << setprecision(4);
    cout << "Best Loglikelihood = " << bestLikelihood << endl;
    cout << "Best Tree" << endl;
    cout << bestStringTree << endl;
    output << "Best Tree" << endl;
    output << bestStringTree << endl;
    output << "Best Loglikelihood = " << bestLikelihood << endl;

    model->setAllParameters( bestModelParameters );
    model->printParameters( cout );
    model->printParameters( output );
    model->printPerturbationParameters( cout );
    model->printPerturbationParameters( output );

    model->getModelParameters().saveToFile( bestModelParametersFile );
    bestTreeFile << bestStringTree << terminator << endl;
    
#ifdef DEBUG
    InferenceTree checktree( bestStringTree );
    assert( checktree.toString( false ) == bestStringTree );
    checktree.loadDataAndModel( currentTree->t_table, model );
    checktree.setBranchVector( bestBranchVector );
    assert( ( int )( checktree.loglikelihood() * 100 ) == ( int )( bestLikelihood *
    100 ) );
#endif
    delete proposedTree;
    return ( 1.0 );
}



void MCMCphase::run( int argc, char * * argv ) {
  
    if ( argc != 2 ) {
        cerr << "usage : " << argv[0] << " control_file" << endl;
        exit( 1 );
    }

    // time structure to record the begining and the end of the simulation
    time_t start;
    time_t end;

    // sequenceTable object : to read the data file and store the molecular
    // data
    SequenceTable * table = NULL;

    // Create a FileParser from the control-file
    FileParser * fileParser = new FileParser( argv[1] );

    // Retrieve the control-file parameters
    ParametersSet parameters = fileParser->retrieveParametersSet();

    delete fileParser;

    // Initial tree
    InferenceTree * startingTree = NULL;

    // random seed
    time_t seed;

    double branchLengthUpperBound, initialBranchStep;
    string outgroup, terminatorChar;

    bool randStartingTree;
    string startingTreeFile;
    bool randStartingModel;
    string startingModelFile;
   
    initialBranchStep = parameters( "PERTURBATION" ).doubleParameter( "Initial branch step proposal parameter" );
    branchLengthUpperBound =
     parameters( "PERTURBATION" ).doubleParameter("Branch length upper bound");

    outgroup = parameters( "DATAFILE" ).stringParameter( "Outgroup" );

    /** ************************************************************************
     * Create global output file
     ************************************************************************ */
    string filename;
    string outputFileBaseName( parameters.stringParameter( "Output file" ) );
    // Find out if the ".output" file already exists and if it does
    // confirm overwrite
    filename = outputFileBaseName + string( ".output" );
    FileParser::confirmOpenFile( output, filename );

    // Find out if the ".samples" file already exists and if it does
    // confirm overwrite
    filename = outputFileBaseName + string( ".samples" );
    FileParser::confirmOpenFile( samplefile, filename );

    // Find out if the ".plot" file already exists and if it does
    // confirm overwrite
    filename = outputFileBaseName + string( ".plot" );
    FileParser::confirmOpenFile( plotfile, filename );
    plotfile.setf(ios::fixed);
    plotfile << setprecision(3);

    // Find out if the ".mp" file already exists and if it does
    // confirm overwrite (models are responsible for the precision and
    // the alignment of the figures in this file)
    filename = outputFileBaseName + string( ".mp" );
    FileParser::confirmOpenFile( mpfile, filename );

    // Find out if the ".bl" file already exists and if it does
    // confirm overwrite
    filename = outputFileBaseName + string( ".bl" );
    FileParser::confirmOpenFile( branchfile, filename );
    branchfile.setf(ios::fixed);
    branchfile << setprecision(4);

    // Find out if the ".besttree" file already exists and if it does
    // confirm overwrite, the tree is responsible for the precision of
    // branch lengths
    filename = outputFileBaseName + string( ".besttree" );
    FileParser::confirmOpenFile( bestTreeFile, filename );

    // Find out if the ".bestmodel" file already exists and if it does
    // confirm overwrite, models are responsible for the precision of the
    // figures
    filename = outputFileBaseName + string( ".bestmodel" );
    FileParser::confirmOpenFile( bestModelParametersFile, filename );

    /** ************************************************************************
     * End Create global output files
     ************************************************************************ */


    /** ************************************************************************
     * Data File Initialization
     ************************************************************************ */

    // Read in molecular data and remove unknown characters from the data matrix
    string dataFile = parameters( "DATAFILE" ).stringParameter( "Data file" );
    table = new SequenceTable( parameters( "DATAFILE" ) );
    //table->normalise( valid_chars, 7 );
    
    // print out molecular data to output file and standard output
    table->print( output );
    table->print( cout );
    output << endl;
    cout << endl;

    /** ************************************************************************
     * End Data File Initialization
     ************************************************************************ */


    /** ************************************************************************
     * Model Initialization
     ************************************************************************ */
    Model * startingModel = NULL;
    Singleton < ModelFactory > & modelFactory = Singleton < ModelFactory >::instance();
    startingModel = modelFactory.create( parameters( "MODEL" ).stringParameter( "Model" ),
         parameters( "MODEL" ) );
    randStartingModel = parameters.boolParameter("Random start model parameters");
    if ( randStartingModel ){
        startingModel->initialisation( table );
    }
    else{
        startingModel->initialisation( NULL );
        startingModelFile =
            parameters.stringParameter("User's starting model parameters file");
        // Create a FileParser from the model parameters-file
        FileParser * modelFileParser = new FileParser( startingModelFile );
        // Retrieve the model parameters
        ParametersSet modelParameters = modelFileParser->retrieveParametersSet();
        delete modelFileParser;
        //initialise the model with them
        startingModel->setModelParameters(modelParameters);
    }
    startingModel->initialisePerturbator( parameters( "PERTURBATION" ) );

    /** ************************************************************************
     * End Model Initialization
     ************************************************************************ */
    randStartingTree = parameters.boolParameter( "Random start tree" );
    if ( !randStartingTree ) {
        startingTreeFile = parameters.stringParameter( "User's starting tree file" );
    }

    int numberBurninIterations = parameters.intParameter( "Burnin iterations" );
    int numberSamplingIterations = parameters.intParameter( "Sampling iterations" );
    int samplingPeriod = parameters.intParameter( "Sampling period" );


    string outputFormat( parameters.stringParameter( "Output format" ) );
    for_each<string::iterator,int(*)(int)>( outputFormat.begin(), outputFormat.end(), tolower );
    if ( outputFormat == "phylip" ){
        terminatorChar = string( ";" );
    }
    else {
        terminatorChar = string( "" );
    }
                
   /** ************************************************************************
    * Print out parameters to output file
    ************************************************************************ */
    if ( parameters.find( "Random seed" ) == parameters.end() ){
        cout << "Random seed not found, CPU time used" << endl;
        output << "Random seed not found, CPU time used" << endl;
        seed = time( 0 );
    }
    else{
        seed = parameters.intParameter("Random seed");
    }
    cout << "Random seed = " << seed << endl;
    output << "Random seed = " << seed << endl;
    
    cout << "Data file          = " << dataFile << endl;
    cout << "Number of species  = " << table->getNumberSpecies() << endl;
    cout << "Outgroup           = " << outgroup << endl;
    cout << "Model              = " << startingModel->getName() << endl;
    cout << "Burnin iterations      = " << numberBurninIterations << endl;
    cout << "Sampling iterations    = " << numberSamplingIterations << endl;
    cout << "Sampling period        = " << samplingPeriod << endl;    
    
    output << "Data file          = " << dataFile << endl;
    output << "Number of species  = " << table->getNumberSpecies() << endl;
    output << "Outgroup           = " << outgroup << endl;
    output << "Model              = " << startingModel->getName() << endl;
    output << "Burnin iterations      = " << numberBurninIterations << endl;
    output << "Sampling iterations    = " << numberSamplingIterations << endl;
    output << "Sampling period        = " << samplingPeriod << endl;    
        
    cout << "Random start model parameters             = "
         << ( randStartingModel ? "yes" : "no" ) << endl;
    output << "Random start model parameters             = "
           << ( randStartingModel ? "yes" : "no" ) << endl;
    if ( !randStartingModel ) {
        cout << "User's starting model parameters file = "
             << startingModelFile << endl;
        output << "User's starting model parameters file = "
               << startingModelFile << endl;
    }
    cout << "Random start tree                         = "
         << ( randStartingTree ? "yes" : "no" ) << endl;
    output << "Random start tree                         = "
           << ( randStartingTree ? "yes" : "no" ) << endl;
    if ( !randStartingTree ) {
        cout << "User's starting tree file             = "
             << startingTreeFile << endl;
        output << "User's starting tree file             = "
               << startingTreeFile << endl;
    }
    cout << "Output files root       = " << outputFileBaseName << endl;
    output << "Output files root       = " << outputFileBaseName << endl;

    /** ************************************************************************
     * End Print Out
     ************************************************************************ */


    // Create a random start tree
    Singleton < randombox > & randBox = Singleton < randombox >::instance();
    randBox.setSeed( seed );
    if ( randStartingTree )
        startingTree = randomTree( table, branchLengthUpperBound *.2 , outgroup );
    else { // Or use the user specified start tree
        string stringTree;
        char next = ' ';
        ifstream userTreeFile( startingTreeFile.c_str() );
        userTreeFile >> next;
        while ( next != ';' ) {
            stringTree = stringTree + next;
            userTreeFile >> next;
        }
        startingTree = new InferenceTree( stringTree );
        startingTree->standardizeTree( outgroup );
    }
    // Output the initial tree and its loglikelihood
    cout << "Initial Tree" << endl;    
    cout << startingTree->toString( false ) << endl;
    output << "Initial Tree" << endl;
    output << startingTree->toString( false ) << endl;

    // Load the model and data
    startingTree->loadDataAndModel( table, startingModel );

    cout << "Lnlikelihood = " << startingTree->loglikelihood() << endl;
    startingModel->printParameters( cout );
    output << "Lnlikelihood = " << startingTree->loglikelihood() << endl;
    startingModel->printParameters( output );

    // start the timer
    start = time( 0 );    
    
    // launch the chain and wait....
    runMCMC( startingTree, startingModel, initialBranchStep, numberBurninIterations,
    numberSamplingIterations, samplingPeriod, outgroup, branchLengthUpperBound, terminatorChar );

    delete startingTree;
    delete startingModel;
    
    // end the timer
    end = time( 0 );

    cout << "Time Taken = " << (int)( end - start ) << " seconds" << endl;
    output << "Time Taken = " << (int)( end - start ) << " seconds" << endl;
    cout << "Number of random calls = " << randBox.number_of_calls() << endl;
}


InferenceTree * MCMCphase::randomTree( SequenceTable * table, double maxBranchLengths,
                                       const string & outgroup ) {

    InferenceTree * tree = new InferenceTree();
    vector < double > branchLengths;
    vector < int > array( table->getNumberSpecies() );
    vector < int > array2( table->getNumberSpecies() );
    string leaf[table->getNumberSpecies()];

    Singleton < randombox > & randBox = Singleton < randombox >::instance();

    // Read in all species names
    for ( unsigned int i = 0; i < table->getNumberSpecies(); i++ ) {
        leaf[i] = table->species[i];
        array2[i] = i;
    }
    // Create a tree by addition of species to randomly selected branches
    array = statlib::random_permutation( array2 );
    // Branch the three first species to the root node
    for ( unsigned int i = 0; i < 3; ++i ) {
        tree->addLeaf( leaf[array[i]], i,
                       maxBranchLengths * randBox.ran(), randBox.ran() );
    }
    // Branch iteratively remaining species from random edges
    for ( unsigned int i = 3; i < table->getNumberSpecies(); ++i ) {
        tree->addLeaf( leaf[array[i]],
        ( int )( floor( randBox.ran() * tree->getNumberBranches() ) ),
        maxBranchLengths * randBox.ran(), randBox.ran() );
    }
    tree->standardizeTree( outgroup );
    return tree;
}

int main( int argc, char * argv[] ) {
    MCMCphase mcmcphase;
    
/*    void* modelModule;
    modelModule = dlopen("Gtr.dll", RTLD_NOW);
    if ( ! modelModule ){
        cerr << dlerror() << endl;
	exit(1);
    }
    modelModule = dlopen("Rna7a.dll", RTLD_NOW);
    if ( ! modelModule ){
        cerr << dlerror() << endl;
	exit(1);
    }
    modelModule = dlopen("Mixed.dll", RTLD_NOW);
    if ( ! modelModule ){
        cerr << dlerror() << endl;
	exit(1);
    }
*/
    mcmcphase.run( argc, argv );    
    return 0;
}




