#include "MLphase.h"

#include "Util/FileParser.h"
#include "Util/ParametersSet.h"

#include "Models/Model.h"
#include "Util/ModelFactory.h"

#include "SequenceTable.h"

#include "Tree/TreeModelValue.h"
#include "Tree/TreesSet.h"
#include "Tree/ClustersSet.h"


#include "PatternDesign/Singleton.h"
#include "Util/randombox.h"

#include "Util/Optimise.h"
#include "Util/TopologyOptimise.h"

#include <iostream>
#include <fstream>


using namespace std;

ofstream outputFile;

/*
void MLphase::PrintMatrixBool( Array2D < bool > arr, string & label,
ostream & out ) {
    int i, j, ncolumns, nrows;
    ncolumns = arr.columns();
    nrows = arr.rows();

    out << label;
    for ( i = 0; i < ncolumns; i++ ) out << setw( 7 ) << ( i + 1 );
    out << endl;
    for ( i = 0; i < nrows; i++ ) {
        out << setw( label.length() ) << ( i + 1 ) << "  ";
        for ( j = 0; j < ncolumns; j++ ) {
            if ( arr( i, j ) ) out << "   true";
            else
                out << "  false";
        }
        out << endl;
    }
    out << endl;
}
*/

/*
void MLphase::PrintMatrixDouble( Array2D < double > arr, string & label,
int fieldwidth, int precision, ostream & out ) {
    int i, j, ncolumns, nrows;
    ncolumns = arr.columns();
    nrows = arr.rows();

    for ( i = 0; i < ( int )( fieldwidth - label.length() ); i++ ) out << " ";
    out << label << "  ";
    for ( i = 0; i < ncolumns; i++ ) out << setw( fieldwidth ) << ( i + 1 );
    out << endl;

    for ( i = 0; i < nrows; i++ ) {
        out << setw( fieldwidth ) << ( i + 1 ) << "  ";
        out.setf( ios::fixed );
        for ( j = 0; j < ncolumns; j++ )
            out << setw( fieldwidth ) << setprecision( precision ) << arr( i, j );
        out << endl;
        out.unsetf( ios::fixed );
    }
    out << endl;
    return;

}
*/

TreesSet * MLphase::ReadOptimiseTrees( ParametersSet & parameters,
                                       SequenceTable* sequenceTable,
                                       Model* model ) {
    TreesSet * treesSet = NULL;
    vector < double > branchLengths, modelParameters;
    InferenceTree * tree = NULL;
    double maxLikelihood;
    string stringTree;
    
    Singleton<randombox>& randBox = Singleton<randombox>::instance();
        
    string treesFileName = parameters.stringParameter( "Trees file" );
    ifstream treesFile( treesFileName.c_str() );
    if ( !treesFile ) {
        cerr << "I/O Error: Unable to open file \"" << treesFileName << "\"" << endl;
        exit( 1 );
    }

    stringstream treesStream;
    FileParser::skipComments( treesFile, treesStream );
    int numberTrees = parameters.intParameter( "Number of trees" );
    treesSet = new TreesSet( numberTrees );


    //initialisation of the model
    bool optimiseModel = parameters.boolParameter( "Optimise model parameters" );
    
    //if parameters are to be optimised AND initial values are not given
    //initialise from the sequence table
    if ( optimiseModel &&
         ( parameters.find("User's model parameters file") ==
           parameters.end() ) ){
        model->initialisation( sequenceTable );
    }
    //if the parameters are fixed or initial values are suggested then use them
    else{
        model->initialisation( NULL );
        // Create a FileParser from the model parameters-file
        FileParser * modelFileParser = new FileParser(
              parameters.stringParameter( "User's model parameters file") );

        // Retrieve the model parameters
        ParametersSet modelParameters = modelFileParser->retrieveParametersSet();
        // Initialise the model with them
        model->setModelParameters(modelParameters);
    }


    for ( int i = 0; i < numberTrees; ++i ) {
        cout << "Optimising Tree " << ( i + 1 ) << endl;
        stringTree.clear();
        FileParser::readTree(treesStream, stringTree);
        if ( stringTree.length() == 0 ){
            cerr << "Error while reading the tree in the file : "
                 << treesFileName << endl;
            exit(1);
        }
        
        tree = new InferenceTree( stringTree );
        if ( tree->isRooted() ) {
             tree->unroot();
        }
        cout << "Topology : " << tree->toStringTopology( false ) << endl;

        tree->loadDataAndModel( sequenceTable, model );
        tree->getBranchVector( branchLengths );
        //if the provided tree does not have valid branch lengths
        for ( unsigned int j = 0; j < branchLengths.size(); ++j ){
            if (branchLengths[j] <= 0){
                branchLengths[j] = 0.9 * randBox.ran();
            }
        }
        tree->setBranchVector( branchLengths );
        Optimise::optimiseQuasiNewton( *tree, *model, branchLengths,
        optimiseModel, 1e-6, true );
        cout << endl;
        maxLikelihood = tree->loglikelihood();
        model->getAllParameters( modelParameters );
        stringTree = tree->toString( false );
        treesSet->insert( stringTree, branchLengths,
                          modelParameters, maxLikelihood );
        //delete tree;
    }
    return ( treesSet );
}


void MLphase::PrintTrees( TreesSet* trees, string & outgroup, bool rooted,
                          Model* model ) {

    vector < double > modelParameters;
    double maxLikelihood;

    cout << endl;
    cout << "**********************************************************************" << endl;
    cout << "******************************* Trees ********************************" << endl;
    cout << "**********************************************************************" << endl;
    cout << endl;

    outputFile << endl;
    outputFile << "**********************************************************************" << endl;
    outputFile << "******************************* Trees ********************************" << endl;
    outputFile << "**********************************************************************" << endl;
    outputFile << endl;
    
    for ( unsigned int i = 0; i < trees->size(); ++i ) {
        InferenceTree tree( trees->get( i + 1 )->stringTree );
        tree.standardizeTree( outgroup );
        model->setAllParameters( trees->get( i + 1 )->modelParameters );
        maxLikelihood = trees->get( i + 1 )->logLikelihood;

        //output the tree on the screen and the output file
        cout.setf(ios::fixed);
        outputFile.setf(ios::fixed);
        cout << setprecision(4);
        outputFile << setprecision(4);
        cout << ( i + 1 ) << ". " << tree.toString( rooted ) << endl;
        outputFile << ( i + 1 ) << ". " << tree.toString( rooted ) << endl;
        //output the ML on the screen and the output file (same precision)
        cout << "ML = " << maxLikelihood << endl;
        outputFile << "ML = " << maxLikelihood << endl;
        //output the model parameters on the screen and on the output file
        model->printParameters( cout );
        cout << endl;        
        model->printParameters( outputFile );
        outputFile << endl;
    }
}


void MLphase::testKH( TreesSet * trees, bool rooted, const string & outgroup,
                      SequenceTable* table, Model* model ) {
    assert(0);
    
    
    trees = trees; 
    rooted = rooted;
    cout << outgroup;
    table = table;
    model = model;
/*
    string label = string("Tree");
    InferenceTree tree = NULL;
    
    vector < double > modelParameters1;
    vector < double > modelParameters2;
    vector < double > branchLengths;
    vector < double > maxLikelihoodValues;

    int size = trees.size();

    Array2D < double > variance( size, size );

    Array2D < bool > significant( size, size );
    
    maxLikelihoodValues.setsize( size );

    for ( int i = 0; i < size; ++i ) {
        tree = new Tree( trees.get( i + 1 )->tree );
        tree->setBranchVector( trees.get( i + 1 )->branchLengths );
        tree->loadDataAndModel( table, model );
        model->setAllParameters( trees.get( i + 1 )->modelParameters );
        maxLikelihoodValues[i] = tree->loglikelihood();
        variance( i, i ) = 0.0;
        for ( j = ( i + 1 ); j < size; j++ ) {
            if ( tree2 != NULL ) delete tree2;
            tree2 = NULL;
            tree2 = new tree( trees.get( j + 1 )->tree );
            tree2->set_branch_vector( trees.get( j + 1 )->bl );
            tree2->loadModel( * model2 );
            tree2->loadData( table );
            model2->setAllParameters( trees.get( j + 1 )->mp );
            tree2->loglikelihood();
            variance( i, j ) = variance( j, i ) = tree1->variance( * tree2 );
        }
    }

    for ( i = 0; i < size; i++ ) {
        significant( i, i ) = false;
        for ( j = ( i + 1 ); j < size; j++ )
            significant( i, j ) = significant( j, i ) =
            ( fabs( ml_values[i] - ml_values[j] ) >
            ( 1.96 * sqrt( variance( i, j ) ) ) );
    }

    PrintTrees( trees, outgroup, rooted, modelnumber, gd, ngrates, id,
    out );

    cout << endl;
    cout << "**********************************************************************" << endl;
    cout << "*** Variance of the difference in loglikelihoods between the trees ***" << endl;
    cout << "**********************************************************************"
    << endl << endl;
    if ( & cout != & out ) {
        out << endl;
        out << "**********************************************************************" << endl;
        out << "*** Variance of the difference in loglikelihoods between the trees ***" << endl;
        out << "**********************************************************************"
        << endl << endl;
    }

    PrintMatrixDouble( variance, label, 12, 4, cout );
    if ( & cout != & out ) PrintMatrixDouble( variance, label, 12, 4, out );



    cout << endl;
    cout << "**********************************************************************" << endl;
    cout << "*** Difference in Loglikelihood Significant ? ************************" << endl;
    cout << "**********************************************************************"
    << endl << endl;
    if ( & cout != & out ) {
        out << endl;
        out << "**********************************************************************" << endl;
        out << "*** Difference in Loglikelihood Significant ? ************************" << endl;
        out << "**********************************************************************"
        << endl << endl;
    }

    PrintMatrixBool( significant, label, cout );
    if ( & cout != & out ) PrintMatrixBool( significant, label, out );
    */

}



void MLphase::run( int argc, char * * argv ) {

    if ( argc != 2 ) {
        cerr << "usage : " << argv[0] << "control_file" << endl;
        exit( 1 );
    }

    // time structure to record the beginning and the end of the simulation
    time_t start;
    time_t end;

    // Create a FileParser from the control-file
    FileParser * fileParser = new FileParser( argv[1] );

    // Retrieve the control-file parameters
    ParametersSet parameters = fileParser->retrieveParametersSet();

    delete fileParser;
    
    /** ************************************************************************
     * Output File Initialization
     ************************************************************************ */
    string outputFileName = parameters.stringParameter( "Output file" );
    FileParser::confirmOpenFile( outputFile, outputFileName );

    /** ************************************************************************
     * Data file Initialization
     ************************************************************************ */
    // sequence_table object : to read the data file and store the molecular
    // data
    SequenceTable * sequenceTable = NULL;

    string outgroup = parameters( "DATAFILE" ).stringParameter( "Outgroup" );
    string dataFile = parameters( "DATAFILE" ).stringParameter( "Data file" );
    sequenceTable = new SequenceTable( parameters( "DATAFILE" ) );

    
    // print out molecular data to output file and standard output
    sequenceTable->print( outputFile );
    outputFile << endl;
    sequenceTable->print( cout );

    /** ************************************************************************
     * End Data File Initialization
     ************************************************************************ */


    /** ************************************************************************
     * Model Initialization
     ************************************************************************ */
    Model * model = NULL;
    Singleton < ModelFactory > & modelFactory = Singleton < ModelFactory >::instance();
    model = modelFactory.create( parameters( "MODEL" ).stringParameter( "Model" ),
         parameters( "MODEL" ) );
         

    /** initialise the random number generator */
    Singleton<randombox>& randBox = Singleton<randombox>::instance();
    randBox.setSeed( parameters.intParameter( "Random seed" ) );
    
    start = time( 0 );
    if ( parameters("FUNCTION").stringParameter("Function") ==
            string("Optimise user-defined phylogenetic trees") ) {
        TreesSet* treesSet = ReadOptimiseTrees( parameters("FUNCTION"),
                                                sequenceTable, model );
        PrintTrees( treesSet, outgroup, false, model );
        end = time( 0 );
        cout << "Time Taken = " << ( end - start ) << endl;
        outputFile << "Time Taken = " << ( end - start ) << endl;
        outputFile.close();
        delete treesSet;
    }

    if ( parameters("FUNCTION").stringParameter("Function") ==
            string("Search for ML topology") ) {
        TreesSet* treesSet = new TreesSet( 10 );
        bool userClades = parameters("FUNCTION").
                           boolParameter("User defined monophyletic clades");
        string searchMethod =
                parameters("FUNCTION").stringParameter("Topology search");
        int searchMethodId = -1;
        if ( searchMethod == "Simple exhaustive search" ) searchMethodId = 0;
        if ( searchMethod == "Branch-and-bound exhaustive search" ) searchMethodId = 1;
        if ( searchMethod == "Heuristic stepwise addition" ) searchMethodId = 2;
        ClustersSet* clusters = NULL;
        if ( userClades ){
            string clustersFile =
                parameters("FUNCTION").stringParameter("Clade file");
            ifstream clustersStream(clustersFile.c_str());
            clusters = new ClustersSet(clustersStream);
        }
        //initialisation of the model
        bool optimiseModel = parameters("FUNCTION").boolParameter( "Optimise model parameters" );
        //if parameters are to be optimised AND initial values are not given
        //initialise from the sequence table
        if ( optimiseModel &&
             ( parameters("FUNCTION").find("User's model parameters file") ==
               parameters("FUNCTION").end() ) ){
            model->initialisation( sequenceTable );
        }
        //if the parameters are fixed or initial values are suggested
        //then use them
        else{
            model->initialisation( NULL );
            // Create a FileParser from the model parameters-file
            FileParser * modelFileParser = new FileParser(
              parameters("FUNCTION").stringParameter(
                                      "User's model parameters file") );
            // Retrieve the model parameters
            ParametersSet modelParameters = modelFileParser->retrieveParametersSet();
            // Initialise the model with them
            model->setModelParameters(modelParameters);
        }

        switch( searchMethodId ){
            case 0 :
                cout << "Simple exhaustive search";
                if ( userClades ){
                    cout << " with fixed monophyletic clades" << endl;
                    TopologyOptimise::exhaustiveSearch( sequenceTable, model,
                                                        optimiseModel, treesSet,
                                                        outputFile, *clusters );
            }
            else{
                cout << endl;
                TopologyOptimise::exhaustiveSearch( sequenceTable, model,
                                                    optimiseModel, treesSet,
                                                    outputFile );
            }
            break;
            case 1 :
                cout << "Branch-and-bound exhaustive search";
                if ( userClades ){
                    cout << " with fixed monophyletic clades" << endl;
                    TopologyOptimise::branchAndBoundSearch( sequenceTable,
                                                          model, optimiseModel,
                                                          treesSet, outputFile,
                                                          *clusters );
                }
                else{
                    cout << endl;
                    TopologyOptimise::branchAndBoundSearch( sequenceTable,
                                                        model, optimiseModel,
                                                        treesSet, outputFile );
                }
            break;
            case 2 :
                cout << "Heuristic stepwise addition";
                if ( userClades ){
                    cout << " with fixed monophyletic clades" << endl;
                    TopologyOptimise::buildTreeStepwise( sequenceTable, model,
                                                        optimiseModel, treesSet,
                                                        outputFile, *clusters );
                }
                else{
                    cout << endl;
                    TopologyOptimise::buildTreeStepwise( sequenceTable, model,
                                                        optimiseModel, treesSet,
                                                        outputFile );
                }
            break;
            default :
                cerr << "The search method : \"" << searchMethod
                     << "\" doesn't exist, exitting..." << endl;
                exit(1);
        } // end of switch statement for methodId
        delete clusters;
        //print the trees in the set
        PrintTrees( treesSet, outgroup, false, model );
        end = time( 0 );
        cout << "Time Taken = " << ( end - start ) << endl;
        outputFile << "Time Taken = " << ( end - start ) << endl;
        outputFile.close();
        delete treesSet;
    }
}


int main( int argc, char * argv[] ) {
    MLphase mlphase;
    mlphase.run( argc, argv );
    return 0;
}




