#include "Models/DnaModel.h"

#include <iostream>
#include <iomanip>
#include <stdio.h>

#include "Sequence/SequenceTable.h"

#include "Models/Perturbator.h"
#include "Util/ModelFactory.h"

#include "PatternDesign/Singleton.h"
#include "Util/statlib.h"
#include "Util/array2D.h"
#include "Util/randombox.h"
#include "Util/ParametersSet.h"


DnaModel::DnaModel( const string & registrationName ) :
        MatrixModel( registrationName ){
}  

DnaModel::DnaModel( ParametersSet & parameters ) : MatrixModel( parameters ){
    if ( discreteGamma ){
        eigenMatrix = new array2D < double > [numberGammaCategories] ( 4, 4 );
        ieigenMatrix = new array2D < double > [numberGammaCategories] ( 4, 4 );
        eigenValues = new vector < double > [numberGammaCategories] ( 4 );
        rateMatrix = new array2D < double > [numberGammaCategories] ( 4, 4 );
    }
    else{
        eigenMatrix = new array2D < double >( 4, 4 );
        ieigenMatrix = new array2D < double >( 4, 4 );
        eigenValues = new vector < double >( 4 );
        rateMatrix = new array2D < double >( 4, 4 );
    }
}

DnaModel::~DnaModel() {
    if ( discreteGamma ){
        if ( eigenMatrix != NULL ){
            delete[] eigenMatrix;
            eigenMatrix = NULL;
        }
        if ( ieigenMatrix != NULL ){
            delete[] ieigenMatrix;
            ieigenMatrix = NULL;
        }
        if ( eigenValues != NULL ){
            delete[] eigenValues;
            eigenValues = NULL;
        }
        if ( rateMatrix != NULL ){
            delete[] rateMatrix;
            rateMatrix = NULL;
        }
    }
    else{
        if ( eigenMatrix != NULL ){
            delete eigenMatrix;
            eigenMatrix = NULL;
        }
        if ( ieigenMatrix != NULL ){
            delete ieigenMatrix;
            ieigenMatrix = NULL;
        }
        if ( eigenValues != NULL ){
            delete eigenValues;
            eigenValues = NULL;
        }
        if ( rateMatrix != NULL ){
            delete rateMatrix;
            rateMatrix = NULL;
        }
    }
}

string DnaModel::getState( int stateNumber, int symbolCategory ) const {
    assert( symbolCategory == 0 );
    switch ( stateNumber ) {
        // Watson-Crick base pairs
        case 0 :
            return ( string( "A" ) );
        case 1 :
            return ( string( "C" ) );
        case 2 :
            return ( string( "G" ) );
        case 3 :
            return ( string( "T" ) );
        default :
            return("");
    }
}
          
int DnaModel::getSymbolNumber( const string & base, int symbolCategory ) const {
    assert(symbolCategory==0);
    if (base.length() != 1)
        return 8;
    switch ( base[0] ) {
        case 'a' :
        case 'A' :
            return 0;
        case 'c' :
        case 'C' :
            return 1;
        case 'g' :
        case 'G' :
            return 2;
        case 'u' :
        case 't' :
        case 'U' :
        case 'T' :
            return 3;
        case 'r' :
        case 'R' :
            return 4; // Unknown purine       (A or G)
        case 'y' :
        case 'Y' :
            return 5; // Unknown pyramidine   (C or U)
        case 'n' :
        case 'N' :
        case 'x' :
        case 'X' :
        case '?' :
            return 6; // Unknown nucleotide   (A,C,G or T)
        case '-' :
            return 7; // gaps are treated as unknown nucleotides
        default :
            return 8; // Out of range (unrecognized symbol)
    }
}



vector < double > DnaModel::retrieveEmpiricalRates( SequenceTable * sequenceTable,
int modelId, double * propInvariant ) {

    double reference;
    vector < double > rate5( 5 );
    
    const array2D< string > & sequences = sequenceTable->getSequences( modelId );
    const vector< pair<string, unsigned int> > & invariantBases = sequenceTable->getInvariantBases( modelId );
    int s1, s2;

    reference = 0.0; // A->G
    rate5[0] = 0.0; // A->C
    rate5[1] = 0.0; // A->U
    rate5[2] = 0.0; // C->G
    rate5[3] = 0.0; // C->U
    rate5[4] = 0.0; // G->U
        
    for ( unsigned int seq = 0;
          seq < sequenceTable->getNumberSpecies(); ++seq ) {
        for ( unsigned int seq2 = seq + 1;
              seq2 < sequenceTable->getNumberSpecies(); ++seq2 ) {
            for ( unsigned int site = 0;
                  site < sequences.numberColumns();
                  ++site ) {
                s1 = getSymbolNumber( sequences( seq, site ) );
                s2 = getSymbolNumber( sequences( seq2, site ) );
                if ( (s1 == -1) || (s2 == -1) ){
                     cerr << "Unrecognized symbol \"";
                     if (s1 == -1){
                         cerr << sequences( seq, site );
                     }
                     else{
                         cerr << sequences( seq2, site );
                     }
                     cerr << "\" during initialisation of model " << modelId << endl;
                     exit(1);
                }       
                if ( ( s1 < 4 ) && ( s2 < 4 ) && ( s1 != s2 ) ) {
                    if ( ( ( s1 == 0 ) && ( s2 == 1 ) ) ||
                    ( ( s1 == 1 ) && ( s2 == 0 ) ) )
                        ++(rate5[0]);
                    if ( ( ( s1 == 0 ) && ( s2 == 2 ) ) ||
                    ( ( s1 == 2 ) && ( s2 == 0 ) ) )
                        ++reference;
                    if ( ( ( s1 == 0 ) && ( s2 == 3 ) ) ||
                    ( ( s1 == 3 ) && ( s2 == 0 ) ) )
                        ++(rate5[1]);
                    if ( ( ( s1 == 1 ) && ( s2 == 2 ) ) ||
                    ( ( s1 == 2 ) && ( s2 == 1 ) ) )
                        ++(rate5[2]);
                    if ( ( ( s1 == 1 ) && ( s2 == 3 ) ) ||
                    ( ( s1 == 3 ) && ( s2 == 1 ) ) )
                        ++(rate5[3]);
                    if ( ( ( s1 == 2 ) && ( s2 == 3 ) ) ||
                    ( ( s1 == 3 ) && ( s2 == 2 ) ) )
                        ++(rate5[4]);
                }
            }
        }
    }
    if ( reference == 0.0 ) {
        cerr << "Error while retrieving empirical rates. "
        << "Not enough single site nucleotid ?" << endl;
        exit(1);
    }
    
    // The rate to A->G is the reference rate
    rate5[0] /= reference; // A->C
    rate5[1] /= reference; // A->U
    rate5[2] /= reference; // C->G
    rate5[3] /= reference; // C->U
    rate5[4] /= reference; // G->U
    if ( propInvariant != NULL ) {
        int numberInvariant = 0;
        for ( vector< pair< string, unsigned int > >::const_iterator iter = invariantBases.begin();
              iter != invariantBases.end(); ++iter ){
            numberInvariant += (*iter).second;
        }
        * propInvariant = (double)numberInvariant/
                          (double)(numberInvariant + sequences.numberColumns());
    }
    return rate5;
}

vector < double > DnaModel::retrieveEmpiricalFrequencies
( SequenceTable * sequenceTable, int modelId ) {

    vector < double > freq4( 4 );

    for ( int i = 0; i < 4; ++i ) {
        freq4[i] = 0.0;
    }

    const array2D< string > & sequences = sequenceTable->getSequences( modelId );
    const vector< pair< string, unsigned int > > & invariantBases = sequenceTable->getInvariantBases( modelId );
    int s;
    
    // count the number of each nucleotid in the SequenceTable, purine,
    // pyrimidine and gaps are ignored.
    for ( unsigned int sequence = 0; sequence < sequenceTable->getNumberSpecies(); ++sequence ){
        for ( unsigned int site = 0; site < sequenceTable->getSequencesLength(modelId); ++site ) {
            s = getSymbolNumber( sequences( sequence, site ) );
            if (s == -1){
                 cerr << "Unrecognized symbol \"" << sequences( sequence, site )
                      << "\" during initialisation of model " << modelId << endl;
                 exit(1);
            }
            for ( int nucleotide = 0; nucleotide < 4; ++nucleotide ) {
                if ( s == nucleotide ) {
                    freq4[nucleotide] += 1.0;
                }
            }
        }
    }
    for ( vector< pair< string, unsigned int> >::const_iterator iter = invariantBases.begin();
          iter != invariantBases.end(); ++iter ){
        s = getSymbolNumber( (*iter).first );
        for ( int nucleotide = 0; nucleotide < 4; ++nucleotide ) {
            if ( s == nucleotide ) {
                freq4[nucleotide] += 1.0;
            }
        }        
    }
    
    double sum = freq4[0] + freq4[1] + freq4[2] + freq4[3];
    for ( int nucleotide = 0; nucleotide < 4; ++nucleotide ) {
        freq4[nucleotide] /= sum;
    }
    return freq4;
}

string DnaModel::getSymbol( int basenumber, int symbolCategory ) const {
    assert(symbolCategory==0);
    switch ( basenumber ) {
        case 0 :
            return ( string( "A" ) );
        case 1 :
            return ( string( "C" ) );
        case 2 :
            return ( string( "G" ) );
        case 3 :
            return ( string( "T" ) );
        case 4 :
            return ( string( "R" ) ); // Unknown Purine     (A or G)
        case 5 :
            return ( string( "Y" ) ); // Unknown Pyramidine (C or T)
        case 6 :
            return ( string( "N" ) ); // Unknown Nucleotide (A , C , G or T)
        case 7 :
            return ( string( "-" ) ); // Gaps
        default :
            return ( string( "Z" ) ); // Unrecognized symbol
    }
}


double DnaModel::probability( int oldState, int newState, double time,
int category, int symbolCategory ) const {
    assert(symbolCategory==0);
    int k, cat;
    double probability = 0.0;
    if ( ( invariantCategory == 1 ) && ( category == 0 ) ) {
        if ( oldState == newState )
            return 1.0;
        else
            return 0.0;
    }
    cat = category - invariantCategory;
    for ( k = 0; k < 4; k++ ) {
        probability += ( ieigenMatrix[cat] ( oldState, k ) *
        eigenMatrix[cat] ( k, newState ) *
        exp( eigenValues[cat] [k] * time ) );
    }
    //just to be sure
    //assert ( ( probability >= -0.000001 ) && ( probability <= 1.000001 ) );
    probability = MIN( probability, 1.0 );
    probability = MAX( probability, 0.0 );
    return ( probability );
}


double DnaModel::diffProbability( int oldState, int newState, double time,
int category, int symbolCategory ) const {
    assert(symbolCategory==0);
    int k, cat;
    double diffProbability = 0.0;
    if ( ( invariantCategory == 1 ) && ( category == 0 ) )
        return 0.0;
    cat = category - invariantCategory;
    for ( k = 0; k < 4; k++ ) {
        diffProbability += ( eigenValues[cat] [k] *
        ieigenMatrix[cat] ( oldState, k ) *
        eigenMatrix[cat] ( k, newState ) *
        exp( eigenValues[cat] [k] * time ) );
    }
    return ( diffProbability );
}


double DnaModel::secondDiffProbability( int oldState, int newState, double time,
int category, int symbolCategory ) const {
    assert(symbolCategory==0);
    int k, cat;
    double diff2Probability = 0.0;
    if ( ( invariantCategory == 1 ) && ( category == 0 ) )
        return 0.0;
    cat = category - invariantCategory;
    for ( k = 0; k < 4; k++ ) {
        diff2Probability +=
        ( pow( eigenValues[cat] [k], 2.0 ) *
        ieigenMatrix[cat] ( oldState, k ) *
        eigenMatrix[cat] ( k, newState ) *
        exp( eigenValues[cat] [k] * time ) );
    }
    return ( diff2Probability );
}


void DnaModel::initEquivalencyTable(){
    int i, j;
    equivalencyTable.resize( 8, 4 );
    // Set the patterns for valid states
    for ( i = 0; i < 4; i++ ) {
        for ( j = 0; j < 4; j++ ){
            equivalencyTable( i, j ) = 0.0;
        }
        equivalencyTable( i, i ) = 1.0;
    }
    // Match a Purine
    equivalencyTable( 4, 0 ) = 1.0; // A
    equivalencyTable( 4, 1 ) = 0.0;
    equivalencyTable( 4, 2 ) = 1.0; // G
    equivalencyTable( 4, 3 ) = 0.0;
    // Match a Pyramidine
    equivalencyTable( 5, 0 ) = 0.0;
    equivalencyTable( 5, 1 ) = 1.0; // C
    equivalencyTable( 5, 2 ) = 0.0;
    equivalencyTable( 5, 3 ) = 1.0; // U
    // Match all nucleotides (unknown)
    equivalencyTable( 6, 0 ) = 1.0; // A
    equivalencyTable( 6, 1 ) = 1.0; // C
    equivalencyTable( 6, 2 ) = 1.0; // G
    equivalencyTable( 6, 3 ) = 1.0; // U
    // Gap (No information)
    equivalencyTable( 7, 0 ) = 1.0;
    equivalencyTable( 7, 1 ) = 1.0;
    equivalencyTable( 7, 2 ) = 1.0;
    equivalencyTable( 7, 3 ) = 1.0;
}


void DnaModel::setAverageMutationRate( double newAverageRate, int symbolCategory ){
    assert( symbolCategory == 0 );
    assert( newAverageRate > 0.0 );
    averageMutationRate = newAverageRate;
    double sum = 0.0;
    double mr;
    int i, j;

    mr = averageMutationRate / ( 1.0 - proportionInvariantSites );

    if ( discreteGamma ) {
        double beta = alpha / mr;
        statlib::gamma_percentile_means( mutationRate, alpha, beta,
        numberGammaCategories );
        for ( int rateCategory = 0 ; rateCategory < numberGammaCategories;
              ++rateCategory ) {
            int freqCat = frequencyCategory[rateCategory + invariantCategory];
            int matrixCat = matrixCategory[rateCategory];
            sum = 0.0;
            for ( i = 0; i < getNumberStates() -1; ++i ) {
                for ( j = i + 1; j < getNumberStates(); ++j ) {
                    int index = matrixIndex(i, j);
                    switch ( index ) {
                        //if reference
                        case -2 : sum += ( 2.0 * frequencies[freqCat] [i] *
                                      frequencies[freqCat] [j] );
                                  break;
                        //if null
                        case -1 : break;
                        //otherwise (>=0)
                        default : sum += ( 2.0 * frequencies[freqCat] [i] *
                                      frequencies[freqCat] [j] *
                                      rateRatios[matrixCat] [index] );
                    }
                }
            }
            mutationRate[rateCategory] /= sum;
        }
    }
    else {
        sum = 0.0;
        for ( i = 0; i < getNumberStates() - 1; ++i ) {
            for ( j = i + 1; j < getNumberStates(); ++j ) {
                int index = matrixIndex(i, j);
                switch ( index ) {
                    //if reference
                    case -2 : sum += ( 2.0 * frequencies[0 + invariantCategory] [i] *
                                       frequencies[0 + invariantCategory] [j] );
                              break;
                    //if null
                    case -1 : break;
                    //otherwise (>=0)
                    default : sum += ( 2.0 * frequencies[0 + invariantCategory] [i] *
                                       frequencies[0 + invariantCategory] [j] *
                                       rateRatios[0] [index] );
                }
            }
        }
        mutationRate[0] = mr / sum;
    }
    setEigenMatrix();
}


// print the parameters in readable format
void DnaModel::printParameters( ostream & outputStream ) const {
  
    //print common parameters
    MatrixModel::printParameters( outputStream );

    outputStream.setf(ios::fixed);
    outputStream << setprecision(5);
    for ( int freqCat = 0; freqCat < numberFrequenciesCategories; ++freqCat ) {
        if (numberFrequenciesCategories>=2){
            outputStream << "set of frequencies " << freqCat + 1 << ":" << endl;
        }
        else{
            outputStream << "frequencies:" << endl;
        }
        for ( int i = 0; i < getNumberStates(); ++i ) {
            outputStream << "f[" << getState( i ) << "] = " <<
            frequencies[freqCat] [i] << "  ";
        }
        outputStream << endl;
    }
    outputStream << endl;
    
    for ( int rateMat = 0 ; rateMat < numberRatesMatrices; ++rateMat ) {
        if (numberRatesMatrices>=2){
            outputStream << "set of rate ratios " << rateMat + 1 << ":" << endl;
        }
        else{
            outputStream << "rate ratios:" << endl;
        }
        outputStream << "    ";
        for ( int i = 0; i < getNumberStates(); ++i ) {
            outputStream << "      " << getState( i ) << "     ";
        }
        outputStream << endl;
        for ( int i = 0; i < getNumberStates(); ++i ) {
            outputStream << "  " << getState( i ) << "  ";
            for ( int j = 0; j < getNumberStates(); ++j ) {
                int index = matrixIndex( i, j);
                switch ( index ) {
                    //if diagonal
                    case -3 : outputStream << "  --------  ";
                              break;
                    //if reference
                    case -2 : outputStream << "   1.00000  ";
                              break;
                    //if null (not possible with implemented DNA models
                    case -1 : assert(0);
                              break;
                    //otherwise
                    default : outputStream << "  " << setw( 8 ) <<
                                  rateRatios [rateMat] [index] << "  ";
                              break;
                }
            }
            outputStream << endl; //end line
        }
        outputStream << endl; //end rate ratios matrix
    }
    
    if ( (numberRatesMatrices>=2) || (numberFrequenciesCategories>=2) ){
        for ( int gammaCat = 0; gammaCat < MAX(numberGammaCategories,1); ++ gammaCat ) {
            if ( numberGammaCategories>1 ){
                outputStream << "transition rate matrix for the gamma category "
	                         << gammaCat + 1;
                if (numberFrequenciesCategories>=2){
                    outputStream << ", frequencies set " <<
                         frequencyCategory[gammaCat + invariantCategory] + 1;
                }
                if (numberRatesMatrices>=2){
                    outputStream << ", rate ratios set "
                                 << matrixCategory[gammaCat] + 1;
                }
                outputStream << ":" << endl;
            }
            else{
                outputStream << "transition rate matrix";	
                if (numberFrequenciesCategories>=2){
                    outputStream << ", frequencies set " <<
                         frequencyCategory[invariantCategory] + 1;
                }
                outputStream << ":" << endl;
            }
            outputStream << "    ";
            for ( int i = 0; i < getNumberStates(); ++i ) {
                outputStream << "      " << getState( i ) << "     ";
            }
            outputStream << endl;
            for ( int i = 0; i < getNumberStates(); ++i ) {
                outputStream << ' ' << getState( i ) << ' ';
                for ( int j = 0; j < getNumberStates(); ++j ) {
                    outputStream << "  " << setw( 9 )
                             << rateMatrix[gammaCat] (i, j) << "  ";
                }
                outputStream << endl; //end line
            }
            outputStream << endl; //end transition rate matrix
        }
        outputStream << endl;
    }
    else{
        outputStream << "transition rate matrix (scaled so that "
                     << "the average substitution rate is 1.0):" << endl;
        double sum = 0.0;
        for ( int i = 0; i < getNumberStates() - 1; ++i ) {
            for ( int j = i + 1; j < getNumberStates(); ++j ) {
                int index = matrixIndex(i, j);
                switch ( index ) {
                    //if reference
                    case -2 : sum += ( 2.0 * frequencies[0] [i] *
                                       frequencies[0] [j] );
                              break;
                    //if null or diagonal
                    case -1 : assert(0);
                              break;
                    case -3 : break;
                    //otherwise (>=0)
                    default : sum += ( 2.0 * frequencies[0] [i] *
                                       frequencies[0] [j] *
                                       rateRatios[0] [index] );
                }
            }
        }
        outputStream << "    ";
        for ( int i = 0; i < getNumberStates(); ++i ) {
            outputStream << "      " << getState( i ) << "     ";
        }
        outputStream << endl;
        for ( int i = 0; i < getNumberStates(); ++i ) {
            outputStream << ' ' << getState( i ) << ' ';
            for ( int j = 0; j < getNumberStates(); ++j ) {
                int index = matrixIndex(i, j);
                switch (index){
                    //if diagonal
                    case -3 : outputStream << "  ---------  ";
                              break;
                    //if reference
                    case -2 : outputStream << "  " << setw( 9 )
                                           << frequencies[0] [j] / sum << "  ";
                              break;
                    //if null
                    case -1 : assert(0);
                    //otherwise (>=0)
                    default : outputStream << "  " << setw( 9 )
                                           << rateRatios[0] [index] * frequencies[0] [j] / sum << "  ";

                }
            }
            outputStream << endl; //end line
        }
        outputStream << endl; //end transition rate matrix
    }
    outputStream << endl;
    outputStream.unsetf(ios::fixed);
}
