#include <assert.h>

#include <iostream>
#include <iomanip>
#include <stdio.h>

#include "PatternDesign/Singleton.h"

#include "Util/ModelFactory.h"
#include "Models/Perturbator.h"
#include "Models/HKY85.h"

#include "Sequence/SequenceTable.h"

#include "Util/statlib.h"
#include "Util/randombox.h"
#include "Util/ParametersSet.h"


#define SQR(n) (n*n)

using namespace std;

//register the model to the model factory with the name HKY85
HKY85 HKY85::prototype( "HKY85" );

HKY85::HKY85( const string & registrationName ) : DnaModel( registrationName ){
//the private constructor is called by the prototype only
}

HKY85::HKY85( ParametersSet & parameters ) : DnaModel( parameters ){
//the normal constructor is called by the prototype's clone method
}

void HKY85::initialisation( SequenceTable * sequenceTable, int modelId ) {
    //initialise the number of parameters
    numberRatesRatiosPerMatrix = 1;
    numberFrequencies = 4;
    //initialise the space for the parameters
    rateRatios = new vector < double > [numberRatesMatrices] (1);
    frequencies = new vector < double > [numberFrequenciesCategories] (4);
    
    //initialise the matrixIndex for correspondance between a matrix element
    //and an index in the rate ratios vector
    initMatrixIndex();
    
    //call the basic initialisation method
    DnaModel::initialisation( sequenceTable, modelId );
    
    //if a sequence table is provided initialise rate ratios and frequencies
    //accordingly
    if (sequenceTable){
        //estimation of the model frequencies from the empirical frequencies
        frequencies[0] = retrieveEmpiricalFrequencies( sequenceTable, modelId );
        for ( int i = 1 ; i < numberFrequenciesCategories; i++ ) {
            frequencies[i] = frequencies[0];
        }

        //estimation of all the rate ratios (and proportion of invariant sites
        //if relevant) from the empirical sequences
        vector < double > rate5;
        if ( invariantCategory ) {
            rate5 = retrieveEmpiricalRates( sequenceTable, modelId,
            & proportionInvariantSites );
        }
        else {
            rate5 = retrieveEmpiricalRates( sequenceTable, modelId );
        }
        //the transversion rate is the only free parameter
        rateRatios[0][0] = (rate5[0]+rate5[1]+rate5[2]+rate5[4])/(2*(1.0+rate5[3]));
        //if the discrete gamma model uses more than one set of rate ratios
        //initialise the other ones to the same values
        for ( int i = 1 ; i < numberRatesMatrices; i++ ) {
            rateRatios[i] = rateRatios[0];
        }
    }
    //init the equivalency table (correspondance symbol/states)
    initEquivalencyTable();
    //first initialisation of the mutation rate matrix
    //(factor(s) used with the subtitution matrix to reach the specified
    //average substitution rate) and eigen system
    setAverageMutationRate( averageMutationRate );
}
  
HKY85::~HKY85() {
}

string HKY85::getName( void ) const {
    char num[3];
    string name( "HKY85" );
    if ( discreteGamma ) {
        sprintf( num, "%i", numberGammaCategories );
        name = name + string( " + dG" ) + string( num );
    }
    if ( invariantCategory == 1 )
        name = name + string( " + I" );
    return ( name );
}

void HKY85::initMatrixIndex() {
    matrixIndex.resize( 4, 4 );
    for ( int i = 0 ; i < 4 ; ++i ) {
        matrixIndex(i, i) = -3;
    }
    //reference
    matrixIndex(0, 2) = matrixIndex(2, 0) = -2;
    matrixIndex(1, 3) = matrixIndex(3, 1) = -2;
    //transversion
    matrixIndex(0, 1) = matrixIndex(1, 0) = 0;
    matrixIndex(0, 3) = matrixIndex(3, 0) = 0;
    matrixIndex(1, 2) = matrixIndex(2, 1) = 0;
    matrixIndex(2, 3) = matrixIndex(3, 2) = 0;
}

void HKY85::setEigenMatrix() {
    for ( int category = 0; category < MAX( numberGammaCategories, 1 );
    category++ ) {
        // fill the rateMatrix (cf HKY85.h)
        int freqCat = frequencyCategory[category+invariantCategory];
        int rateCat = matrixCategory[category];
        rateMatrix[category] ( 0, 1 ) =
        mutationRate[category] * frequencies[freqCat] [1] *
        rateRatios[rateCat] [0]; // A->C
        rateMatrix[category] ( 0, 2 ) =
        mutationRate[category] * frequencies[freqCat] [2];
        rateMatrix[category] ( 0, 3 ) =
        mutationRate[category] * frequencies[freqCat] [3] *
        rateRatios[rateCat] [0]; // A->U
        rateMatrix[category] ( 0, 0 ) =
        -( ( rateMatrix[category] ) ( 0, 1 ) +
        ( rateMatrix[category] ) ( 0, 2 ) +
        ( rateMatrix[category] ) ( 0, 3 ) );
        rateMatrix[category] ( 1, 0 ) =
        mutationRate[category] * frequencies[freqCat] [0] *
        rateRatios[rateCat] [0]; // C->A
        rateMatrix[category] ( 1, 2 ) =
        mutationRate[category] * frequencies[freqCat] [2] *
        rateRatios[rateCat] [0]; // C->G
        rateMatrix[category] ( 1, 3 ) =
        mutationRate[category] * frequencies[freqCat] [3]; // C->U
        rateMatrix[category] ( 1, 1 ) =
        -( ( rateMatrix[category] ) ( 1, 0 ) +
        ( rateMatrix[category] ) ( 1, 2 ) +
        ( rateMatrix[category] ) ( 1, 3 ) );
        rateMatrix[category] ( 2, 0 ) =
        mutationRate[category] * frequencies[freqCat] [0]; // G->A
        rateMatrix[category] ( 2, 1 ) =
        mutationRate[category] * frequencies[freqCat] [1] *
        rateRatios[rateCat] [0]; // G->C
        rateMatrix[category] ( 2, 3 ) =
        mutationRate[category] * frequencies[freqCat] [3] *
        rateRatios[rateCat] [0]; // G->U
        rateMatrix[category] ( 2, 2 ) =
        -( ( rateMatrix[category] ) ( 2, 0 ) +
        ( rateMatrix[category] ) ( 2, 1 ) +
        ( rateMatrix[category] ) ( 2, 3 ) );
        rateMatrix[category] ( 3, 0 ) =
        mutationRate[category] * frequencies[freqCat] [0] *
        rateRatios[rateCat] [0]; // U->A
        rateMatrix[category] ( 3, 1 ) =
        mutationRate[category] * frequencies[freqCat] [1]; // U->C
        rateMatrix[category] ( 3, 2 ) =
        mutationRate[category] * frequencies[freqCat] [2] *
        rateRatios[rateCat] [0]; // U->G
        rateMatrix[category] ( 3, 3 ) =
        -( ( rateMatrix[category] ) ( 3, 0 ) +
        ( rateMatrix[category] ) ( 3, 1 ) +
        ( rateMatrix[category] ) ( 3, 2 ) );

        for ( int i = 0; i < 4; ++i ) {
            eigenValues[category] [i] = 0.0;
            for ( int j = 0; j < 4; ++j ) {
                eigenMatrix[category] ( i, j ) = 0.0;
                ieigenMatrix[category] ( i, j ) = 0.0;
            }
        }
        
        double transitionRate = mutationRate[category];
        double transversionRate = mutationRate[category] * rateRatios[rateCat][0];
        
        double purineFreq = frequencies[freqCat][0] + frequencies[freqCat][2];
        double pyrimidineFreq = frequencies[freqCat][1] + frequencies[freqCat][3];

        eigenValues[category] [0] = 0.0;
        eigenValues[category] [1] =
        -transitionRate * pyrimidineFreq - transversionRate * purineFreq;
        eigenValues[category] [2] =
        -transitionRate * purineFreq - transversionRate * pyrimidineFreq;
        eigenValues[category] [3] = -transversionRate;


        // Update the eigenmatrix by hand
        eigenMatrix[category] ( 0, 0 ) = frequencies[freqCat][0];
        eigenMatrix[category] ( 0, 1 ) = frequencies[freqCat][1];
        eigenMatrix[category] ( 0, 2 ) = frequencies[freqCat][2];
        eigenMatrix[category] ( 0, 3 ) = frequencies[freqCat][3];

        eigenMatrix[category] ( 1, 0 ) = 0.0;
        eigenMatrix[category] ( 1, 1 ) = -1.0;
        eigenMatrix[category] ( 1, 2 ) = 0.0;
        eigenMatrix[category] ( 1, 3 ) = 1.0;

        eigenMatrix[category] ( 2, 0 ) = -1.0;
        eigenMatrix[category] ( 2, 1 ) = 0.0;
        eigenMatrix[category] ( 2, 2 ) = 1.0;
        eigenMatrix[category] ( 2, 3 ) = 0.0;

        eigenMatrix[category] ( 3, 0 ) =
        -( frequencies[freqCat][0] * pyrimidineFreq ) / purineFreq;
        eigenMatrix[category] ( 3, 1 ) = frequencies[freqCat][1];
        eigenMatrix[category] ( 3, 2 ) =
        -( frequencies[freqCat][2] * pyrimidineFreq ) / purineFreq;
        eigenMatrix[category] ( 3, 3 ) = frequencies[freqCat][3];

        // Inverse of the eigenmatrix
        ieigenMatrix[category] ( 0, 0 ) = 1.0;
        ieigenMatrix[category] ( 0, 1 ) = 0.0;
        ieigenMatrix[category] ( 0, 2 ) = -frequencies[freqCat][2] / purineFreq;
        ieigenMatrix[category] ( 0, 3 ) = -1.0;

        ieigenMatrix[category] ( 1, 0 ) = 1.0;
        ieigenMatrix[category] ( 1, 1 ) = -frequencies[freqCat][3] / pyrimidineFreq;
        ieigenMatrix[category] ( 1, 2 ) = 0.0;
        ieigenMatrix[category] ( 1, 3 ) = purineFreq / pyrimidineFreq;

        ieigenMatrix[category] ( 2, 0 ) = 1.0;
        ieigenMatrix[category] ( 2, 1 ) = 0.0;
        ieigenMatrix[category] ( 2, 2 ) = frequencies[freqCat][0] / purineFreq;
        ieigenMatrix[category] ( 2, 3 ) = -1.0;

        ieigenMatrix[category] ( 3, 0 ) = 1.0;
        ieigenMatrix[category] ( 3, 1 ) = frequencies[freqCat][1] / pyrimidineFreq;
        ieigenMatrix[category] ( 3, 2 ) = 0.0;
        ieigenMatrix[category] ( 3, 3 ) = purineFreq / pyrimidineFreq;
    } //for each category (except invariant)
}

