#include "Models/K80.h"

#include <assert.h>

#include <iostream>
#include <iomanip>
#include <stdio.h>

#include "PatternDesign/Singleton.h"

#include "Util/ModelFactory.h"
#include "Models/Perturbator.h"

#include "Sequence/SequenceTable.h"

#include "Util/statlib.h"
#include "Util/randombox.h"
#include "Util/ParametersSet.h"


#define SQR(n) (n*n)

using namespace std;

//register the model to the model factory with the name K80
K80 K80::prototype( "K80" );

K80::K80( const string & registrationName ) : DnaModel( registrationName ){
//the private constructor is called by the prototype only
}

K80::K80( ParametersSet & parameters ) : DnaModel( parameters ){
//the normal constructor is called by the prototype's clone method
}

void K80::initialisation( SequenceTable * sequenceTable, int modelId ) {
    //initialise the number of parameters
    numberRatesRatiosPerMatrix = 1;
    numberFrequencies = 0;
    //initialise the space for the parameters
    rateRatios = new vector < double > [numberRatesMatrices] (1);
    frequencies = NULL;
    
    //initialise the matrixIndex for correspondance between a matrix element
    //and an index in the rate ratios vector
    initMatrixIndex();
    
    //call the basic initialisation method
    DnaModel::initialisation( sequenceTable, modelId );
    
    //if a sequence table is provided initialise rate ratios and frequencies
    //accordingly
    if (sequenceTable){
        //estimation of all the rate ratios (and proportion of invariant sites
        //if relevant) from the empirical sequences
        vector < double > rate5;
        if ( invariantCategory ) {
            rate5 = retrieveEmpiricalRates( sequenceTable, modelId,
            & proportionInvariantSites );
        }
        else {
            rate5 = retrieveEmpiricalRates( sequenceTable, modelId );
        }
        //the transversion rate is the only free parameter
        rateRatios[0][0] = (rate5[0]+rate5[1]+rate5[2]+rate5[4])/(2*(1.0+rate5[3]));
        //if the discrete gamma model uses more than one set of rate ratios
        //initialise the other ones to the same values
        for ( int i = 1 ; i < numberRatesMatrices; i++ ) {
            rateRatios[i] = rateRatios[0];
        }
    }
    //init the equivalency table (correspondance symbol/states)
    initEquivalencyTable();
    //first initialisation of the mutation rate matrix
    //(factor(s) used with the subtitution matrix to reach the specified
    //average substitution rate) and eigen system
    setAverageMutationRate( averageMutationRate );
}
  
K80::~K80() {
}

string K80::getName( void ) const {
    char num[3];
    string name( "K80" );
    if ( discreteGamma ) {
        sprintf( num, "%i", numberGammaCategories );
        name = name + string( " + dG" ) + string( num );
    }
    if ( invariantCategory == 1 )
        name = name + string( " + I" );
    return ( name );
}

void K80::setAverageMutationRate( double newAverageRate, int symbolCategory ) {
    assert(symbolCategory==0);
    double mr, sum;
    assert( newAverageRate > 0.0 );
    averageMutationRate = newAverageRate;

    // compute mutation rate for non-invariant sites
    mr = averageMutationRate / ( 1.0 - proportionInvariantSites );
    // if a gamma distribution is used, update the mutation rate
    // of all gamma categories
    if ( discreteGamma ) {
        // compute the gamma scale parameter beta
        double beta = alpha / mr;
        statlib::gamma_percentile_means( mutationRate, alpha, beta,
        numberGammaCategories );
        for ( int i = 0 ; i < numberGammaCategories; ++i ) {
            sum = .25 + .5 * rateRatios[matrixCategory[i]] [0];
            mutationRate[i] /= sum;
        }
    }
    else {
        //        mutationRate.resize( 1 );
        sum = .25 + .5 * rateRatios[0] [0];
        mutationRate[0] = mr / sum;
    }
    setEigenMatrix();
}


void K80::printParameters( ostream & outputStream ) const{
    //print common parameters
    MatrixModel::printParameters( outputStream );
    
    outputStream << "f[A] = .25   f[C] = .25   f[G] = .25   f[U] = .25" << endl;
    outputStream << endl;
    outputStream.setf(ios::fixed);
    outputStream << setprecision(5);
    for ( int rateMat = 0; rateMat < numberRatesMatrices; ++rateMat ) {
        if (numberRatesMatrices>=2){
            outputStream << "set of rate ratios " << rateMat + 1 << ":" << endl;
        }
        else{
            outputStream << "rate ratios:" << endl;
        }
        outputStream << "    ";
        for ( int i = 0; i < getNumberStates(); ++i ) {
            outputStream << "      " << getState( i ) << "     ";
        }
        outputStream << endl;
        for ( int i = 0; i < getNumberStates(); ++i ) {
            outputStream << "  " << getState( i ) << "  ";
            for ( int j = 0; j < getNumberStates(); ++j ) {
                int index = matrixIndex( i, j);
                switch ( index ) {
                    //if diagonal
                    case -3 : outputStream << "  --------  ";
                              break;
                    //if reference
                    case -2 : outputStream << "   1.00000  ";
                              break;
                    //if null (not possible with implemented DNA models
                    case -1 : assert(0);
                              break;
                    //otherwise
                    default : outputStream << "  " << setw( 8 ) <<
                                  rateRatios [rateMat] [index] << "  ";
                              break;
                }
            }
            outputStream << endl; //end line
        }
        outputStream << endl; //end rate ratios matrix
    }
    if (numberRatesMatrices>=2){
        for ( int gammaCat = 0; gammaCat < MAX(numberGammaCategories,1); ++ gammaCat ) {
            if ( numberGammaCategories>1 ){
                outputStream << "transition rate matrix for the gamma category "
	                         << gammaCat + 1;
                if (numberRatesMatrices>=2){
                    outputStream << ", rate ratios set "
                                 << matrixCategory[gammaCat] + 1;
                }
                outputStream << ":" << endl;
            }
            else{
                outputStream << "transition rate matrix:" << endl;
            }
            outputStream << "    ";
            for ( int i = 0; i < getNumberStates(); ++i ) {
                outputStream << "      " << getState( i ) << "     ";
            }
            outputStream << endl;
            for ( int i = 0; i < getNumberStates(); ++i ) {
                outputStream << ' ' << getState( i ) << ' ';
                for ( int j = 0; j < getNumberStates(); ++j ) {
                    outputStream << "  " << setw( 9 )
                                 << rateMatrix[gammaCat] (i, j) << "  ";
                }
                outputStream << endl; //end line
            }
            outputStream << endl; //end transition rate matrix
        }
        outputStream << endl;
    }
    else{
        outputStream << "transition rate matrix (scaled so that "
                     << "the average substitution rate is 1.0):" << endl;
        double sum = .25 + .5 * rateRatios[0] [0];
        for ( int i = 0; i < getNumberStates(); ++i ) {
            outputStream << "      " << getState( i ) << "     ";
        }
        outputStream << endl;
        for ( int i = 0; i < getNumberStates(); ++i ) {
            outputStream << ' ' << getState( i ) << ' ';
            for ( int j = 0; j < getNumberStates(); ++j ) {
                int index = matrixIndex(i, j);
                switch(index){
                    //if diagonal
                    case -3 : outputStream << "  ---------  ";
                              break;
                    //if reference
                    case -2 : outputStream << "  " << setw( 9 )
                                           << .25 / sum << "  ";
                              break;
                    //if null
                    case -1 : assert(0);
                    //otherwise (>=0)
                    default : outputStream << "  " << setw( 9 )
                                           << .25 * rateRatios[0] [index] / sum << "  ";

                }
            }
            outputStream << endl; //end line
        }
        outputStream << endl; //end transition rate matrix
    }
    outputStream << endl;
    outputStream.unsetf(ios::fixed);
}

void K80::initMatrixIndex() {
    matrixIndex.resize( 4, 4 );
    for ( int i = 0 ; i < 4 ; ++i ) {
        matrixIndex(i, i) = -3;
    }
    //reference (transitions)
    matrixIndex(0, 2) = matrixIndex(2, 0) = -2;
    matrixIndex(1, 3) = matrixIndex(3, 1) = -2;
    //transversion
    matrixIndex(0, 1) = matrixIndex(1, 0) = 0;
    matrixIndex(0, 3) = matrixIndex(3, 0) = 0;
    matrixIndex(1, 2) = matrixIndex(2, 1) = 0;
    matrixIndex(2, 3) = matrixIndex(3, 2) = 0;
}


void K80::setEigenMatrix() {
    for ( int category = 0; category < MAX( numberGammaCategories, 1 );
    category++ ) {
        // fill the rateMatrix (cf Hky.h)
        int rateCat = matrixCategory[category];
        rateMatrix[category] ( 0, 1 ) = mutationRate[category] * .25 *
                                        rateRatios[rateCat] [0];
        rateMatrix[category] ( 0, 2 ) = mutationRate[category] * .25;
        rateMatrix[category] ( 0, 3 ) = rateMatrix[category] ( 0, 1 ) ;
        rateMatrix[category] ( 0, 0 ) = -2 * rateMatrix[category]( 0, 1 ) -
                                        rateMatrix[category]( 0, 2 );
        rateMatrix[category] ( 1, 0 ) = rateMatrix[category] ( 0, 1 );
        rateMatrix[category] ( 1, 2 ) = rateMatrix[category] ( 0, 1 );
        rateMatrix[category] ( 1, 3 ) = rateMatrix[category] ( 0, 2 );
        rateMatrix[category] ( 1, 1 ) = rateMatrix[category] ( 0, 0 );

        rateMatrix[category] ( 2, 0 ) = rateMatrix[category] ( 0, 2 );
        rateMatrix[category] ( 2, 1 ) = rateMatrix[category] ( 0, 1 );
        rateMatrix[category] ( 2, 3 ) = rateMatrix[category] ( 0, 1 );
        rateMatrix[category] ( 2, 2 ) = rateMatrix[category] ( 0, 0 );

        rateMatrix[category] ( 3, 0 ) = rateMatrix[category] ( 0, 1 );
        rateMatrix[category] ( 3, 1 ) = rateMatrix[category] ( 0, 2 );
        rateMatrix[category] ( 3, 2 ) = rateMatrix[category] ( 0, 1 );
        rateMatrix[category] ( 3, 3 ) = rateMatrix[category] ( 0, 0 );

        for ( int i = 0; i < 4; ++i ) {
            eigenValues[category] [i] = 0.0;
            for ( int j = 0; j < 4; ++j ) {
                eigenMatrix[category] ( i, j ) = 0.0;
                ieigenMatrix[category] ( i, j ) = 0.0;
            }
        }
        
        double transitionRate = mutationRate[category];
        double transversionRate = mutationRate[category] * rateRatios[rateCat][0];
        

        eigenValues[category] [0] = 0.0;
        eigenValues[category] [1] =
        -transitionRate * .5 - transversionRate * .5;
        eigenValues[category] [2] =
        -transitionRate * .5 - transversionRate * .5;
        eigenValues[category] [3] = -transversionRate;


        // Update the eigenmatrix by hand
        eigenMatrix[category] ( 0, 0 ) = .25;
        eigenMatrix[category] ( 0, 1 ) = .25;
        eigenMatrix[category] ( 0, 2 ) = .25;
        eigenMatrix[category] ( 0, 3 ) = .25;

        eigenMatrix[category] ( 1, 0 ) = 0.0;
        eigenMatrix[category] ( 1, 1 ) = -1.0;
        eigenMatrix[category] ( 1, 2 ) = 0.0;
        eigenMatrix[category] ( 1, 3 ) = 1.0;

        eigenMatrix[category] ( 2, 0 ) = -1.0;
        eigenMatrix[category] ( 2, 1 ) = 0.0;
        eigenMatrix[category] ( 2, 2 ) = 1.0;
        eigenMatrix[category] ( 2, 3 ) = 0.0;

        eigenMatrix[category] ( 3, 0 ) = -.25;
        eigenMatrix[category] ( 3, 1 ) = .25;
        eigenMatrix[category] ( 3, 2 ) = -.25;
        eigenMatrix[category] ( 3, 3 ) = .25;

        // Inverse of the eigenmatrix
        ieigenMatrix[category] ( 0, 0 ) = 1.0;
        ieigenMatrix[category] ( 0, 1 ) = 0.0;
        ieigenMatrix[category] ( 0, 2 ) = -.5;
        ieigenMatrix[category] ( 0, 3 ) = -1.0;

        ieigenMatrix[category] ( 1, 0 ) = 1.0;
        ieigenMatrix[category] ( 1, 1 ) = -.5;
        ieigenMatrix[category] ( 1, 2 ) = 0.0;
        ieigenMatrix[category] ( 1, 3 ) = 1.0;

        ieigenMatrix[category] ( 2, 0 ) = 1.0;
        ieigenMatrix[category] ( 2, 1 ) = 0.0;
        ieigenMatrix[category] ( 2, 2 ) = .5;
        ieigenMatrix[category] ( 2, 3 ) = -1.0;

        ieigenMatrix[category] ( 3, 0 ) = 1.0;
        ieigenMatrix[category] ( 3, 1 ) = .5;
        ieigenMatrix[category] ( 3, 2 ) = 0.0;
        ieigenMatrix[category] ( 3, 3 ) = 1.0;
    } //for each category (except invariant)
}


Model * K80::clone( ParametersSet & parameters ) const {
    if( parameters.find( "Number of frequencies categories" ) !=
            parameters.end() ){
        assert(parameters.intParameter("Number of frequencies categories") == 0);
    }
    else{
        parameters["Number of frequencies categories"] = "0";
    }
    return new K80( parameters );
}

