#include "Models/MatrixModel.h"

#include <iostream>
#include <iomanip>
#include <stdio.h>

#include "SequenceTable.h"

#include "Models/Perturbator.h"
#include "Util/ModelFactory.h"

#include "PatternDesign/Singleton.h"
#include "Util/statlib.h"
#include "Util/array2D.h"
#include "Util/randombox.h"
#include "Util/ParametersSet.h"

using namespace std;

MatrixModel::MatrixModel( const string & registrationName ) {
    Singleton < ModelFactory > & modelFactory = Singleton < ModelFactory >::instance();
    modelFactory.subscribe( this, registrationName );
}

MatrixModel::MatrixModel( ParametersSet & parameters ) {

    perturbator = NULL;

    //default value for the number of rates matrices equals to 1
    //model without rates (numberRatesRatioPerMatrix = 0)
    //must make sure this value is filled with 0 before construction
    //(for instance in the clone method)
    if ( parameters.find( "Number of rates matrices" ) ==
    parameters.end() ) {
        numberRatesMatrices = 1;
    }
    else {
        numberRatesMatrices = parameters.intParameter( "Number of rates matrices" );
    }

    //default value for the number of frequencies categories equals to 1
    //model without frequencies (numberFrequencies = 0)
    //must make sure this value is filled with 0 before construction
    //(for instance in the clone method)
    if ( parameters.find( "Number of frequencies categories" ) ==
    parameters.end() ) {
        numberFrequenciesCategories = 1;
    }
    else {
        numberFrequenciesCategories = parameters.intParameter( "Number of frequencies categories" );
    }

    if ( parameters.find( "Discrete gamma distribution of rates" ) ==
    parameters.end() ) {
        discreteGamma = false;
    }
    else {
        discreteGamma = parameters.boolParameter( "Discrete gamma distribution of rates" );
    }

    // invariant category, the invariant will be computed later if a
    // sequenceTable is used to initialise the Model
    if ( ( parameters.find( "Invariant sites" ) != parameters.end() ) &&
    ( parameters.boolParameter( "Invariant sites" ) ) ) {
        invariantCategory = 1;
        proportionInvariantSites = 0.2;
    }
    else {
        invariantCategory = 0;
        proportionInvariantSites = 0.0;
    }

    // Number of categories to use to approximate the gamma distribution
    if ( !discreteGamma ) {
        // space allocation
        numberGammaCategories = 0;
        // Total number categories
        numberRateCategories = 1 + invariantCategory;
        // if the model uses some substitution rates
        if (numberRatesMatrices != 0){
            //only one rate matrix allowed if we do not use a gamma
            //distribution
            assert( numberRatesMatrices == 1 );
            matrixCategory.resize( 1 );
            matrixCategory[0] = 0;
        }
        else{
            matrixCategory.resize( 0 );
        }
        // if the model uses some frequencies
        if ( numberFrequenciesCategories != 0 ){
            // only one frequencies category allowed if we do not use a
            // gamma distribution and no invariant
            if ( !invariantCategory ){
                assert( numberFrequenciesCategories == 1 );
                frequencyCategory.resize(1);
                frequencyCategory[0] = 0;
            }
            // two allowed categories otherwise (but not compulsory)
            else{
                assert( (numberFrequenciesCategories == 1)||(numberFrequenciesCategories==2) );
                frequencyCategory.resize(2);
                frequencyCategory[0] = 0;
                frequencyCategory[1] = numberFrequenciesCategories-1;
            }
        }
        else{  //no frequencies in the model
            frequencyCategory.resize(0);
        }
    }
    else {
        numberGammaCategories = parameters.intParameter( "Number of gamma categories" );
        // At least 1 gamma category
        assert( numberGammaCategories >= 1 );
        // the gamma shape parameter, it seems that this default value gives
        // better results during a ML inference than a more realistic one
        // (like .3)
        alpha = 0.1;
        // space allocation
        // Total number of rate categories
        numberRateCategories = numberGammaCategories + invariantCategory;

        // if the model uses some substitution rates
        if ( numberRatesMatrices != 0 ){
            assert( (numberRatesMatrices >= 1) && (numberRatesMatrices <= numberGammaCategories) );
            matrixCategory.resize( numberGammaCategories );
            //if no indication provided
            //initialise the matrixCategory vector used to match a rate category
            //with a matrix
            if ( parameters.find( "Matrix of category 1" ) ==
                parameters.end() ) {
                //allow only two simple case only one matrice for all the categories
                // or one matrice for each category
                assert( ( numberRatesMatrices == 1 ) ||
                ( numberRatesMatrices == numberGammaCategories ) );
                if ( numberRatesMatrices == 1 ) {
                    for ( int i = 0; i < numberGammaCategories; ++i ) {
                        this->matrixCategory[i] = 0;
                    }
                }
                if ( numberRatesMatrices == numberGammaCategories ) {
                    for ( int i = 0; i < numberGammaCategories; ++i ) {
                        this->matrixCategory[i] = i;
                    }
                }
            }
            else {
                int mat = 0;
	            char label[25];
                sprintf( label, "Matrix of category ");
                char* numberPos = label;
                while (*(++numberPos));
                for ( int i = 0; i < numberGammaCategories; ++i ) {
                    sprintf( numberPos, "%d", i + 1);
                    matrixCategory[i] = parameters.intParameter( label )-1;
                    if ( matrixCategory[i] < mat ) {
                        cout << "WARNING ! Strange allocation of a rate matrix for "
                        << "the rate category " << i+1 << endl;
                    }
                    mat = matrixCategory[i];
                    if ( ( mat < 0 ) || ( mat >= numberRatesMatrices ) ) {
                        cerr << "ERROR ! assigning rate matrice " << mat+1
                        << "for the rate category " << i + 1 << endl;
                        exit( 1 );
                    }
                }
            }
        }
        else{
            matrixCategory.resize( 0 );
        }

        // if the model uses some frequencies
        if ( numberFrequenciesCategories != 0 ){
            // initialise the frequencyCategory vector used to match a
            // gamma (or invariant) category with a set of frequencies
            assert( (numberFrequenciesCategories >= 1) && (numberFrequenciesCategories <= numberRateCategories) );
            frequencyCategory.resize( numberRateCategories );
            //if no indication provided
            //initialise the frequencyCategory vector used to match a rate category
            //with a matrix
            if ( parameters.find( "Frequencies for category 1" ) ==
                parameters.end() ) {
                // allow only two simple case only one set of frequencies for all the
                // rate categories(number of gamma categories + invariant category )
                // or one set for each category
                assert( ( numberFrequenciesCategories == 1 ) ||
                ( numberFrequenciesCategories == numberRateCategories ) );
                if ( numberFrequenciesCategories == 1 ) {
                    for ( int i = 0; i < numberRateCategories; ++i ) {
                        this->frequencyCategory[i] = 0;
                    }
                }
                if ( numberFrequenciesCategories == numberRateCategories ) {
                    for ( int i = 0; i < numberRateCategories; ++i ) {
                        this->frequencyCategory[i] = i;
                    }
                }
            }
            else {
                if (invariantCategory){
                    frequencyCategory[0] = parameters.intParameter( "Frequencies for invariant category" ) - 1;
                    assert( (frequencyCategory[0]>=0) && (frequencyCategory[0]<numberFrequenciesCategories) );
                    if ( frequencyCategory[0] != 0 ){
                        cout << "WARNING ! Strange allocation for the frequencies of"
                             << "the invariant category, 1 is usually used" << endl;
                    }
                }
                int cat = 0;
	            char label[35];
                sprintf( label, "Frequencies for category ");
                char* numberPos = label;
                while (*(++numberPos));
                for ( int i = 0; i < MAX(numberGammaCategories,1); ++i ) {
                    sprintf( numberPos, "%d", i + 1 );
                    frequencyCategory[i+invariantCategory] = parameters.intParameter( label )-1;
                    if ( frequencyCategory[i+invariantCategory] < cat ) {
                        cout << "WARNING ! Strange allocation of frequencies for "
                        << "the rate category " << i+1 << endl;
                    }
                    cat = frequencyCategory[i+invariantCategory];
                    if ( ( cat < 0 ) || ( cat >= numberFrequenciesCategories ) ) {
                        cerr << "ERROR ! assigning rate matrice " << cat+1
                        << "for the rate category " << i + 1 << endl;
                        exit( 1 );
                    }
                }
            }
        }
        else{
            frequencyCategory.resize( numberRateCategories );
        }
    }
    // space allocation
    mutationRate.resize( numberRateCategories - invariantCategory );
}

MatrixModel::~MatrixModel() {
}

void MatrixModel::initialisation( SequenceTable * sequenceTable, int modelId ) {
    modelId = modelId;
    //frequencies and rateRatios must have the good size
    if ( sequenceTable == NULL ) {
        for ( int i = 0; i <  numberFrequencies; ++i ) {
            frequencies[0] [i] = 1.0 / numberFrequencies;
        }
        for ( int i = 1; i < numberFrequenciesCategories; ++i ) {
            frequencies[i] = frequencies[0];
        }

        for ( int i = 0; i < numberRatesRatiosPerMatrix; ++i ) {
            rateRatios[0] [i] = 1.0;
        }
        for ( int i = 1; i < numberRatesMatrices; ++i ) {
            rateRatios[i] = rateRatios[0];
        }
    }
    else {
        //in this case each model must initialise itself from the sequenceTable
    }
    averageMutationRate = 1.0;
}


ParametersSet MatrixModel::getModelParameters(){
    ParametersSet parameters("MODEL",false);
    ParametersSet* currentCategory;

    char tmpString[40];
    char dblString[40];

    parameters["Model name"] = getName();
    if (numberFrequenciesCategories>=1){
        parameters.addCategory("FREQUENCIES");
        if (numberFrequenciesCategories>=2){
            sprintf(tmpString,"%d",numberFrequenciesCategories);
            parameters("FREQUENCIES")["Number of frequencies categories"] =
                tmpString;
            for ( int i = 0; i < numberFrequenciesCategories; ++i ){
                sprintf(tmpString,"FREQUENCIESCATEGORY%d",i+1);
                parameters("FREQUENCIES").addCategory(tmpString);
                currentCategory = &parameters("FREQUENCIES")(tmpString);
                for (int j = 0; j < numberFrequencies; ++j){
                    sprintf(dblString, "%.5f",frequencies[i][j]);
                    (*currentCategory)["F("+getSymbol(j)+")"] = dblString;
                }
            }
        }
        else{
            for (int j = 0; j < numberFrequencies; ++j){
                sprintf(dblString, "%.5f",frequencies[0][j]);
                parameters("FREQUENCIES")["F("+getSymbol(j)+")"] = dblString;
            }
        }
    }
    
    if (numberRatesMatrices>=1){
        parameters.addCategory("RATESRATIOS");
        if (numberRatesMatrices>=2){
            sprintf(tmpString,"%d",numberRatesMatrices);
            parameters("RATESRATIOS")["Number of rates matrices"] = tmpString;
            for ( int i = 0; i < numberRatesMatrices; ++i ){
                sprintf(tmpString,"RATESRATIOSCATEGORY%d",i+1);
                parameters("RATESRATIOS").addCategory(tmpString);
                currentCategory = &parameters("RATESRATIOS")(tmpString);
                for (int j = 0; j < numberRatesRatiosPerMatrix; ++j){
                    sprintf(dblString, "%.5f",rateRatios[i][j]);
                    sprintf(tmpString,"Rate ratio %d",j+1);	    
                    (*currentCategory)[tmpString] = dblString;
                }
            }
        }
        else{
            for (int j = 0; j < numberRatesRatiosPerMatrix; ++j){
                sprintf(dblString, "%.5f",rateRatios[0][j]);
                sprintf(tmpString,"Rate ratio %d",j+1);	    
                parameters("RATESRATIOS")[tmpString] = dblString;
            }
        }
    }

    parameters.addCategory("RATESCATEGORIES");
    parameters("RATESCATEGORIES")["Invariant category"] =
        (invariantCategory ? "yes" : "no");
    if (invariantCategory){
        sprintf(dblString, "%.6f",proportionInvariantSites);
        parameters("RATESCATEGORIES")["Proportion of invariant sites"] =
            dblString;
    }
    parameters("RATESCATEGORIES")["Gamma distribution of rates"] =
        (discreteGamma ? "yes" : "no");
    if ( discreteGamma ){
        sprintf(tmpString,"%d",numberGammaCategories);
        parameters("RATESCATEGORIES")["Number of gamma categories"] =
            tmpString;
        sprintf(dblString, "%.6f",alpha);
        parameters("RATESCATEGORIES")["Alpha parameter"] = dblString;
    }
    if ( (numberFrequenciesCategories>=2) || (numberRatesMatrices>=2) ){
        for ( int i = 0; i < numberRateCategories; ++i ){
            sprintf(tmpString,"RATESCATEGORY%d",i+1);
            parameters("RATESCATEGORIES").addCategory(tmpString);
            currentCategory = &parameters("RATESCATEGORIES")(tmpString);
            if  ( numberFrequenciesCategories >= 2 ){
                sprintf(tmpString,"%d",frequencyCategory[i]+1);
                (*currentCategory)["Frequencies for the category"] = tmpString;
            }
            if ( ( numberRatesMatrices >= 2 ) &&
                 ( ( i != 0 ) || !invariantCategory ) ){
                sprintf(tmpString,"%d",matrixCategory[i-invariantCategory]+1);
                (*currentCategory)["Rate ratios for the category"] = tmpString;
            }
        }
    }
    return parameters;
}


/** ************************************************************************
 * setModelParameters
 * @semantics      set the parameters of the model according to the content
 *                 of the parameters set
 * @preconditions  the parameters in the set are consistent
 ************************************************************************ */
void MatrixModel::setModelParameters( ParametersSet& parameters ){

    ParametersSet* currentCategory;
    
    char tmpString[40];

    if (getName() != parameters.stringParameter("Model name")){
        cerr << "You are trying to initialise the model " << getName() << endl;
        cerr << "with parameters for the model "
             << parameters.stringParameter("Model name") << endl;
        cerr << "Exit..." << endl;
        exit(1);
    }
    if ( numberFrequenciesCategories >= 1 ){
        if ( numberFrequenciesCategories >= 2 ){
            assert( numberFrequenciesCategories ==
                parameters("FREQUENCIES").intParameter("Number of frequencies categories") );
            for ( int i = 0; i < numberFrequenciesCategories; ++i ){
                sprintf(tmpString,"FREQUENCIESCATEGORY%d",i+1);
                currentCategory = &parameters("FREQUENCIES")(tmpString);
                double sum = 0.0;
                for (int j = 0; j < numberFrequencies; ++j){
                    frequencies[i][j] = currentCategory->doubleParameter("F("+getSymbol(j)+")");
                    sum += frequencies[i][j];
                }
                if(fabs(sum-1.0)>=0.02){
                    cerr << "WARNING: frequencies do not sum to one in your model file" << endl;
                }
            }
        }
        else{
            double sum = 0.0;
            for (int j = 0; j < numberFrequencies; ++j){
                frequencies[0][j] = parameters("FREQUENCIES").doubleParameter("F("+getSymbol(j)+")");
                sum += frequencies[0][j];
            }
            if(fabs(sum-1.0)>=0.02){
                cerr << "WARNING: frequencies do not sum to one in your model file" << endl;
            }
        }
    }
    if ( numberRatesMatrices >= 1){
        if ( numberRatesMatrices >= 2 ){
            assert( numberRatesMatrices ==
            parameters("RATESRATIOS").intParameter("Number of rates matrices") );
            for ( int i = 0; i < numberRatesMatrices; ++i ){
                sprintf(tmpString,"RATESRATIOSCATEGORY%d",i+1);
                currentCategory = &parameters("RATESRATIOS")(tmpString);
                for (int j = 0; j < numberRatesRatiosPerMatrix; ++j){
                    sprintf(tmpString,"Rate ratio %d",j+1);	    
                    rateRatios[i][j] = currentCategory->doubleParameter(tmpString);
                }
            }
        }
        else{
            for (int j = 0; j < numberRatesRatiosPerMatrix; ++j){
                sprintf(tmpString,"Rate ratio %d",j+1);	    
                rateRatios[0][j] = parameters("RATESRATIOS").doubleParameter(tmpString);
            }
        }
    }

    if ( (invariantCategory == 1) !=
        parameters("RATESCATEGORIES").boolParameter("Invariant category") ){
        cerr << "The name of your model "
             << parameters.stringParameter("Model name") << endl;
        cerr << "does not match with the \"Invariant category\" field "
             << "in your model file" << endl;
        cerr << "Exit..." << endl;
        exit(1);        
    }
    if (invariantCategory){
        proportionInvariantSites =
            parameters("RATESCATEGORIES").doubleParameter("Proportion of invariant sites");
    }
    if( discreteGamma !=
        parameters("RATESCATEGORIES").boolParameter("Gamma distribution of rates") ){
        cerr << "The name of your model "
             << parameters.stringParameter("Model name") << endl;
        cerr << "does not match with the \"Gamma distribution of rates\" field "
             << "in your model file" << endl;
        cerr << "Exit..." << endl;
        exit(1);        
    }
    if ( discreteGamma ){
        if ( numberGammaCategories !=
            parameters("RATESCATEGORIES").intParameter("Number of gamma categories") ){
            cerr << "The name of your model "
                 << parameters.stringParameter("Model name") << endl;
            cerr << "does not match with the \"Number of gamma categories\" field "
                 << "in your model file" << endl;
            cerr << "Exit..." << endl;
            exit(1);        
        }
        alpha = parameters("RATESCATEGORIES").doubleParameter("Alpha parameter");
    }
    if ( (numberFrequenciesCategories>=2) || (numberRatesMatrices>=2) ){
        for ( int i = 0; i < numberRateCategories; ++i ){
            sprintf(tmpString,"RATESCATEGORY%d",i+1);
            currentCategory = &parameters("RATESCATEGORIES")(tmpString);
            if  ( numberFrequenciesCategories >= 2 ){
                frequencyCategory[i] = currentCategory->intParameter("Frequencies for the category") - 1;
                assert( (frequencyCategory[i]>=0) &&
                    (frequencyCategory[i]<numberFrequenciesCategories) );
            }
            if ( ( numberRatesMatrices >= 2 ) &&
                 ( ( i != 0 ) || !invariantCategory ) ){
                matrixCategory[i-invariantCategory] = currentCategory->intParameter("Rate ratios for the category") - 1;
                assert( ( matrixCategory[i-invariantCategory]>=0 ) &&
                        ( matrixCategory[i-invariantCategory]<numberRatesMatrices ) );
            }
        }
    }
    setAverageMutationRate( averageMutationRate );
}


// Functions to get and set the number og constant sites
double MatrixModel::getProportionInvariantSites( int symbolCategory ) const {
    assert( symbolCategory == 0 );
    assert( invariantCategory == 1 );
    return proportionInvariantSites;
}


void MatrixModel::setProportionInvariantSites( double newProp, int symbolCategory ) {
    assert( symbolCategory == 0 );
    assert ( invariantCategory == 1 );
    assert( ( newProp >= 0 ) && ( newProp <= 1 ) );
    proportionInvariantSites = newProp;
    setAverageMutationRate( averageMutationRate );
}


double MatrixModel::getGammaShapeParameter( int symbolCategory ) const {
    assert( symbolCategory == 0 );
    assert( discreteGamma );
    return alpha;
}


void MatrixModel::setGammaShapeParameter( double newAlpha, int symbolCategory ) {
    assert( symbolCategory == 0 );
    assert( discreteGamma );
    assert( newAlpha > 0.0 );
    alpha = newAlpha;
    setAverageMutationRate( averageMutationRate );
}

double MatrixModel::getAverageMutationRate( int symbolCategory ) const {
    assert(symbolCategory==0);
    return averageMutationRate;
}


void MatrixModel::getAllParameters( vector < double > & pars ) const {


    int npars = ( numberFrequencies - 1 ) * numberFrequenciesCategories +
    numberRatesRatiosPerMatrix * numberRatesMatrices +
    ( discreteGamma ? 1 : 0 ) + invariantCategory;

    pars.resize( npars );

    int index = -1;
    double sum;

    for ( int freqCat = 0 ; freqCat < numberFrequenciesCategories ;
    ++freqCat ) {
        // The first parameters determine the base-pair frequencies
        sum = ( 1.0 - frequencies[freqCat] [numberFrequencies - 1] ) /
        frequencies[freqCat] [numberFrequencies - 1];
#ifdef DEBUG
        assert( !isnan( sum ) );
        assert( !isinf( sum ) );
#endif
        for ( int freq = 0; freq < numberFrequencies - 1; ++freq ) {
            pars[++index] = sqrt( ( 1.0 + sum ) * frequencies[freqCat] [freq] );
        }
    }

    for ( int rateMat = 0 ; rateMat < numberRatesMatrices ; ++rateMat ) {
        for ( int ratio = 0; ratio < numberRatesRatiosPerMatrix; ++ratio ) {
            pars[++index] = sqrt( rateRatios[rateMat] [ratio] );
#ifdef DEBUG
            assert( !isinf( rateRatios[rateMat] [ratio] ) );
            assert( !isnan( rateRatios[rateMat] [ratio] ) );
            assert( rateRatios[rateMat] [ratio] >= 0.0 );
#endif
        }
    }

    if ( discreteGamma ) {
        pars[++index] = sqrt( alpha );

    }
    if ( invariantCategory == 1 ) {
        pars[++index] = sqrt( proportionInvariantSites /
        ( 1.0 - proportionInvariantSites ) );
    }
}


void MatrixModel::setAllParameters( const vector < double > & pars ) {

    assert( (int)pars.size() == ( numberFrequencies - 1 )  * numberFrequenciesCategories +
    numberRatesRatiosPerMatrix * numberRatesMatrices +
    ( discreteGamma ? 1 : 0 ) + invariantCategory );

    double sum;

    int index = 0;
    for ( int freqCat = 0 ; freqCat <
    numberFrequenciesCategories ; ++freqCat ) {
        sum = 1.0;
        for ( int i = 0; i < numberFrequencies - 1; ++i ) {
            sum += ( pars[index + i] * pars[index + i] );
        }
        for ( int i = 0; i < numberFrequencies - 1; ++i ) {
            frequencies[freqCat] [i] = ( pars[index + i] * pars[index + i] ) / sum;
#ifdef DEBUG
            assert( !isnan( frequencies[freqCat] [i] ) );
            assert( !isinf( frequencies[freqCat] [i] ) );
#endif
        }
        frequencies[freqCat] [numberFrequencies - 1] = 1.0 / sum;

        index += numberFrequencies - 1;
    }

    for ( int rateMat = 0 ; rateMat < numberRatesMatrices ; ++rateMat ) {
        for ( int ratio = 0; ratio < numberRatesRatiosPerMatrix; ++ratio ) {
            rateRatios[rateMat] [ratio] = ( pars[index] * pars[index] );
#ifdef DEBUG
            if ( isnan( rateRatios[rateMat] [ratio] ) ) {
                cout << "rate ratio = " << ratio << endl;
                cout << "Index     = " << index << endl;
                cout << "Parameter = " << pars[index] << endl;
                printParameters( cout );
                assert( !isnan( rateRatios[rateMat] [ratio] ) );
            }
            assert( !isinf( rateRatios[rateMat] [ratio] ) );
#endif
            ++index;
        }
    }

    if ( discreteGamma ) {
        alpha = pow( pars[index], 2.0 );
        ++index;
    }
    if ( invariantCategory == 1 ) {
        proportionInvariantSites =
        pow( pars[index], 2.0 ) / ( 1.0 + pow( pars[index], 2.0 ) );
    }
    setAverageMutationRate( averageMutationRate );
}


unsigned int MatrixModel::getNumberFreeParameters(){
    return ( ( numberFrequencies - 1 ) * numberFrequenciesCategories +
             numberRatesRatiosPerMatrix * numberRatesMatrices +
             ( discreteGamma ? 1 : 0 ) + invariantCategory);
}



double MatrixModel::getRateCategoryProbability( int category, int symbolCategory ) const {
    assert( symbolCategory == 0 );

    if ( invariantCategory == 1 ) {
        if ( category == 0 ) {
            return proportionInvariantSites;
        }
        assert ( ( category > 0 ) && ( category < numberRateCategories ) );
        if ( discreteGamma ) {
            return ( 1.0 - proportionInvariantSites ) /
            ( ( double )numberGammaCategories );
        }
        else {
            return ( 1.0 - proportionInvariantSites );
        }
    }
    // no invariant
    else {
        assert( ( category >= 0 ) && ( category < numberRateCategories ) );
        if ( discreteGamma ) {
            return ( 1.0 / ( double )numberGammaCategories );
        }
        else {
            return ( 1.0 );
        }
    }
}


void MatrixModel::initialisePerturbator( ParametersSet & parameters ) {

    char label[50];

    perturbator = new Perturbator();
        
    if (numberFrequencies != 0){
        int freqPrio = parameters.intParameter("Frequencies, proposal priority");
        double minFreqAcceptanceRate = .21;
        double maxFreqAcceptanceRate = .25;
        double freqStep = 1000.0;
        //dynamic adaptation, acceptance rate range provided...
        if ( parameters.find("Frequencies, proposal minimum acceptance rate") !=
                    parameters.end() ){
                minFreqAcceptanceRate = parameters.doubleParameter
                    ("Frequencies, proposal minimum acceptance rate");
                maxFreqAcceptanceRate = parameters.doubleParameter
                    ("Frequencies, proposal maximum acceptance rate");
            if (minFreqAcceptanceRate < 0.0){
                cerr << "Proposal minimum acceptance rate lower than 0.0 for frequencies : "
                     << minFreqAcceptanceRate << endl;
                exit(1);
            }
            if (maxFreqAcceptanceRate > 1.0){
                cerr << "Proposal maximum acceptance rate greater than 1.0 for frequencies : "
                     << maxFreqAcceptanceRate << endl;
                exit(1);
            }
            if ( (minFreqAcceptanceRate>=maxFreqAcceptanceRate) ){
                cerr << "Error with the proposal acceptance rates for frequencies" << endl;
                cerr << "minimum acceptance rate = " << minFreqAcceptanceRate << endl;
                cerr << "maximum acceptance rate = " << maxFreqAcceptanceRate << endl;
                cerr << minFreqAcceptanceRate << " >= "
                     << maxFreqAcceptanceRate << endl;
                exit(1);
            }            
        }
        // if the dynamical adaptation is turned off, the tuning parameter
        // is compulsory (otherwise it is optionnal)
        if ( (minFreqAcceptanceRate == 0.0) && (maxFreqAcceptanceRate==1.0) ){
            freqStep = parameters.doubleParameter
                        ("Frequencies, initial Dirichlet tuning parameter");
        }
        else{
            if ( parameters.find("Frequencies, initial Dirichlet tuning parameter") !=
                 parameters.end() ){
                freqStep = parameters.doubleParameter
                        ("Frequencies, initial Dirichlet tuning parameter");
            }
        }
        if ( ( freqStep <= 100.0 ) || ( freqStep >= 100000.0 ) ){
            cerr << "initial Dirichlet tuning parameter out "
                 << "of range [100.0, 100000.0] for frequencies" << endl;
            exit(1);
        }
        if (freqPrio != 0){
            sprintf ( label, "Frequencies, Set ");
            char* numberPos = label;
            while (*(++numberPos));
            for ( int i = 0; i < numberFrequenciesCategories; ++i ) {
                sprintf( numberPos, "%d", i+1 );
                perturbator->registerFrequencies( label, &(frequencies[i]),
                  freqPrio, freqStep, minFreqAcceptanceRate, maxFreqAcceptanceRate );
            }
        }
    }
    
    if (numberRatesMatrices != 0){
        int ratesPrio = parameters.intParameter("Rate ratios, proposal priority");
        double minRatesAcceptanceRate = .21;
        double maxRatesAcceptanceRate = .25;
        double ratesStep = .2;        
        //dynamic adaptation, acceptance rate range provided...
        if ( parameters.find("Rate ratios, proposal minimum acceptance rate") !=
                    parameters.end() ){
                minRatesAcceptanceRate = parameters.doubleParameter
                    ("Rate ratios, proposal minimum acceptance rate");
                maxRatesAcceptanceRate = parameters.doubleParameter
                    ("Rate ratios, proposal maximum acceptance rate");
            if (minRatesAcceptanceRate < 0.0){
                cerr << "Proposal minimum acceptance rate lower than 0.0 for rate ratios : "
                     << minRatesAcceptanceRate << endl;
                exit(1);
            }
            if (maxRatesAcceptanceRate > 1.0){
                cerr << "Proposal maximum acceptance rate greater than 1.0 for rate ratios : "
                     << maxRatesAcceptanceRate << endl;
                exit(1);
            }
            if ( (minRatesAcceptanceRate>=maxRatesAcceptanceRate) ){
                cerr << "Error with the proposal acceptance rates for rate ratios" << endl;
                cerr << "minimum acceptance rate = " << minRatesAcceptanceRate << endl;
                cerr << "maximum acceptance rate = " << maxRatesAcceptanceRate << endl;
                cerr << minRatesAcceptanceRate << " >= "
                     << maxRatesAcceptanceRate << endl;
                exit(1);
            }            
        }
        // if the dynamical adaptation is turned off, the tuning parameter
        // is compulsory (otherwise it is optionnal)
        if ( (minRatesAcceptanceRate == 0.0) && (maxRatesAcceptanceRate==1.0) ){
            ratesStep = parameters.doubleParameter
                        ("Rate ratios, initial step");
        }
        else{
            if ( parameters.find("Rate ratios, initial step") !=
                 parameters.end() ){
                ratesStep = parameters.doubleParameter
                        ("Rate ratios, initial step");
            }
        }
                
        if ( ( ratesStep == 0.0 ) || ( ratesStep >= 5.0 ) ){
            cerr << "Initial step out of range ]0, 5.0]"
                 << "for rate ratios" << endl;
            exit(1);
        }
        if (ratesPrio != 0){
            sprintf ( label, "Rate Ratio, Set ");
            char* numberPos1 = label;
            while (*(++numberPos1));
            for ( int i = 0; i < numberRatesMatrices; ++i ) {
                sprintf( numberPos1, "%d, no ", i+1 );
                char* numberPos2 = numberPos1;
                while (*(++numberPos2));
                for ( int j = 0; j < numberRatesRatiosPerMatrix; ++j ){
                    sprintf( numberPos2, "%d", j+1 );
                    perturbator->registerGaussParameter( label,
                               &(rateRatios[i][j]),
                               ratesPrio, ratesStep, 0.0, 1000.0,
                               minRatesAcceptanceRate, maxRatesAcceptanceRate );
                }
            }
        }
    }
    
    if ( invariantCategory ) {
        int invariantPrio = parameters.intParameter
                                          ("Invariant parameter, proposal priority");
        double minInvariantAcceptanceRate = .21;
        double maxInvariantAcceptanceRate = .25;
        double invariantStep = .05;
        
        //dynamic adaptation, acceptance rate range provided...
        if ( parameters.find("Invariant parameter, proposal minimum acceptance rate") !=
                    parameters.end() ){
                minInvariantAcceptanceRate = parameters.doubleParameter
                    ("Invariant parameter, proposal minimum acceptance rate");
                maxInvariantAcceptanceRate = parameters.doubleParameter
                    ("Invariant parameter, proposal maximum acceptance rate");
            if (minInvariantAcceptanceRate < 0.0){
                cerr << "Proposal minimum acceptance rate lower than 0.0 for the invariant parameter : "
                     << minInvariantAcceptanceRate << endl;
                exit(1);
            }
            if (maxInvariantAcceptanceRate > 1.0){
                cerr << "Proposal maximum acceptance rate greater than 1.0 for the invariant parameter : "
                     << maxInvariantAcceptanceRate << endl;
                exit(1);
            }
            if ( (minInvariantAcceptanceRate>=maxInvariantAcceptanceRate) ){
                cerr << "Error with the proposal acceptance rates for the invariant parameter" << endl;
                cerr << "minimum acceptance rate = " << minInvariantAcceptanceRate << endl;
                cerr << "maximum acceptance rate = " << maxInvariantAcceptanceRate << endl;
                cerr << minInvariantAcceptanceRate << " >= "
                     << maxInvariantAcceptanceRate << endl;
                exit(1);
            }            
        }
        // if the dynamical adaptation is turned off, the tuning parameter
        // is compulsory (otherwise it is optionnal)
        if ( (minInvariantAcceptanceRate == 0.0) && (maxInvariantAcceptanceRate==1.0) ){
            invariantStep = parameters.doubleParameter
                        ("Invariant parameter, initial step");
        }
        else{
            if ( parameters.find("Invariant parameter, initial step") !=
                 parameters.end() ){
                invariantStep = parameters.doubleParameter
                        ("Invariant parameter, initial step");
            }
        }
        if ( ( invariantStep <= 0.0 ) || ( invariantStep > .1 ) ){
            cerr << "Initial step out of range ]0, 0.1]"
                 << "for the invariant parameter" << endl;
            exit(1);
        }
        if ( invariantPrio != 0 ){
            perturbator->registerGaussParameter( "Invariant",
                                   &proportionInvariantSites,
                                   invariantPrio, invariantStep, 0.0, 1.0,
                                   minInvariantAcceptanceRate, maxInvariantAcceptanceRate );
        }
    }
      
    if ( discreteGamma ) {
        int alphaPrio = parameters.intParameter("Gamma parameter, proposal priority");
        double minAlphaAcceptanceRate = .21;
        double maxAlphaAcceptanceRate = .25;
        double alphaStep = .2;
        
        //dynamic adaptation, acceptance rate range provided...
        if ( parameters.find("Gamma parameter, proposal minimum acceptance rate") !=
                    parameters.end() ){
                minAlphaAcceptanceRate = parameters.doubleParameter
                    ("Gamma parameter, proposal minimum acceptance rate");
                maxAlphaAcceptanceRate = parameters.doubleParameter
                    ("Gamma parameter, proposal maximum acceptance rate");
            if (minAlphaAcceptanceRate < 0.0){
                cerr << "Proposal minimum acceptance rate lower than 0.0 for the gamma parameter : "
                     << minAlphaAcceptanceRate << endl;
                exit(1);
            }
            if (maxAlphaAcceptanceRate > 1.0){
                cerr << "Proposal maximum acceptance rate greater than 1.0 for the gamma parameter : "
                     << maxAlphaAcceptanceRate << endl;
                exit(1);
            }
            if ( (minAlphaAcceptanceRate>=maxAlphaAcceptanceRate) ){
                cerr << "Error with the proposal acceptance rates for the gamma parameter" << endl;
                cerr << "minimum acceptance rate = " << minAlphaAcceptanceRate << endl;
                cerr << "maximum acceptance rate = " << maxAlphaAcceptanceRate << endl;
                cerr << minAlphaAcceptanceRate << " >= "
                     << maxAlphaAcceptanceRate << endl;
                exit(1);
            }            
        }
        // if the dynamical adaptation is turned off, the tuning parameter
        // is compulsory (otherwise it is optionnal)
        if ( (minAlphaAcceptanceRate == 0.0) && (maxAlphaAcceptanceRate==1.0) ){
            alphaStep = parameters.doubleParameter
                        ("Gamma parameter, initial step");
        }
        else{
            if ( parameters.find("Gamma parameter, initial step") !=
                 parameters.end() ){
                alphaStep = parameters.doubleParameter
                        ("Gamma parameter, initial step");
            }
        }
        if ( ( alphaStep <= 0.0 ) || ( alphaStep > 10000.0 ) ){
            cerr << "Initial step out of range ]0, 10000.0]"
                 << "for the gamma parameter" << endl;
            exit(1);
        }
        if (alphaPrio!=0){
            perturbator->registerGaussParameter( "Gamma", &alpha,
                           alphaPrio, alphaStep, 0.0, 200000.0,
                           minAlphaAcceptanceRate, maxAlphaAcceptanceRate );
        }
    }
}

void MatrixModel::printParameters( ostream & outputStream ) const{
    outputStream.setf(ios::fixed);
    outputStream << setprecision( 4 );
    outputStream << "Average Mutation Rate = " << averageMutationRate << endl;
    if ( numberGammaCategories > 1 ) {
        outputStream << "Gamma shape parameter = " << alpha << endl;
    }
    if ( invariantCategory == 1 ) {
        outputStream << "Proportion of invariant sites = "
                     << proportionInvariantSites << endl;
    }
    outputStream.unsetf(ios::fixed);
}  

void MatrixModel::printPerturbationParameters( ostream & outputStream ) {
    assert( perturbator );
    perturbator->printParameters( outputStream );
}


double MatrixModel::perturb() {
    double res = perturbator->perturb();
    setAverageMutationRate( averageMutationRate );
    return res;
}


bool MatrixModel::validatePerturbation( bool validation ) {
    bool res = perturbator->validatePerturbation( validation );
    if (!validation){
        setAverageMutationRate( averageMutationRate );
    }
    return res;    
}

void MatrixModel::stopBurn(){
    perturbator->stopBurn();
}

void MatrixModel::printLine( ostream & outputStream ) {
    int i;
    outputStream.setf(ios::fixed);
    outputStream << setprecision( 4 );
    if ( invariantCategory ){
        outputStream << setw( 10 ) << proportionInvariantSites << ' ';
    }
    if (discreteGamma){
        outputStream << setw( 14 ) << alpha << ' ';
    }
    outputStream << setprecision( 4 );
    for (int freqCat = 0; freqCat < numberFrequenciesCategories; ++ freqCat ){
        for ( i = 0; i < numberFrequencies; ++i ) {
            outputStream << setw( 9 ) << frequencies[freqCat] [i] << ' ';
        }
    }
    for ( int rateMat = 0 ; rateMat < numberRatesMatrices; ++rateMat ) { 
        for ( int rate = 0 ; rate < numberRatesRatiosPerMatrix; ++rate ) {
            outputStream << setw( 9 ) << rateRatios[rateMat] [rate] << ' ';
        }
    }
    outputStream.unsetf(ios::fixed);
}

void MatrixModel::fromLine( const vector<double>& parameters ) {
    int i;
    
    assert( parameters.size() == getNumberLineParameters() );
    int index = -1;
    
    if ( invariantCategory ){
        proportionInvariantSites = parameters[++index];
    }
    if (discreteGamma){
        alpha = parameters[++index];
    }
    for (int freqCat = 0; freqCat < numberFrequenciesCategories; ++ freqCat ){
        double sum = 0.0;
        for ( i = 0; i < numberFrequencies-1; ++i ) {
            frequencies[freqCat] [i] = parameters[++index];
            sum += frequencies[freqCat][i];
        }
        frequencies[freqCat][numberFrequencies-1] = 1.0 - sum;
        if ( fabs(1.0 - sum - parameters[++index]) > 0.002 ){
            cerr << "WARNING, frequencies do not sum to 1 while "
                 << "reading parameters from a line" << endl;
        }
    }
    for ( int rateMat = 0 ; rateMat < numberRatesMatrices; ++rateMat ) { 
        for ( int rate = 0 ; rate < numberRatesRatiosPerMatrix; ++rate ) {
            rateRatios[rateMat] [rate]  = parameters[++index];
        }
    }
    setAverageMutationRate( averageMutationRate );
}

unsigned int MatrixModel::getNumberLineParameters(){
    return ( invariantCategory +
             (discreteGamma ? 1 : 0 ) +
             numberFrequenciesCategories * numberFrequencies +
             numberRatesMatrices * numberRatesRatiosPerMatrix );
}
