#pragma implementation
#include "Models/Mixed.h"

#include "Models/MixedPerturbator.h"
#include "Util/ModelFactory.h"

#include "PatternDesign/Singleton.h"

#include "Util/randombox.h"
#include "Util/ParametersSet.h"

Mixed Mixed::prototype( "MIXED" );

Mixed::Mixed( const string & registrationName ) {
    Singleton < ModelFactory > & modelFactory = Singleton < ModelFactory >::instance();
    modelFactory.subscribe( this, registrationName );
}


Mixed::Mixed( ParametersSet & parameters ) {
    Singleton < ModelFactory > & modelFactory = Singleton < ModelFactory >::instance();

    Model* tempModel;
    string tempModelName;
    
    char categoryName[25];
    model.clear();
    int numberModels = parameters.intParameter("Number of models");
    for ( int i = 0; i < numberModels; ++i ){
        sprintf( categoryName, "MODEL%d", i+1 );
        tempModelName = parameters( categoryName ).stringParameter( "Model" );
        tempModel = modelFactory.create( tempModelName,
                                         parameters( categoryName ) );
        model.push_back(tempModel);
    }
    perturbator = NULL;
}

Mixed::~Mixed() {
    //delete all the models
    for ( vector<Model*>::iterator iter = model.begin();
          iter != model.end(); ++iter ) {
        delete (*iter);
    }
}


/** ************************************************************************
 * getModelParameters
 * @semantics  create a ParametersSet and fill it with the parameters of
 *             the model
************************************************************************ */
ParametersSet Mixed::getModelParameters(){
    ParametersSet parameters("Mixed model parameters",false);
    
    parameters["Model name"] = getName();
    char param[50];
    char averageMutationRateRatio[20];
    for ( unsigned int i = 1; i < model.size(); ++i ){
        sprintf( averageMutationRateRatio, "%.5f",
                 model[i]->getAverageMutationRate(0));
        sprintf( param, "Model%d/model1 average mutation rate ratio" , i+1 );
        parameters[param] = averageMutationRateRatio;
    }
	
    char categoryName[15];
    for ( unsigned int i = 0; i < model.size(); ++i ){
        sprintf( categoryName, "MODEL%d", i+1);      
        parameters.addCategory("");
        ParametersSet& newParametersSet = parameters("");
        newParametersSet = model[i]->getModelParameters();
        newParametersSet.setName( categoryName );
    }
    return parameters;
}

  
void Mixed::setModelParameters( ParametersSet& parameters ){
    if (parameters.stringParameter("Model name")!=getName()){
        cerr << "You are trying to initialise the model " << getName() << endl;
        cerr << "with parameters for the model "
             << parameters.stringParameter("Model name") << endl;
        cerr << "Exit..." << endl;
        exit(1);
    }
    model[0]->setAverageMutationRate(1.0, 0);
    model[0]->setModelParameters(parameters("MODEL1"));
    char param[60];
    for ( unsigned int i = 1; i < model.size(); ++i ){
        sprintf(param, "Model%d/model1 average mutation rate ratio", i+1);
        model[i]->setAverageMutationRate( parameters.doubleParameter(param), 0 );
        sprintf(param, "MODEL%d",i+1);
        model[i]->setModelParameters(parameters(param));
    }
}


void Mixed::getAllParameters( vector < double > & parameters ) const {

    vector<double> partialParameters;

    parameters.clear();

    for ( unsigned int i = 1; i < model.size(); ++i ){
        parameters.push_back( sqrt( model[i]->getAverageMutationRate(0) ) );
    }

    for ( unsigned int i = 0; i < model.size(); ++i ){
        model[i]->getAllParameters( partialParameters );
        parameters.insert( parameters.end(), partialParameters.begin(),
                           partialParameters.end() );
    }
}

void Mixed::setAllParameters( const vector < double > & newParameters ) {

    int current = 0;
    
    vector <double> partialParameters;

    for ( unsigned int i = 1; i < model.size(); ++i ){
        model[i]->setAverageMutationRate( pow( newParameters[current], 2.0 ), 0 );
        ++current;
    }

    for( unsigned int i = 0; i < model.size(); ++i ){
        model[i]->getAllParameters( partialParameters );
        for (unsigned int j = 0; j < partialParameters.size(); ++j){
            partialParameters[j] = newParameters[current];
            ++current;
        }
        model[i]->setAllParameters(partialParameters);
    }
    
    assert( current = (int)newParameters.size() );
}

unsigned int Mixed::getNumberFreeParameters(){
    unsigned int ret = model[0]->getNumberFreeParameters();
    for ( unsigned int i = 1; i < model.size(); ++i ){
        ret += model[i]->getNumberFreeParameters();
    }
    return ret;
}

void Mixed::initialisePerturbator( ParametersSet & parameters ) {

    perturbator = new MixedPerturbator( model.size() );

    int averageRatesPrio = parameters.intParameter("Average rates, proposal priority");
    double minAverageRatesAcceptanceRate = .21;
    double maxAverageRatesAcceptanceRate = .25;    
    double averageRatesStep = .2;
    //dynamic adaptation, acceptance rate range provided...
    if ( parameters.find("Average rates, proposal minimum acceptance rate") !=
                parameters.end() ){
            minAverageRatesAcceptanceRate = parameters.doubleParameter
                ("Average rates, proposal minimum acceptance rate");
            maxAverageRatesAcceptanceRate = parameters.doubleParameter
                ("Average rates, proposal maximum acceptance rate");
            if (minAverageRatesAcceptanceRate < 0.0){
                cerr << "Proposal minimum acceptance rate lower than 0.0 for the average rates : "
                     << minAverageRatesAcceptanceRate << endl;
                exit(1);
            }
            if (maxAverageRatesAcceptanceRate > 1.0){
                cerr << "Proposal maximum acceptance rate greater than 1.0 for the average rates : "
                     << maxAverageRatesAcceptanceRate << endl;
                exit(1);
            }
            if ( (minAverageRatesAcceptanceRate>=maxAverageRatesAcceptanceRate) ){
                cerr << "Error with the proposal acceptance rates for the average rates" << endl;
                cerr << "minimum acceptance rate = " << minAverageRatesAcceptanceRate << endl;
                cerr << "maximum acceptance rate = " << maxAverageRatesAcceptanceRate << endl;
                cerr << minAverageRatesAcceptanceRate << " >= "
                     << maxAverageRatesAcceptanceRate << endl;
                exit(1);
            }            
    }
    // if the dynamical adaptation is turned off, the tuning parameter
    // is compulsory (otherwise it is optionnal)
    if ( (minAverageRatesAcceptanceRate == 0.0) && (maxAverageRatesAcceptanceRate==1.0) ){
        averageRatesStep = parameters.doubleParameter
                    ("Average rates, initial step");
    }
    else{
        if ( parameters.find("Average rates, initial step") !=
             parameters.end() ){
            averageRatesStep = parameters.doubleParameter
                    ("Average rates, initial step");
        }
    }
    if ( ( averageRatesStep <= 0.0 ) || ( averageRatesStep > 1.0 ) ){
        cerr << "Initial step out of range ] 0 ; 1.0 ]"
             << "for the average rates" << endl;
        exit(1);
    }
    if (averageRatesPrio != 0){
        char param[25];
        char label[50];
        sprintf( param, "PERTURBATION" );
        char* numberPos = param;
        while ( *(++numberPos) );

        for( unsigned int i = 0; i < model.size(); ++i ){
            sprintf( numberPos, "%d", i+1 );
            model[i]->initialisePerturbator( parameters( param ) );
            sprintf( label, "Model %d priority", i+1 );
            perturbator->registerModel( model[i], parameters.intParameter(label),
                                    averageRatesPrio, averageRatesStep,
                                    minAverageRatesAcceptanceRate, maxAverageRatesAcceptanceRate );
        }
    }
}

void Mixed::printPerturbationParameters( ostream & outputStream ) {
    perturbator->printParameters( outputStream );
}

double Mixed::perturb() {
    return perturbator->perturb();
}

void Mixed::stopBurn(){
    perturbator->stopBurn();
    for ( unsigned int i = 0; i < model.size(); ++i ){
        model[i]->stopBurn();
    }
}


bool Mixed::validatePerturbation( bool validation ) {
    return perturbator->validatePerturbation( validation );
}

void Mixed::initialisation( SequenceTable * sequenceTable, int modelId ) {
    assert( modelId == 0);
    for ( unsigned int i = 0; i < model.size(); ++i ){
        model[i]->initialisation( sequenceTable, i );
    }
}

void Mixed::printLine( ostream & outputStream ) {
    for (unsigned int i = 0; i < model.size(); ++i ){
        if ( i!= 0 ){
            outputStream << model[i]->getAverageMutationRate( 0 ) << ' ';
        }
        model[i]->printLine( outputStream );
        outputStream << ' ';
    }
}

void Mixed::fromLine( const vector<double>& parameters ){
    vector<double>::const_iterator iter = parameters.begin();
    for (unsigned int i = 0; i < model.size(); ++i ){
        if ( i!= 0 ){
            model[i]->setAverageMutationRate( *iter, 0 );
            ++iter;
        }
        model[i]->fromLine(
                vector<double>(iter,iter+model[i]->getNumberLineParameters()) );
       iter = iter + model[i]->getNumberLineParameters();
    }
}


unsigned int Mixed::getNumberLineParameters(){
    unsigned int ret = model.size()-1;
    for (unsigned int i = 0; i < model.size(); ++i ){
        ret += model[i]->getNumberLineParameters();
    }
    return ret;
}

