#include "Models/PerturbatorFrequenciesParameter.h"

#include <iostream>

#include "Util/statlib.h"

PerturbatorFrequenciesParameter::PerturbatorFrequenciesParameter(
      string name, vector<double>* pvalues, double tuning,
      double minAcceptance, double maxAcceptance ):
PerturbatorParameter(name){
    this->pvalues = pvalues;
    this->tuning = tuning;
    this->minAcceptance = minAcceptance;
    this->maxAcceptance = maxAcceptance;
    assert(minAcceptance >= 0.0);
    assert(maxAcceptance <= 1.0);
    assert(minAcceptance<maxAcceptance);
    if ( (minAcceptance == 0.0) && (maxAcceptance == 1.0) ){
        adaptativePerturbation = false;
    }
    tuningModifier = 1.6;
    lastModification = 0;
}


bool PerturbatorFrequenciesParameter::validatePerturbation( bool validation ){
    //if the perturbation is refused
    if (!validation){
        //restore the old value
        *pvalues = oldValues;
    }
    else{
        ++numberAcceptedPerturbation;
    }
    
    //if the dynamical step modification is on
    if( adaptativePerturbation ){
        if ( numberPerturbation == 200 ){
            double rate = (double)numberAcceptedPerturbation/
                          (double)numberPerturbation;
            //acceptance rate too high, reduce the dirichlet tuning parameter
            if ( rate > maxAcceptance ){
                if (lastModification == 1){
                    //reduce the tuning modifier if close to the good tuning
                    //(when it jumps around the targeted acceptance rate)
                    tuningModifier -= (tuningModifier-1.0)*0.3;
                }
                tuning = tuning / tuningModifier;
                lastModification = -1;
//                cout << name << " acceptance too high, new tuning " << tuning << endl;
            }
            else{
                //acceptance rate too low, increase the dirichlet tuning parameter
                if ( rate < minAcceptance ){
                    //reduce the tuning modifier if close to the good tuning
                    //(when it jumps around the targeted acceptance rate)
                    if (lastModification == -1){
                        tuningModifier -= (tuningModifier-1.0)*0.3;
                    }
                    tuning = tuning * tuningModifier;
                    lastModification = 1;
//                    cout << name << " acceptance too low, new step " << tuning << endl;
                }
            }
#ifdef DEBUG
            stepfile << tuning << " " << tuningModifier << endl;
#endif            
            numberAcceptedPerturbation = 0;
            numberPerturbation = 0;
        }
    }
    return true;
}


double PerturbatorFrequenciesParameter::perturb(){
    ++numberPerturbation;
    //store the old value in case the perturbation is not accepted
    oldValues = *pvalues;

    //modify the value   
    double numer = 0.0;
    double denom = 0.0;

    vector < double > dirichp1( oldValues.size() );
    vector < double > dirichp2( oldValues.size() );
    for ( unsigned int i = 0; i < oldValues.size(); ++i ) {
        dirichp1[i] = oldValues[i] * tuning;
    }
    *pvalues = rand.sdirichlet( dirichp1 );

    for ( unsigned int j = 0;
          j < (*pvalues).size(); ++j ){
        dirichp2[j] = (*pvalues)[j] * tuning;
    }
    numer = statlib::ldirichlet( oldValues, dirichp2 );
    denom = statlib::ldirichlet( *pvalues, dirichp1 );
    return ( numer - denom );
}

void PerturbatorFrequenciesParameter::printParameter( ostream& outputStream ){
    outputStream << name << ", acceptance rate : "
                 << (double)numberAcceptedPerturbation/(double)numberPerturbation
                 << ", final dirichlet tuning parameter : "<< tuning << endl;
}
