#include "Models/PerturbatorGaussParameter.h"

#include <iostream>

PerturbatorGaussParameter::PerturbatorGaussParameter(string name, double* pvalue,
   double step, double min, double max,
   double minAcceptance, double maxAcceptance ):
PerturbatorParameter(name){

    this->min = min;
    this->max = max;
    this->pvalue = pvalue;
    this->step = step;
    this->minAcceptance = minAcceptance;
    this->maxAcceptance = maxAcceptance;
    assert(minAcceptance >= 0.0);
    assert(maxAcceptance <= 1.0);
    assert(minAcceptance<maxAcceptance);
    if ( (minAcceptance == 0.0) && (maxAcceptance == 1.0) ){
        adaptativePerturbation = false;
    }
    assert(min<max);
    assert(step <= (max-min)/10);
    stepModifier = 1.6;
    lastModification = 0;
}

double PerturbatorGaussParameter::perturb(){
    ++numberPerturbation;
    //store the old value in case the perturbation is not accepted
    oldValue = *pvalue;
    //modify the value
    *pvalue += rand.gasdev() * step;
    //stay within boundaries
    double delta = max - min;
    if ( *pvalue <= (min-delta) ){
        int n = (int)( (max-*pvalue)/(2.0*delta) );
        *pvalue += 2.0 * delta * (double)n;
    }
    else{
        if ( *pvalue >= (max+delta) ){
            int n = (int)( (*pvalue-min)/(2.0*delta) );
            *pvalue -= 2.0 * delta * (double)n;
        }
    }
    assert( *pvalue > (min-delta) );
    assert( *pvalue < (max+delta) );
    //reflection
    if (*pvalue<min) *pvalue = 2.0*min-*pvalue;
    if (*pvalue>max) *pvalue = 2.0*max-*pvalue;
    assert( *pvalue >= min );
    assert( *pvalue <= max );
    return 0.0;
}
        
bool PerturbatorGaussParameter::validatePerturbation( bool validation ){
    //if the perturbation is refused
    if (!validation){
        //restore the old value
        *pvalue = oldValue;
    }
    else{
        ++numberAcceptedPerturbation;
    }
    
    //if the dynamical step modification is on
    if( adaptativePerturbation ){
        if ( numberPerturbation == 200 ){
            double rate = (double)numberAcceptedPerturbation/
                          (double)numberPerturbation;
            if ( rate > maxAcceptance ){
                if (lastModification == -1){
                    //reduce the tuning modifier if close to the good tuning
                    //(when it jumps around the targeted acceptance rate)
                     stepModifier -= (stepModifier-1.0)*0.3;
                }
                if ( step < (max-min) / 10 ){
                    step = step * stepModifier;
                    lastModification = 1;
//                    cout << name << " acceptance too high, new step " << step << endl;
                }
            }
            else{
                if ( rate < minAcceptance ){
                    if (lastModification == 1){
                        //reduce the tuning modifier if close to the good tuning
                        //(when it jumps around the targeted acceptance rate)
                        stepModifier -= (stepModifier-1.0)*0.3;
                    }
                    step = step / stepModifier;
                    lastModification = -1;
//                    cout << name << " acceptance too low, new step " << step << endl;
                }
            }
#ifdef DEBUG
            stepfile << step << " " << stepModifier << endl;
#endif
            numberAcceptedPerturbation = 0;
            numberPerturbation = 0;
        }
    }
    return true;
}

void PerturbatorGaussParameter::printParameter( ostream& outputStream ){
    outputStream << name << ", acceptance rate : "
                 << (double)numberAcceptedPerturbation/(double)numberPerturbation
                 << ", final step : "<< step << endl;
}


