#include <assert.h>

#include <iostream>
#include <iomanip>
#include <stdio.h>

#include "PatternDesign/Singleton.h"

#include "Util/ModelFactory.h"
#include "Models/Perturbator.h"
#include "Models/REV.h"

#include "Sequence/SequenceTable.h"

#include "Util/statlib.h"
#include "Util/randombox.h"
#include "Util/ParametersSet.h"


#define SQR(n) (n*n)

using namespace std;

//register the model to the model factory with the name REV
REV REV::prototype( "REV" );

REV::REV( const string & registrationName ) : DnaModel( registrationName ){
//the private constructor is called by the prototype only
}


REV::REV( ParametersSet & parameters ) : DnaModel( parameters ){
//the normal constructor is called by the prototype's clone method
}

void REV::initialisation( SequenceTable * sequenceTable, int modelId ) {
    //initialise the number of parameters
    numberRatesRatiosPerMatrix = 5;
    numberFrequencies = 4;
    //initialise the space for the parameters
    rateRatios = new vector < double > [numberRatesMatrices] (5);
    frequencies = new vector < double > [numberFrequenciesCategories] (4);
    
    //initialise the matrixIndex for correspondance between a matrix element
    //and an index in the rate ratios vector
    initMatrixIndex();
    
    //call the basic initialisation method
    DnaModel::initialisation( sequenceTable, modelId );
    
    //if a sequence table is provided initialise rate ratios and frequencies
    //accordingly
    if (sequenceTable){
        //estimation of the model frequencies from the empirical frequencies
        frequencies[0] = retrieveEmpiricalFrequencies( sequenceTable, modelId );
        for ( int i = 1 ; i < numberFrequenciesCategories; i++ ) {
            frequencies[i] = frequencies[0];
        }

        //estimation of all the rate ratios (and proportion of invariant sites
        //if relevant) from the empirical sequences
        if ( invariantCategory ) {
            rateRatios[0] = retrieveEmpiricalRates( sequenceTable, modelId,
            & proportionInvariantSites );
        }
        else {
            rateRatios[0] = retrieveEmpiricalRates( sequenceTable, modelId );
        }
        //if the discrete gamma model uses more than one set of rate ratios
        //initialise the other ones to the same values
        for ( int i = 1 ; i < numberRatesMatrices; i++ ) {
            rateRatios[i] = rateRatios[0];
        }
    }
    //init the equivalency table (correspondance symbol/states)
    initEquivalencyTable();
    //first initialisation of the mutation rate matrix
    //(factor(s) used with the subtitution matrix to reach the specified
    //average substitution rate) and eigen system
    setAverageMutationRate( averageMutationRate );
}  

REV::~REV() {
}

string REV::getName( void ) const {
    char num[3];
    string name( "REV" );
    if ( discreteGamma ) {
        sprintf( num, "%i", numberGammaCategories );
        name = name + string( " + dG" ) + string( num );
    }
    if ( invariantCategory == 1 )
        name = name + string( " + I" );
    return ( name );
}

void REV::initMatrixIndex() {
    matrixIndex.resize( 4, 4 );
    for ( int i = 0 ; i < 4 ; ++i ) {
        matrixIndex(i, i) = -3;
    }
    //reference
    matrixIndex(0, 2) = matrixIndex(2, 0) = -2;
    //transversion
    matrixIndex(0, 1) = matrixIndex(1, 0) = 0;
    matrixIndex(0, 3) = matrixIndex(3, 0) = 1;
    matrixIndex(1, 2) = matrixIndex(2, 1) = 2;
    matrixIndex(2, 3) = matrixIndex(3, 2) = 4;
    //second transition
    matrixIndex(1, 3) = matrixIndex(3, 1) = 3;
}

void REV::setEigenMatrix() {
    vector< double > junk( 4 );
    for ( int category = 0; category < MAX( numberGammaCategories, 1 );
    category++ ) {
        // fill the rateMatrix (cf REV.h)
        int freqCat = frequencyCategory[category+invariantCategory];
        int rateCat = matrixCategory[category];
        rateMatrix[category] ( 0, 1 ) =
        mutationRate[category] * frequencies[freqCat] [1] *
        rateRatios[rateCat] [0]; // A->C
        rateMatrix[category] ( 0, 2 ) =
        mutationRate[category] * frequencies[freqCat] [2];
        rateMatrix[category] ( 0, 3 ) =
        mutationRate[category] * frequencies[freqCat] [3] *
        rateRatios[rateCat] [1]; // A->U
        rateMatrix[category] ( 0, 0 ) =
        -( ( rateMatrix[category] ) ( 0, 1 ) +
        ( rateMatrix[category] ) ( 0, 2 ) +
        ( rateMatrix[category] ) ( 0, 3 ) );
        rateMatrix[category] ( 1, 0 ) =
        mutationRate[category] * frequencies[freqCat] [0] *
        rateRatios[rateCat] [0]; // C->A
        rateMatrix[category] ( 1, 2 ) =
        mutationRate[category] * frequencies[freqCat] [2] *
        rateRatios[rateCat] [2]; // C->G
        rateMatrix[category] ( 1, 3 ) =
        mutationRate[category] * frequencies[freqCat] [3] *
        rateRatios[rateCat] [3]; // C->U
        rateMatrix[category] ( 1, 1 ) =
        -( ( rateMatrix[category] ) ( 1, 0 ) +
        ( rateMatrix[category] ) ( 1, 2 ) +
        ( rateMatrix[category] ) ( 1, 3 ) );
        rateMatrix[category] ( 2, 0 ) =
        mutationRate[category] * frequencies[freqCat] [0]; // G->A
        rateMatrix[category] ( 2, 1 ) =
        mutationRate[category] * frequencies[freqCat] [1] *
        rateRatios[rateCat] [2]; // G->C
        rateMatrix[category] ( 2, 3 ) =
        mutationRate[category] * frequencies[freqCat] [3] *
        rateRatios[rateCat] [4]; // G->U
        rateMatrix[category] ( 2, 2 ) =
        -( ( rateMatrix[category] ) ( 2, 0 ) +
        ( rateMatrix[category] ) ( 2, 1 ) +
        ( rateMatrix[category] ) ( 2, 3 ) );
        rateMatrix[category] ( 3, 0 ) =
        mutationRate[category] * frequencies[freqCat] [0] *
        rateRatios[rateCat] [1]; // U->A
        rateMatrix[category] ( 3, 1 ) =
        mutationRate[category] * frequencies[freqCat] [1] *
        rateRatios[rateCat] [3]; // U->C
        rateMatrix[category] ( 3, 2 ) =
        mutationRate[category] * frequencies[freqCat] [2] *
        rateRatios[rateCat] [4]; // U->G
        rateMatrix[category] ( 3, 3 ) =
        -( ( rateMatrix[category] ) ( 3, 0 ) +
        ( rateMatrix[category] ) ( 3, 1 ) +
        ( rateMatrix[category] ) ( 3, 2 ) );

        for ( int i = 0; i < 4; ++i ) {
            eigenValues[category] [i] = 0.0;
            junk[i] = 0.0;
            for ( int j = 0; j < 4; ++j ) {
                eigenMatrix[category] ( i, j ) = 0.0;
                ieigenMatrix[category] ( i, j ) = 0.0;
            }
        }
        // Find the eigenValues and eigenVectors of the rate matrix
        LeftEigenSystem( rateMatrix[category], eigenMatrix[category],
        eigenValues[category], junk );
        // Find the inverse of the matrix whose rows are the left
        // eigen vectors of the rate matrix
        ieigenMatrix[category] = inverse( eigenMatrix[category] );
    } //for each category (except invariant)
}

