#pragma implementation

#include <assert.h>

#include <iostream>
#include <iomanip>


#include "Models/RNA16.h"

#include "statlib.h"

#define SQR(n) (n*n)
#define SIGN(n) (n>=0.0 ? 1.0 : -1.0)


using namespace std;

//register the model to the model factory with the name RNA16
RNA16 RNA16::prototype( "RNA16" );

RNA16::RNA16( const string & registrationName ) :
RnaModel( registrationName ) {
//the private constructor is called by the prototype only
}

RNA16::RNA16( ParametersSet & parameters )
: RnaModel( parameters, 16 ) {
//the normal constructor is called by the prototype's clone method
//RNA16 is a RnaModel, its matrix size is 16*16
}


void RNA16::initialisation( SequenceTable * sequenceTable, int modelId ) {
    //initialise the number of parameters
    numberRatesRatiosPerMatrix = 119;
    numberFrequencies = 16;
    //initialise the space for the parameters
    rateRatios = new vector < double > [numberRatesMatrices] ( 119 );
    frequencies = new vector < double > [numberFrequenciesCategories] ( 16 );

    //initialise the matrixIndex for correspondance between a matrix element
    //and an index in the rate ratios vector
    initMatrixIndex();

    //call the basic initialisation method
    RnaModel::initialisation( sequenceTable, modelId );
    
    //if a sequence table is provided initialise rate ratios and frequencies
    //accordingly
    if ( sequenceTable ) {
        vector < double > freq16;
        array2D < double > rate16;

        //frequencies[0] is equal to the emprical frequencies of the 16
        //possible symbols
        frequencies[0] = retrieveEmpiricalFrequencies( sequenceTable, modelId );
        //if the discrete gamma model uses more than one set of frequencies
        //initialise the other ones to the same values
        for ( int i = 1; i < numberFrequenciesCategories; ++i ) {
            frequencies[i] = frequencies[0];
        }

        
        //estimation of all the rate ratios (and proportion of invariant sites
        //if relevant) from the empirical sequences
        if ( invariantCategory ) {
            rate16 = retrieveEmpiricalRates( sequenceTable, modelId, & proportionInvariantSites );
        }
        else {
            rate16 = retrieveEmpiricalRates( sequenceTable, modelId );
        }
        //fill the rate vector according to the estimation
        for ( int i = 0; i < 15 ; ++i ) {
            for ( int j = i + 1; j < 16 ; ++j ) {
                int index = matrixIndex(i, j);
                //if the rate is not the reference rate (0, 2)
                if ( index >= 0 ) {
                    rateRatios[0] [index] = rate16( i, j );
                }
            }
        }
        //if the discrete gamma model uses more than one set of rate ratios
        //initialise the other ones to the same values
        for ( int i = 1; i < numberRatesMatrices; ++i ) {
            rateRatios[i] = rateRatios[0];
        }
    }
    //init the equivalency table (correspondance symbol/states)
    initEquivalencyTable();
    //first initialisation of the mutation rate matrix
    //(factor(s) used with the subtitution matrix to reach the specified
    //average substitution rate) and eigen system
    setAverageMutationRate( averageMutationRate );
}


// Destructor
RNA16::~RNA16() {
}


// The models unique name
string RNA16::getName( void ) const {
    string name;
    char num[3];

    name = string( "RNA16" );
    if ( discreteGamma ) {
        sprintf( num, "%i", numberGammaCategories );
        name = name + string( " + dG" ) + string( num );
    }
    if ( invariantCategory == 1 ) {
        name = name + string( " + I" );
    }
    return ( name );
}




void RNA16::initMatrixIndex() {
    matrixIndex.resize( 16, 16 );
    int index = -1;
    for ( int i = 0 ; i < 15 ; ++i ) {
        for ( int j = i + 1 ; j < 16 ; ++j ) {
            if ( ( i == 0 ) && ( j == 2 ) ) {
                //the reference ratio is given the negative index of -2
                matrixIndex(i, j) = matrixIndex(j, i) = -2;
            }
            else {
                ++index;
                matrixIndex(i, j) = matrixIndex(j, i) = index;
            }
        }
    }
    assert( index == 118 );
    //initialise the diagonal
    for ( int i = 0 ; i < 16 ; ++i ) {
        matrixIndex(i, i) = -3;
    }
}


void RNA16::initEquivalencyTable() {
    int i, j;
    
    equivalencyTable.resize( 64, 16 );

    // Clear the table.
    for ( i = 0; i < 64; ++i ) {
        for ( j = 0; j < 16; ++j ) {
            equivalencyTable( i, j ) = 0.0;
        }
    }

    //valid states
    for ( i = 0; i < 16; ++i ) {
        equivalencyTable( i, i ) = 1.0;
    }
    
    equivalencyTable( 16, 0 ) = equivalencyTable( 16, 6 ) =
    equivalencyTable( 16, 7 ) = equivalencyTable( 16, 8 ) = 1.0; // AX
    equivalencyTable( 17, 6 ) = equivalencyTable( 17, 7 ) = 1.0; // AR (AG or AA)
    equivalencyTable( 18, 0 ) = equivalencyTable( 18, 8 ) = 1.0; // AY (AU or AC)

    equivalencyTable( 19, 5 ) = equivalencyTable( 19, 11 ) =
    equivalencyTable( 19, 12 ) = equivalencyTable( 19, 13 ) = 1.0; // CX
    equivalencyTable( 20, 5 ) = equivalencyTable( 20, 11 ) = 1.0; // CR (CG or CA)
    equivalencyTable( 21, 12 ) = equivalencyTable( 21, 13 ) = 1.0; // CY (CC or CU)

    equivalencyTable( 22, 1 ) = equivalencyTable( 22, 2 ) =
    equivalencyTable( 22, 9 ) = equivalencyTable( 22, 10 ) = 1.0; // GX
    equivalencyTable( 23, 9 ) = equivalencyTable( 23, 10 ) = 1.0; // GR (GA or GG)
    equivalencyTable( 24, 1 ) = equivalencyTable( 24, 2 ) = 1.0; // GY (GC or GU)

    equivalencyTable( 25, 3 ) = equivalencyTable( 25, 4 ) =
    equivalencyTable( 25, 14 ) = equivalencyTable( 25, 15 ) = 1.0; // UX
    equivalencyTable( 26, 3 ) = equivalencyTable( 26, 4 ) = 1.0; // UR (UA or UG)
    equivalencyTable( 27, 14 ) = equivalencyTable( 27, 15 ) = 1.0; // UY (UC or UU)

    equivalencyTable( 28, 3 ) = equivalencyTable( 28, 6 ) =
    equivalencyTable( 28, 9 ) = equivalencyTable( 28, 11 ) = 1.0; // XA
    equivalencyTable( 29, 9 ) = equivalencyTable( 29, 10 ) = 1.0; // RA (GA or GG)
    equivalencyTable( 30, 3 ) = equivalencyTable( 30, 11 ) = 1.0; // YA (CA or UA)

    equivalencyTable( 31, 2 ) = equivalencyTable( 31, 8 ) =
    equivalencyTable( 31, 12 ) = equivalencyTable( 31, 14 ) = 1.0; // XC
    equivalencyTable( 32, 2 ) = equivalencyTable( 32, 8 ) = 1.0; // RC (GC or AC)
    equivalencyTable( 33, 12 ) = equivalencyTable( 33, 14 ) = 1.0; // YC (CC or UC)

    equivalencyTable( 34, 4 ) = equivalencyTable( 34, 5 ) =
    equivalencyTable( 34, 7 ) = equivalencyTable( 34, 10 ) = 1.0; // XG
    equivalencyTable( 35, 7 ) = equivalencyTable( 35, 10 ) = 1.0; // RG (AG or GG)
    equivalencyTable( 36, 4 ) = equivalencyTable( 36, 5 ) = 1.0; // YG (CG or UG)

    equivalencyTable( 37, 0 ) = equivalencyTable( 37, 1 ) =
    equivalencyTable( 37, 13 ) = equivalencyTable( 37, 15 ) = 1.0; // XU
    equivalencyTable( 38, 0 ) = equivalencyTable( 38, 1 ) = 1.0; // RU (AU or GU)
    equivalencyTable( 39, 13 ) = equivalencyTable( 39, 15 ) = 1.0; // YU (CU or UU)


    equivalencyTable( 41, 3 ) = equivalencyTable( 41, 4 ) =
    equivalencyTable( 41, 5 ) = equivalencyTable( 41, 6 ) =
    equivalencyTable( 41, 7 ) = equivalencyTable( 41, 9 ) =
    equivalencyTable( 41, 10 ) = equivalencyTable( 41, 11 ) = 1.0; //XR (XA,XG)

    equivalencyTable( 42, 0 ) = equivalencyTable( 42, 1 ) =
    equivalencyTable( 42, 2 ) = equivalencyTable( 42, 8 ) =
    equivalencyTable( 42, 12 ) = equivalencyTable( 42, 13 ) =
    equivalencyTable( 42, 14 ) = equivalencyTable( 42, 15 ) = 1.0; //XY (XC,XU)

    equivalencyTable( 43, 3 ) = equivalencyTable( 43, 4 ) =
    equivalencyTable( 43, 5 ) = equivalencyTable( 43, 11 ) =
    equivalencyTable( 43, 12 ) = equivalencyTable( 43, 13 ) =
    equivalencyTable( 43, 14 ) = equivalencyTable( 43, 15 ) = 1.0; //YX (CX,UX)

    equivalencyTable( 44, 3 ) = equivalencyTable( 44, 4 ) =
    equivalencyTable( 44, 5 ) = equivalencyTable( 44, 11 ) = 1.0; //YR (CA,CG,UA,UG)

    equivalencyTable( 45, 2 ) = equivalencyTable( 45, 5 ) =
    equivalencyTable( 45, 10 ) = equivalencyTable( 45, 12 ) = 1.0; //YY (CC,CG,GC,GG)

    equivalencyTable( 46, 0 ) = equivalencyTable( 46, 1 ) =
    equivalencyTable( 46, 2 ) = equivalencyTable( 46, 6 ) =
    equivalencyTable( 46, 7 ) = equivalencyTable( 46, 8 ) =
    equivalencyTable( 46, 9 ) = equivalencyTable( 46, 10 ) = 1.0; //RX (AX,GX)

    equivalencyTable( 47, 6 ) = equivalencyTable( 47, 7 ) =
    equivalencyTable( 47, 9 ) = equivalencyTable( 47, 10 ) = 1.0; //RR (AA,AG,GA,GG)

    equivalencyTable( 48, 0 ) = equivalencyTable( 48, 1 ) =
    equivalencyTable( 48, 2 ) = equivalencyTable( 48, 8 ) = 1.0; //RY (AC,AU,GC,GU)

    equivalencyTable( 49, 6 ) = equivalencyTable( 49, 9 ) =
    equivalencyTable( 49, 11 ) = equivalencyTable( 49, 13 ) = 1.0; //-A

    equivalencyTable( 50, 7 ) = equivalencyTable( 50, 10 ) =
    equivalencyTable( 50, 5 ) = equivalencyTable( 50, 4 ) = 1.0; //-G

    equivalencyTable( 51, 8 ) = equivalencyTable( 51, 2 ) =
    equivalencyTable( 51, 12 ) = equivalencyTable( 51, 14 ) = 1.0; //-C

    equivalencyTable( 52, 0 ) = equivalencyTable( 52, 1 ) =
    equivalencyTable( 52, 13 ) = equivalencyTable( 52, 15 ) = 1.0; //-U


    equivalencyTable( 54, 3 ) = equivalencyTable( 54, 4 ) =
    equivalencyTable( 54, 5 ) = equivalencyTable( 54, 6 ) =
    equivalencyTable( 54, 7 ) = equivalencyTable( 54, 9 ) =
    equivalencyTable( 54, 10 ) = equivalencyTable( 54, 11 ) = 1.0; //-R (-A,-G)

    equivalencyTable( 55, 0 ) = equivalencyTable( 55, 1 ) =
    equivalencyTable( 55, 2 ) = equivalencyTable( 55, 8 ) =
    equivalencyTable( 55, 12 ) = equivalencyTable( 55, 13 ) =
    equivalencyTable( 55, 14 ) = equivalencyTable( 55, 15 ) = 1.0; //-Y (-C,-U)

    equivalencyTable( 56, 0 ) = equivalencyTable( 56, 6 ) =
    equivalencyTable( 56, 7 ) = equivalencyTable( 56, 8 ) = 1.0; //A-

    equivalencyTable( 57, 1 ) = equivalencyTable( 57, 2 ) =
    equivalencyTable( 57, 9 ) = equivalencyTable( 57, 10 ) = 1.0; //G-

    equivalencyTable( 58, 5 ) = equivalencyTable( 58, 11 ) =
    equivalencyTable( 58, 12 ) = equivalencyTable( 58, 13 ) = 1.0; //C-

    equivalencyTable( 59, 3 ) = equivalencyTable( 59, 4 ) =
    equivalencyTable( 59, 14 ) = equivalencyTable( 59, 15 ) = 1.0; //U-

    equivalencyTable( 61, 0 ) = equivalencyTable( 61, 1 ) =
    equivalencyTable( 61, 2 ) = equivalencyTable( 61, 6 ) =
    equivalencyTable( 61, 7 ) = equivalencyTable( 61, 8 ) =
    equivalencyTable( 61, 9 ) = equivalencyTable( 61, 10 ) = 1.0; //R- (A-,G-)

    equivalencyTable( 62, 3 ) = equivalencyTable( 62, 4 ) =
    equivalencyTable( 62, 5 ) = equivalencyTable( 62, 11 ) =
    equivalencyTable( 62, 12 ) = equivalencyTable( 62, 13 ) =
    equivalencyTable( 62, 14 ) = equivalencyTable( 62, 15 ) = 1.0; //Y- (C-,U-)

    for ( i = 0; i < 16; ++i ) {
        equivalencyTable( 40, i ) = 1.0; //XX
        equivalencyTable( 53, i ) = 1.0; //-X
        equivalencyTable( 60, i ) = 1.0; //X-
        equivalencyTable( 63, i ) = 1.0; //--
    }
}


// print the parameters in readable format
void RNA16::printParameters( ostream & outputStream ) const {
  
    //print common parameters
    MatrixModel::printParameters( outputStream );

    outputStream.setf(ios::fixed);
    outputStream << setprecision(5);
    for ( int freqCat = 0; freqCat < numberFrequenciesCategories; ++freqCat ) {
        if (numberFrequenciesCategories>=2){
            outputStream << "set of frequencies " << freqCat + 1 << ":" << endl;
        }
        else{
            outputStream << "frequencies:" << endl;
        }
        for ( int i = 0; i < getNumberStates(); ++i ) {
            outputStream << "f[" << getState( i ) << "] = " <<
            frequencies[freqCat] [i] << "  ";
        }
    }
    outputStream << endl;

    outputStream << setprecision(4);
    for ( int rateMat = 0 ; rateMat < numberRatesMatrices; ++rateMat ) {
        if (numberRatesMatrices>=2){
            outputStream << "set of rate ratios " << rateMat + 1 << ":" << endl;
        }
        else{
            outputStream << "rate ratios:" << endl;
        }
        outputStream << "    ";
        for ( int i = 0; i < 16; ++i ) {
            outputStream << "    " << getSymbol( i ) << "   ";
        }
        outputStream << endl;
        for ( int i = 0; i < 16; ++i ) {
            outputStream << ' ' << getSymbol( i ) << ' ';
            for ( int j = 0; j <16; ++j ) {
                int index = matrixIndex( i, j);
                switch ( index ) {
                    //if diagonal
                    case -3 : outputStream << "  ------ ";
                              break;
                    //if reference
                    case -2 : outputStream << "  1.0000 ";
                              break;
                    //if null
                    case -1 : outputStream << "  XXXXXX ";
                              break;
                    //otherwise
                    default : outputStream << setw( 8 ) <<
                              rateRatios [rateMat] [index] << ' ';
                              break;
                }
            }
            outputStream << endl; //end line
        }
        outputStream << endl; //end rate ratios matrix
    }
    
    if ( (numberRatesMatrices>=2) || (numberFrequenciesCategories>=2) ){
        for ( int gammaCat = 0; gammaCat < MAX(numberGammaCategories,1); ++ gammaCat ) {
            if ( numberGammaCategories>1 ){
                outputStream << "transition rate matrix for the gamma category "
	                         << gammaCat + 1;
                if (numberFrequenciesCategories>=2){
                    outputStream << ", frequencies set " <<
                         frequencyCategory[gammaCat + invariantCategory] + 1;
                }
                if (numberRatesMatrices>=2){
                    outputStream << ", rate ratios set "
                                 << matrixCategory[gammaCat] + 1;
                }
                outputStream << ":" << endl;
            }
            else{
                outputStream << "transition rate matrix";	
                if (numberFrequenciesCategories>=2){
                    outputStream << ", frequencies set " <<
                         frequencyCategory[invariantCategory] + 1;
                }
                outputStream << ":" << endl;
            }
            outputStream << "     ";
            for ( int i = 0; i < getNumberStates(); ++i ) {
                outputStream << "    " << getState( i ) << "    ";
            }
            outputStream << endl;
            for ( int i = 0; i < getNumberStates(); ++i ) {
                outputStream << ' ' << getState( i ) << ' ';
                for ( int j = 0; j < getNumberStates(); ++j ) {
                    outputStream << setw( 9 ) << rateMatrix[gammaCat] (i, j) << ' ';
                }
                outputStream << endl; //end line
            }
            outputStream << endl; //end transition rate matrix
        }
        outputStream << endl;
    }
    else{
        outputStream << "transition rate matrix (scaled so that "
                     << "the average substitution rate is 1.0):" << endl;
        double sum = 0.0;
        for ( int i = 0; i < getNumberStates() - 1; ++i ) {
            for ( int j = i + 1; j < getNumberStates(); ++j ) {
                int index = matrixIndex(i, j);
                switch ( index ) {
                    //if reference
                    case -2 : sum += ( 2.0 * frequencies[0] [i] *
                                       frequencies[0] [j] );
                              break;
                    //if null
                    case -1 : case -3: break;
                    //otherwise (>=0)
                    default : sum += ( 2.0 * frequencies[0] [i] *
                                       frequencies[0] [j] *
                                       rateRatios[0] [index] );
                }
            }
        }
        outputStream << "     ";
        for ( int i = 0; i < getNumberStates(); ++i ) {
            outputStream << "    " << getState( i ) << "    ";
        }
        outputStream << endl;
        for ( int i = 0; i < getNumberStates(); ++i ) {
            outputStream << ' ' << getState( i ) << ' ';
            for ( int j = 0; j < getNumberStates(); ++j ) {
                int index = matrixIndex(i, j);
                switch ( index ) {
                    //if diagonal
                    case -3 : outputStream << "   ------ ";
                              break;
                    //if reference
                    case -2 : outputStream <<  setw( 9 )
                                           << frequencies[0] [j] / sum << ' ';
                              break;
                    //if null
                    case -1 : outputStream << setw( 9 ) << 0.0  << ' ';
                              break;
                    //otherwise (>=0)
                    default : outputStream << setw( 9 )
                            << rateRatios[0] [index] * frequencies[0] [j] / sum
                            << ' ';
                }
            }
            outputStream << endl; //end line
        }
        outputStream << endl; //end transition rate matrix
    }
    outputStream << endl;
    outputStream.unsetf(ios::fixed);
}  

   
string RNA16::getState( int stateNumber, int symbolCategory ) const {
    assert( symbolCategory == 0 );
    switch ( stateNumber ) {
        // Watson-Crick base pairs
        case 0 :
            return ( string( "AU" ) );
        case 1 :
            return ( string( "GU" ) );
        case 2 :
            return ( string( "GC" ) );
        case 3 :
            return ( string( "UA" ) );
        case 4 :
            return ( string( "UG" ) );
        case 5 :
            return ( string( "CG" ) );
            // Mismatches
        case 6 :
            return ( string( "AA" ) );
        case 7 :
            return ( string( "AG" ) );
        case 8 :
            return ( string( "AC" ) );
        case 9 :
            return ( string( "GA" ) );
        case 10 :
            return ( string( "GG" ) );
        case 11 :
            return ( string( "CA" ) );
        case 12 :
            return ( string( "CC" ) );
        case 13 :
            return ( string( "CU" ) );
        case 14 :
            return ( string( "UC" ) );
        case 15 :
            return ( string( "UU" ) );
        default :
            return("");
    }
}      


void RNA16::setEigenMatrix() {
    for ( int category = 0; category < MAX( numberGammaCategories, 1 );
          ++category ) {
        int freqCat = frequencyCategory[category + invariantCategory];
        int matrixCat = matrixCategory[category];

        // Initialise the rate matrix
        for ( int i = 0; i < 15; ++i ) {
            for ( int j = i + 1; j < 16; ++j ) {
                int index = matrixIndex(i, j);
                switch (index){
                    //if reference rate
                    case -2 : rateMatrix[category] ( i, j ) =
                           frequencies[freqCat] [j] * mutationRate[category];
                              rateMatrix[category] ( j, i ) =
                           frequencies[freqCat] [i] * mutationRate[category];
                              break;
                    //if null rate
                    case -1 : rateMatrix[category] ( i, j ) = 0.0;
                              rateMatrix[category] ( j, i ) = 0.0;
                              break;
                    //otherwise
                    default : rateMatrix[category] ( i, j ) =
                                  rateRatios[matrixCat] [index] *
                             frequencies[freqCat] [j] * mutationRate[category];
                              rateMatrix[category] ( j, i ) =
                                  rateRatios[matrixCat] [index] *
                             frequencies[freqCat] [i] * mutationRate[category];
#ifdef DEBUG
                             assert( !isnan(rateRatios[matrixCat][index]) );
#endif
                             break;
                }
            }
        }
        for ( int i = 0; i < 16; ++i ) {
            double sum = 0.0;
            for ( int j = 0; j < 16; ++j ) {
                if ( i != j ){
                    sum += rateMatrix[category] ( i, j );
                }
            }
            rateMatrix[category] ( i, i ) = -sum;
        }

        vector < double > junk( 16 );
        for ( int i = 0; i < 16; ++i ) {
            eigenValues[category] [i] = 0.0;
            junk[i] = 0.0;
            for ( int j = 0; j < 16; ++j ){
                eigenMatrix[category] ( i, j ) = 0.0;
                ieigenMatrix[category] ( i, j ) = 0.0;
            }
        }
        // Find the eigenvalues and eigenvectors of the rate matrix
        LeftEigenSystem( rateMatrix[category], eigenMatrix[category],
        eigenValues[category], junk );
        // Find the inverse of the matrix whose rows are the left eigen
        // vectors of the rate matrix
        ieigenMatrix[category] = inverse( eigenMatrix[category] );
    }
}




