#include <assert.h>

#include <iostream>
#include <iomanip>


#include "Models/RNA16A.h"

#include "statlib.h"

#define SQR(n) (n*n)
#define SIGN(n) (n>=0.0 ? 1.0 : -1.0)

using namespace std;

//register the model to the model factory with the name RNA16A
RNA16A RNA16A::prototype( "RNA16A" );

RNA16A::RNA16A( const string & registrationName ) :
RNA16( registrationName ) {
//the private constructor is called by the prototype only
//RNA16A is a restriction of RNA16
}

RNA16A::RNA16A( ParametersSet & parameters )
: RNA16( parameters ) {
//the normal constructor is called by the prototype's clone method
//RNA16A is a restriction of RNA16
}


void RNA16A::initialisation( SequenceTable * sequenceTable, int modelId ) {
    //The initialisation procedure for the 16A model is slightly different
    //than for the 16 model, therefore we redefine it
    //initialise the number of parameters
    numberRatesRatiosPerMatrix = 4;
    numberFrequencies = 16;
    //initialise the space for the parameters
    rateRatios = new vector < double > [numberRatesMatrices] ( 5 );
    frequencies = new vector < double > [numberFrequenciesCategories] ( 16 );

    //initialise the matrixIndex for correspondance between a matrix element
    //and an index in the rate ratios vector
    initMatrixIndex();

    //call the basic initialisation method
    RnaModel::initialisation( sequenceTable, modelId );
    
    //if a sequence table is provided initialise rate ratios and frequencies
    //accordingly
    if ( sequenceTable ) {
        vector < double > freq16;
        array2D < double > rate16;

        //frequencies[0] is equal to the emprical frequencies of the 16
        //possible symbols
        frequencies[0] = retrieveEmpiricalFrequencies( sequenceTable, modelId );
        //if the discrete gamma model uses more than one set of frequencies
        //initialise the other ones to the same values
        for ( int i = 1; i < numberFrequenciesCategories; ++i ) {
            frequencies[i] = frequencies[0];
        }

        
        //estimation of all the rate ratios (and proportion of invariant sites
        //if relevant from the empirical sequences
        if ( invariantCategory ) {
            rate16 = retrieveEmpiricalRates( sequenceTable, modelId, & proportionInvariantSites );
        }
        else {
            rate16 = retrieveEmpiricalRates( sequenceTable, modelId );
        }
        //fill the rate vector according to the estimation
        //reference ratio
        double ref = ( rate16( 0, 2 ) + rate16( 3, 5 ) ) / 2;
        // 0 - single transition rate ratio
        double singleRef = ( rate16( 0, 1 ) + rate16( 1, 2 ) + rate16( 3, 4 ) +
        rate16( 4, 5 ) );
        rateRatios[0][0]= singleRef / (4*ref);
        // 1 - double transversion rate ratio
        double doubleRef = 0.0;
        for ( int i = 0; i < 3; ++i ) {
            for ( int j = 3; j < 6; ++j ) {
                doubleRef += rate16( i, j );
            }
        }
        rateRatios[0] [1] = doubleRef/(9*ref);
        // 2 - pair/mismatch rate ratio
        double pairMismatchRef = 0.0;
        for ( int i = 0; i < 6; ++i ) {
            for ( int j = 6; j < 16; ++j ) {
                pairMismatchRef += rate16( i, j );
            }
        }
        rateRatios[0][2] = pairMismatchRef/(60*ref);
        // 3 - mismatch/mismatch rate ratio
        // to be reviewed because of the null rate
        double mismatchMismatchRef = 0.0;
        for ( int i = 6; i < 16; ++i ) {
            for ( int j = 6; j < 16; ++j ) {
                mismatchMismatchRef += rate16( i, j );
            }
        }
        rateRatios[0][3] = mismatchMismatchRef/(60*ref);
        //if the discrete gamma model uses more than one set of rate ratios
        //initialise the other ones to the same values
        for ( int i = 1; i < numberRatesMatrices; i++ ) {
            rateRatios[i] = rateRatios[0];
        }
    }
    //init the equivalency table (correspondance symbol/states)
    initEquivalencyTable();
    //first initialisation of the mutation rate matrix
    //(factor(s) used with the subtitution matrix to reach the specified
    //average substitution rate) and eigen system
    setAverageMutationRate( averageMutationRate );
}


// Destructor
RNA16A::~RNA16A() {
}


// The models unique name
string RNA16A::getName( void ) const {
    string name;
    char num[3];

    name = string( "RNA16A" );
    if ( discreteGamma ) {
        sprintf( num, "%i", numberGammaCategories );
        name = name + string( " + dG" ) + string( num );
    }
    if ( invariantCategory == 1 ) {
        name = name + string( " + I" );
    }
    return ( name );
}



void RNA16A::initMatrixIndex() {
    matrixIndex.resize( 16, 16 );  
    //the reference ratio is given the negative index of -2
    matrixIndex( 0, 2 ) = matrixIndex(3, 5) = -2;
    // 0 - ingle transition rate ratio
    matrixIndex(0, 1) = matrixIndex(1, 2) =
        matrixIndex(3, 4) = matrixIndex(4, 5) = 0;
    // 1 - double transversion rate ratio
    for ( int i = 0; i < 3; ++i ) {
        for ( int j = 3; j < 6; ++j ) {
            matrixIndex(i, j) = 1;
        }
    }

    string pair1;
    string pair2;
    // 3 - pair/mismatch rate ratio
    for ( int i = 0; i < 6; ++i ) {
        pair1 = getSymbol(i);
        for ( int j = 6; j < 16; ++j ) {
            pair2 = getSymbol(j);
            assert(pair1.size() == 2);
            assert(pair2.size() == 2);
            //if one nucleotide is the same rate = pair/mismatch rate ratio
            if ( (pair1[0]==pair2[0]) || (pair1[1]==pair2[1]) ){
                matrixIndex(i, j) = 2;
            }
            // else null rate ratio -> index = -1
            else{
                matrixIndex(i, j) = -1;
            }
        }
    }
    // 4 - mismatch/mismatch rate ratio
    for ( int i = 6; i < 15; ++i ) {
        pair1 = getSymbol(i);
        for ( int j = i + 1; j < 16; ++j ) {
            pair2 = getSymbol(j);
            assert(pair1.size() == 2);
            assert(pair2.size() == 2);
            //if one nucleotide is the same rate = pair/mismatch rate ratio
            if ( (pair1[0]==pair2[0]) || (pair1[1]==pair2[1]) ){
                matrixIndex(i, j) = 3;
            }
            // else null rate ratio, it is given the index of -1
            else{
                matrixIndex(i, j) = -1;
            }
        }
    }
    //copy on the other triangle of the matrix
    for ( int i = 1; i < 16; ++i ) {
        for ( int j = 0; j < i; ++j ) {
            matrixIndex(i, j) = matrixIndex(j, i);
        }
    }
    //initialise the diagonal
    for ( int i = 0 ; i < 16 ; ++i ) {
        matrixIndex(i, i) = -3;
    }
}
