#include <assert.h>

#include <iostream>
#include <iomanip>

#include "Models/RNA6A.h"

#include "statlib.h"

#define SQR(n) (n*n)
#define SIGN(n) (n>=0.0 ? 1.0 : -1.0)


using namespace std;

//register the model to the model factory with the name RNA6A
RNA6A RNA6A::prototype( "RNA6A" );

RNA6A::RNA6A( const string & registrationName ) :
RnaModel( registrationName ) {
//the private constructor is called by the prototype only
}

RNA6A::RNA6A( ParametersSet & parameters )
: RnaModel( parameters, 6 ) {
//the normal constructor is called by the prototype's clone method
//RNA6A is a RnaModel, its matrix size is 6*6
}


void RNA6A::initialisation( SequenceTable * sequenceTable, int modelId ) {
    //initialise the number of parameters
    numberRatesRatiosPerMatrix = 14;
    numberFrequencies = 6;
    //initialise the space for the parameters
    rateRatios = new vector < double > [numberRatesMatrices] ( 14 );
    frequencies = new vector < double > [numberFrequenciesCategories] ( 6 );

    //initialise the matrixIndex for correspondance between a matrix element
    //and an index in the rate ratios vector
    initMatrixIndex();

    //call the basic initialisation method
    RnaModel::initialisation( sequenceTable, modelId );
    
    //if a sequence table is provided initialise rate ratios and frequencies
    //accordingly
    if ( sequenceTable ) {
        vector < double > freq16;
        array2D < double > rate16;

        //initialise the vector of frequencies from the empirical frequencies
        //of the 16 possible symbols
        freq16 = retrieveEmpiricalFrequencies( sequenceTable, modelId );
        //mismatches are ignored
        double sum = freq16[0];
        for ( int i = 0; i < 6 ; ++i ) {
            sum += freq16[i];
        }
        for ( int i = 0; i < 6 ; ++i ) {
            frequencies[0][i] = freq16[i] / sum;
        }
        //if the discrete gamma model uses more than one set of frequencies
        //initialise the other ones to the same values
        for ( int i = 1; i < numberFrequenciesCategories; ++i ) {
            frequencies[i] = frequencies[0];
        }

        
        //estimation of all the rate ratios (and proportion of invariant sites
        //if relevant from the empirical sequences
        if ( invariantCategory ) {
            rate16 = retrieveEmpiricalRates( sequenceTable, modelId, & proportionInvariantSites );
        }
        else {
            rate16 = retrieveEmpiricalRates( sequenceTable, modelId );
        }
        //fill the rate vector according to the estimation
        for ( int i = 0; i < 5 ; ++i ) {
            for ( int j = i + 1; j < 6 ; ++j ) {
                int index = matrixIndex(i, j);
                //if the rate is not the reference rate (0, 2)
                if ( index >= 0 ) {
                    rateRatios[0] [index] = rate16( i, j );
                }
            }
        }
        for ( int i = 1; i < numberRatesMatrices; i++ ) {
            rateRatios[i] = rateRatios[0];
        }
    }
    //init the equivalency table (correspondance symbol/states)
    initEquivalencyTable();
    //first initialisation of the mutation rate matrix
    //(factor(s) used with the subtitution matrix to reach the specified
    //average substitution rate) and eigen system
    setAverageMutationRate( averageMutationRate );
}


// Destructor
RNA6A::~RNA6A() {
}


// The models unique name
string RNA6A::getName( void ) const {
    string name;
    char num[3];

    name = string( "RNA6A" );
    if ( discreteGamma ) {
        sprintf( num, "%i", numberGammaCategories );
        name = name + string( " + dG" ) + string( num );
    }
    if ( invariantCategory == 1 ) {
        name = name + string( " + I" );
    }
    return ( name );
}




void RNA6A::initMatrixIndex() {
    matrixIndex.resize( 6, 6 );
    int index = -1;
    for ( int i = 0 ; i < 5 ; ++i ) {
        for ( int j = i + 1 ; j < 6 ; ++j ) {
            if ( ( i == 0 ) && ( j == 2 ) ) {
                //the reference ratio is given the negative index of -2
                matrixIndex(i, j) = matrixIndex(j, i) = -2;
            }
            else {
                ++index;
                matrixIndex(i, j) = matrixIndex(j, i) = index;
            }
        }
    }
    assert( index == 13 );
    //initialise the diagonal
    for ( int i = 0 ; i < 6 ; ++i ) {
        matrixIndex(i, i) = -3;
    }
}


void RNA6A::initEquivalencyTable() {
    int i, j;

    // model states : AU, GU, GC, UA, UG, CG .
    equivalencyTable.resize( 64, 6 );

    // Set the patterns for valid states
    for ( i = 0; i < 6; ++i ) {
        for ( j = 0; j < 6; ++j ) {
            equivalencyTable( i, j ) = 0.0;
        }
        equivalencyTable( i, i ) = 1.0;
    }

    // Match all other pairs (unknown)
    for ( i = 6; i < 64; ++i ) {
        for ( j = 0; j < 6; ++j ) {
            equivalencyTable( i, j ) = 1.0;
        }
    }
}



string RNA6A::getState( int stateNumber, int symbolCategory ) const {
    assert( symbolCategory == 0 );
    switch ( stateNumber ) {
        // Watson-Crick base pairs
        case 0 :
            return ( string( "AU" ) );
        case 1 :
            return ( string( "GU" ) );
        case 2 :
            return ( string( "GC" ) );
        case 3 :
            return ( string( "UA" ) );
        case 4 :
            return ( string( "UG" ) );
        case 5 :
            return ( string( "CG" ) );
        default :
            return("");
    }
}


void RNA6A::setEigenMatrix() {
    for ( int category = 0; category < MAX( numberGammaCategories, 1 );
          ++category ) {
        int freqCat = frequencyCategory[category + invariantCategory];
        int matrixCat = matrixCategory[category];


        // Initialise the rate matrix
        for ( int i = 0; i < 5; ++i ) {
            for ( int j = i + 1; j < 6; ++j ) {
                int index = matrixIndex(i, j);
                switch (index){
                    //if reference rate
                    case -2 : rateMatrix[category] ( i, j ) =
                           frequencies[freqCat] [j] * mutationRate[category];
                              rateMatrix[category] ( j, i ) =
                           frequencies[freqCat] [i] * mutationRate[category];
                              break;
                    //if null rate
                    case -1 : rateMatrix[category] ( i, j ) = 0.0;
                              rateMatrix[category] ( j, i ) = 0.0;
                              break;
                    //otherwise
                    default : rateMatrix[category] ( i, j ) =
                                  rateRatios[matrixCat] [index] *
                             frequencies[freqCat] [j] * mutationRate[category];
                              rateMatrix[category] ( j, i ) =
                                  rateRatios[matrixCat] [index] *
                             frequencies[freqCat] [i] * mutationRate[category];
#ifdef DEBUG
                             assert( !isnan(rateRatios[matrixCat][index]) );
#endif
                             break;
                }
            }
        }
        for ( int i = 0; i < 6; ++i ) {
            double sum = 0.0;
            for ( int j = 0; j < 6; ++j ) {
                if ( i != j ){
                    sum += rateMatrix[category] ( i, j );
                }
            }
            rateMatrix[category] ( i, i ) = -sum;
        }

        vector < double > junk( 6 );
        for ( int i = 0; i < 6; ++i ) {
            eigenValues[category] [i] = 0.0;
            junk[i] = 0.0;
            for ( int j = 0; j < 6; ++j ){
                eigenMatrix[category] ( i, j ) = 0.0;
                ieigenMatrix[category] ( i, j ) = 0.0;
            }
        }
        // Find the eigenvalues and eigenvectors of the rate matrix
        LeftEigenSystem( rateMatrix[category], eigenMatrix[category],
        eigenValues[category], junk );
        // Find the inverse of the matrix whose rows are the left eigen
        // vectors of the rate matrix
        ieigenMatrix[category] = inverse( eigenMatrix[category] );
    }
}




