#include <assert.h>

#include <iostream>
#include <iomanip>

#include "Models/RNA6B.h"

#include "statlib.h"

#define SQR(n) (n*n)
#define SIGN(n) (n>=0.0 ? 1.0 : -1.0)


using namespace std;

//register the model to the model factory with the name RNA6B
RNA6B RNA6B::prototype( "RNA6B" );

RNA6B::RNA6B( const string & registrationName ) :
RNA6A( registrationName ) {
//the private constructor is called by the prototype only
//RNA6B is a restriction of RNA6A
}

RNA6B::RNA6B( ParametersSet & parameters )
: RNA6A( parameters ) {
//the normal constructor is called by the prototype's clone method
//RNA6B is a restriction of RNA6A
}


void RNA6B::initialisation( SequenceTable * sequenceTable, int modelId ) {
    //The initialisation procedure for the 6B model is slightly different
    //than for the 6A model, therefore we redefine it
    //initialise the number of parameters
    numberRatesRatiosPerMatrix = 2;
    numberFrequencies = 6;
    //initialise the space for the parameters
    rateRatios = new vector < double > [numberRatesMatrices] ( 2 );
    frequencies = new vector < double > [numberFrequenciesCategories] ( 6 );

    //initialise the matrixIndex for correspondance between a matrix element
    //and an index in the rate ratios vector
    initMatrixIndex();

    //call the basic initialisation method
    RnaModel::initialisation( sequenceTable, modelId );
    
    //if a sequence table is provided initialise rate ratios and frequencies
    //accordingly
    if ( sequenceTable ) {
        vector < double > freq16;
        array2D < double > rate16;

        //initialise the vector of frequencies from the empirical frequencies
        //of the 16 possible symbols
        freq16 = retrieveEmpiricalFrequencies( sequenceTable, modelId );
        //mismatches are ignored
        double sum = freq16[0];
        for ( int i = 0; i < 6 ; ++i ) {
            sum += freq16[i];
        }
        for ( int i = 0; i < 6 ; ++i ) {
            frequencies[0][i] = freq16[i] / sum;
        }
        //if the discrete gamma model uses more than one set of frequencies
        //initialise the other ones to the same values
        for ( int i = 1; i < numberFrequenciesCategories; ++i ) {
            frequencies[i] = frequencies[0];
        }

        
        //estimation of all the rate ratios (and proportion of invariant sites
        //if relevant from the empirical sequences
        if ( invariantCategory ) {
            rate16 = retrieveEmpiricalRates( sequenceTable, modelId, & proportionInvariantSites );
        }
        else {
            rate16 = retrieveEmpiricalRates( sequenceTable, modelId );
        }

        //fill the rate vector according to the estimation
        // reference ratio
        double ref = ( rate16( 0, 2 ) + rate16( 3, 5 ) ) / 2;
        // 0 - single transition rate ratio
        double singleRef = ( rate16( 0, 1 ) + rate16( 1, 2 ) + rate16( 3, 4 ) +
        rate16( 4, 5 ) );
        rateRatios[0][0]= singleRef / (4*ref);
        // 1 - double transversion rate ratio
        double doubleRef = 0.0;
        for ( int i = 0; i < 3; ++i ) {
            for ( int j = 3; j < 6; ++j ) {
                doubleRef += rate16( i, j );
            }
        }
        rateRatios[0] [1] = doubleRef/(9*ref);
        //if the discrete gamma model uses more than one set of rate ratios
        //initialise the other ones to the same values
        for ( int i = 1; i < numberRatesMatrices; i++ ) {
            rateRatios[i] = rateRatios[0];
        }
    }
    //init the equivalency table (correspondance symbol/states)
    initEquivalencyTable();
    //first initialisation of the mutation rate matrix
    //(factor(s) used with the subtitution matrix to reach the specified
    //average substitution rate) and eigen system
    setAverageMutationRate( averageMutationRate );
}


// Destructor
RNA6B::~RNA6B() {
}


// The models unique name
string RNA6B::getName( void ) const {
    string name;
    char num[3];

    name = string( "RNA6B" );
    if ( discreteGamma ) {
        sprintf( num, "%i", numberGammaCategories );
        name = name + string( " + dG" ) + string( num );
    }
    if ( invariantCategory == 1 ) {
        name = name + string( " + I" );
    }
    return ( name );
}




void RNA6B::initMatrixIndex() {
    matrixIndex.resize( 6, 6 );
    //the reference ratio is given the negative index of -2
    matrixIndex( 0, 2 ) = matrixIndex(3, 5) = -2;
    // 0 - single transition rate ratio
    matrixIndex(0, 1) = matrixIndex(1, 2) =
        matrixIndex(3, 4) = matrixIndex(4, 5) = 0;
    // 1 - double transversion rate ratio
    for ( int i = 0; i < 3; ++i ) {
        for ( int j = 3; j < 6; ++j ) {
            matrixIndex(i, j) = 1;
        }
    }
    //copy on the other triangle of the matrix
    for ( int i = 1; i < 6; ++i ) {
        for ( int j = 0; j < i; ++j ) {
            matrixIndex(i, j) = matrixIndex(j, i);
        }
    }
    //initialise the diagonal
    for ( int i = 0 ; i < 6 ; ++i ) {
        matrixIndex(i, i) = -3;
    }
}


