#pragma implementation

#include <assert.h>

#include <iostream>
#include <iomanip>


#include "Models/RNA7A.h"

#include "statlib.h"

#define SQR(n) (n*n)
#define SIGN(n) (n>=0.0 ? 1.0 : -1.0)


using namespace std;

//register the model to the model factory with the name RNA7A
RNA7A RNA7A::prototype( "RNA7A" );

RNA7A::RNA7A( const string & registrationName ) :
RnaModel( registrationName ) {
//the private constructor is called by the prototype only
}

RNA7A::RNA7A( ParametersSet & parameters )
: RnaModel( parameters, 7 ) {
//the normal constructor is called by the prototype's clone method
//RNA7A is a RnaModel, its matrix size is 7*7
}


void RNA7A::initialisation( SequenceTable * sequenceTable, int modelId ) {
    //initialise the number of parameters
    numberRatesRatiosPerMatrix = 20;
    numberFrequencies = 7;
    //initialise the space for the parameters
    rateRatios = new vector < double > [numberRatesMatrices] ( 20 );
    frequencies = new vector < double > [numberFrequenciesCategories] ( 7 );

    //initialise the matrixIndex for correspondance between a matrix element
    //and an index in the rate ratios vector
    initMatrixIndex();

    //call the basic initialisation method
    RnaModel::initialisation( sequenceTable, modelId );
    
    //if a sequence table is provided initialise rate ratios and frequencies
    //accordingly
    if ( sequenceTable ) {
        vector < double > freq16;
        array2D < double > rate16;

        //initialise the vector of frequencies from the empirical frequencies
        //of the 16 possible symbols
        freq16 = retrieveEmpiricalFrequencies( sequenceTable, modelId );
        for ( int i = 0; i < 7 ; ++i ) {
            frequencies[0] [i] = freq16[i];
        }
        //all mismatch are lumped in the same category
        for ( int i = 7; i < 16 ; ++i ) {
            frequencies[0] [6] += freq16[i];
        }
        //if the discrete gamma model uses more than one set of frequencies
        //initialise the other ones to the same values
        for ( int i = 1; i < numberFrequenciesCategories; ++i ) {
            frequencies[i] = frequencies[0];
        }

        
        //estimation of all the rate ratios (and proportion of invariant sites
        //if relevant from the empirical sequences
        if ( invariantCategory ) {
            rate16 = retrieveEmpiricalRates( sequenceTable, modelId, & proportionInvariantSites );
        }
        else {
            rate16 = retrieveEmpiricalRates( sequenceTable, modelId );
        }
        //fill the rate vector according to the estimation
        for ( int i = 0; i < 5 ; ++i ) {
            for ( int j = i + 1; j < 6 ; ++j ) {
                int index = matrixIndex(i, j);
                //if the rate is not the reference rate (0, 2)
                if ( index >= 0 ) {
                    rateRatios[0] [index] = rate16( i, j );
                }
            }
        }
        //all mismatch are lumped in the same category
        for ( int i = 0; i < 6 ; ++i ) {
            rateRatios[0] [matrixIndex(i, 6)] = 0.0;
            for ( int j = 6; j < 16; ++j ) {
                rateRatios[0] [matrixIndex(i, 6)] += rate16( i, j );
            }
        }
        //if the discrete gamma model uses more than one set of rate ratios
        //initialise the other ones to the same values
        for ( int i = 1; i < numberRatesMatrices; i++ ) {
            rateRatios[i] = rateRatios[0];
        }
    }
    //init the equivalency table (correspondance symbol/states)
    initEquivalencyTable();
    //first initialisation of the mutation rate matrix
    //(factor(s) used with the subtitution matrix to reach the specified
    //average substitution rate) and eigen system
    setAverageMutationRate( averageMutationRate );
}


// Destructor
RNA7A::~RNA7A() {
}


// The models unique name
string RNA7A::getName( void ) const {
    string name;
    char num[3];

    name = string( "RNA7A" );
    if ( discreteGamma ) {
        sprintf( num, "%i", numberGammaCategories );
        name = name + string( " + dG" ) + string( num );
    }
    if ( invariantCategory == 1 ) {
        name = name + string( " + I" );
    }
    return ( name );
}




void RNA7A::initMatrixIndex() {
    matrixIndex.resize( 7, 7 );
    int index = -1;
    for ( int i = 0 ; i < 6 ; ++i ) {
        for ( int j = i + 1 ; j < 7 ; ++j ) {
            if ( ( i == 0 ) && ( j == 2 ) ) {
                //the reference ratio is given the negative index of -2
                matrixIndex(i, j) = matrixIndex(j, i) = -2;
            }
            else {
                ++index;
                matrixIndex(i, j) = matrixIndex(j, i) = index;
            }
        }
    }
    assert( index == 19 );
    //initialise the diagonal
    for ( int i = 0 ; i < 7 ; ++i ) {
        matrixIndex(i, i) = -3;
    }
}


void RNA7A::initEquivalencyTable() {
    int i, j;

    // model states : AU, GU, GC, UA, UG, CG, MM .
    equivalencyTable.resize( 64, 7 );

    // Clear the table.
    for ( j = 0; j < 7; ++j ) {
        for ( i = 0; i < 64; ++i ) {
            equivalencyTable( i, j ) = 0.0;
        }
    }

    //valid states
    for ( i = 0; i < 6; ++i ) {
        equivalencyTable( i, i ) = 1.0;
    }


    // mismatch
    for ( i = 6; i < 16; ++i ) {
        equivalencyTable( i, 6 ) = 1.0;
    }

    equivalencyTable( 16, 0 ) = equivalencyTable( 16, 6 ) = 1.0; // AX
    equivalencyTable( 17, 6 ) = 1.0; // AR (AG or AA)
    equivalencyTable( 18, 0 ) = equivalencyTable( 18, 6 ) = 1.0; // AY (AU or AC)

    equivalencyTable( 19, 5 ) = equivalencyTable( 19, 6 ) = 1.0; // CX
    equivalencyTable( 20, 5 ) = equivalencyTable( 20, 6 ) = 1.0; // CR (CG or CA)
    equivalencyTable( 21, 6 ) = 1.0; // CY (CC or CU)

    equivalencyTable( 22, 1 ) = equivalencyTable( 22, 2 ) =
    equivalencyTable( 22, 6 ) = 1.0; // GX
    equivalencyTable( 23, 6 ) = 1.0; // GR (GA or GG)
    equivalencyTable( 24, 1 ) = equivalencyTable( 24, 2 ) = 1.0; // GY (GC or GU)

    equivalencyTable( 25, 3 ) = equivalencyTable( 25, 4 ) =
    equivalencyTable( 25, 6 ) = 1.0; // UX
    equivalencyTable( 26, 3 ) = equivalencyTable( 26, 4 ) = 1.0; // UR (UA or UG)
    equivalencyTable( 27, 6 ) = 1.0; // UY (UC or UU)

    equivalencyTable( 28, 3 ) = equivalencyTable( 28, 6 ) = 1.0; // XA
    equivalencyTable( 29, 6 ) = 1.0; // RA (GA or AA)
    equivalencyTable( 30, 3 ) = equivalencyTable( 30, 6 ) = 1.0; // YA (CA or UA)

    equivalencyTable( 31, 2 ) = equivalencyTable( 31, 6 ) = 1.0; // XC
    equivalencyTable( 32, 2 ) = equivalencyTable( 32, 6 ) = 1.0; // RC (GC or AC)
    equivalencyTable( 33, 6 ) = 1.0; // YC (CC or UC)

    equivalencyTable( 34, 4 ) = equivalencyTable( 34, 5 ) =
    equivalencyTable( 34, 6 ) = 1.0; // XG
    equivalencyTable( 35, 6 ) = 1.0; // RG (AG or GG)
    equivalencyTable( 36, 4 ) = equivalencyTable( 36, 5 ) = 1.0; // YG (CG or UG)

    equivalencyTable( 37, 0 ) = equivalencyTable( 37, 1 ) =
    equivalencyTable( 37, 6 ) = 1.0; // XU
    equivalencyTable( 38, 0 ) = equivalencyTable( 38, 1 ) = 1.0; // RU (AU or GU)
    equivalencyTable( 39, 6 ) = 1.0; // YU (CU or UU)

    for ( i = 0; i < 7; i++ ) {
        equivalencyTable( 40, i ) = 1.0; //XX
    }
    for ( i = 3; i < 7; i++ ) {
        equivalencyTable( 41, i ) = 1.0; //XR (XA,XG,MM)
    }
    equivalencyTable( 42, 0 ) = equivalencyTable( 42, 1 ) =
    equivalencyTable( 42, 2 ) = equivalencyTable( 42, 6 ) = 1.0; // XY (XC or XU)

    for ( i = 3; i < 7; i++ ) {
        equivalencyTable( 43, i ) = 1.0; // YX (CX or UX)
    }

    for ( i = 3; i < 7; i++ ) {
        equivalencyTable( 44, i ) = 1.0; // YR (CA,CG,UA,UG)
    }

    equivalencyTable( 45, 6 ) = 1.0; // YY (CC,CU,UC)

    equivalencyTable( 46, 0 ) = equivalencyTable( 46, 1 ) =
    equivalencyTable( 46, 2 ) = equivalencyTable( 46, 6 ) = 1.0; //RX (AX,GX,MM)

    equivalencyTable( 47, 6 ) = 1.0; // RR (AG,GA)

    equivalencyTable( 48, 0 ) = equivalencyTable( 48, 1 ) =
    equivalencyTable( 48, 2 ) = equivalencyTable( 48, 6 ) = 1.0; // RY
    // (AC, AU, GC, GU)

    equivalencyTable( 49, 3 ) = equivalencyTable( 49, 6 ) = 1.0; //-A
    equivalencyTable( 50, 4 ) = equivalencyTable( 50, 5 ) =
    equivalencyTable( 50, 6 ) = 1.0; //-G
    equivalencyTable( 51, 2 ) = equivalencyTable( 51, 6 ) = 1.0; //-C
    equivalencyTable( 52, 0 ) = equivalencyTable( 52, 1 ) = 1.0; //-U

    for ( i = 0; i < 7; i++ ) {
        equivalencyTable( 53, i ) = 1.0; //-X
    }
    for ( i = 3; i < 7; i++ ) {
        equivalencyTable( 54, i ) = 1.0; //-R (-A,-G,MM)
    }
    equivalencyTable( 55, 0 ) = equivalencyTable( 55, 1 ) =
    equivalencyTable( 55, 2 ) = equivalencyTable( 55, 6 ) = 1.0; // -Y (-C or -U)

    equivalencyTable( 56, 0 ) = equivalencyTable( 56, 6 ) = 1.0; //A-
    equivalencyTable( 57, 1 ) = equivalencyTable( 57, 2 ) =
    equivalencyTable( 57, 6 ) = 1.0; //G-
    equivalencyTable( 58, 5 ) = equivalencyTable( 58, 6 ) = 1.0; //C-
    equivalencyTable( 59, 3 ) = equivalencyTable( 59, 4 ) =
    equivalencyTable( 59, 6 ) = 1.0; //U-

    equivalencyTable( 61, 0 ) = equivalencyTable( 61,
    1 ) = equivalencyTable( 61, 2 ) =
    equivalencyTable( 61, 6 ) = 1.0; //R- (A-, G-)

    for ( i = 3; i < 7; i++ ) {
        equivalencyTable( 62, i ) = 1.0; //Y- (C-, U-)
    }

    for ( i = 0; i < 7; i++ ) {
        equivalencyTable( 60, i ) = 1.0;    //X-
        equivalencyTable( 63, i ) = 1.0;    //--
    }
}

string RNA7A::getState( int stateNumber, int symbolCategory ) const {
    assert( symbolCategory == 0 );
    switch ( stateNumber ) {
        // Watson-Crick base pairs
        case 0 :
            return ( string( "AU" ) );
        case 1 :
            return ( string( "GU" ) );
        case 2 :
            return ( string( "GC" ) );
        case 3 :
            return ( string( "UA" ) );
        case 4 :
            return ( string( "UG" ) );
        case 5 :
            return ( string( "CG" ) );
        case 6 :
            return ( string( "MM" ) );
        default :
            return("");
    }
}


void RNA7A::setEigenMatrix() {
    for ( int category = 0; category < MAX( numberGammaCategories, 1 );
          ++category ) {
        int freqCat = frequencyCategory[category + invariantCategory];
        int matrixCat = matrixCategory[category];


        // Initialise the rate matrix
        for ( int i = 0; i < 6; ++i ) {
            for ( int j = i + 1; j < 7; ++j ) {
                int index = matrixIndex(i, j);
                switch (index){
                    //if reference rate
                    case -2 : rateMatrix[category] ( i, j ) =
                           frequencies[freqCat] [j] * mutationRate[category];
                              rateMatrix[category] ( j, i ) =
                           frequencies[freqCat] [i] * mutationRate[category];
                              break;
                    //if null rate
                    case -1 : rateMatrix[category] ( i, j ) = 0.0;
                              rateMatrix[category] ( j, i ) = 0.0;
                              break;
                    //otherwise
                    default : rateMatrix[category] ( i, j ) =
                                  rateRatios[matrixCat] [index] *
                             frequencies[freqCat] [j] * mutationRate[category];
                              rateMatrix[category] ( j, i ) =
                                  rateRatios[matrixCat] [index] *
                             frequencies[freqCat] [i] * mutationRate[category];
#ifdef DEBUG
                             assert( !isnan(rateRatios[matrixCat][index]) );
#endif
                             break;
                }
            }
        }
        for ( int i = 0; i < 7; ++i ) {
            double sum = 0.0;
            for ( int j = 0; j < 7; ++j ) {
                if ( i != j ){
                    sum += rateMatrix[category] ( i, j );
                }
            }
            rateMatrix[category] ( i, i ) = -sum;
        }

        vector < double > junk( 7 );
        for ( int i = 0; i < 7; ++i ) {
            eigenValues[category] [i] = 0.0;
            junk[i] = 0.0;
            for ( int j = 0; j < 7; ++j ){
                eigenMatrix[category] ( i, j ) = 0.0;
                ieigenMatrix[category] ( i, j ) = 0.0;
            }
        }
        // Find the eigenvalues and eigenvectors of the rate matrix
        LeftEigenSystem( rateMatrix[category], eigenMatrix[category],
        eigenValues[category], junk );
        // Find the inverse of the matrix whose rows are the left eigen
        // vectors of the rate matrix
        ieigenMatrix[category] = inverse( eigenMatrix[category] );
    }
}




