#include <assert.h>

#include <iostream>
#include <iomanip>


#include "Models/RNA7D.h"

#include "statlib.h"

#define SQR(n) (n*n)
#define SIGN(n) (n>=0.0 ? 1.0 : -1.0)


using namespace std;

//register the model to the model factory with the name RNA7D
RNA7D RNA7D::prototype( "RNA7D" );

RNA7D::RNA7D( const string & registrationName ) :
RNA7A( registrationName ) {
//the private constructor is called by the prototype only
//RNA7D is a restriction of RNA7A
}

RNA7D::RNA7D( ParametersSet & parameters )
: RNA7A( parameters ) {
//the normal constructor is called by the prototype's clone method
//RNA7D is a restriction of RNA7A
}


void RNA7D::initialisation( SequenceTable * sequenceTable, int modelId ) {
    //The initialisation procedure for the 7D model is slightly different
    //than for the 7A model, therefore we redefine it
    //initialise the number of parameters
    numberRatesRatiosPerMatrix = 3;
    numberFrequencies = 7;
    //initialise the space for the parameters
    rateRatios = new vector < double > [numberRatesMatrices] ( 3 );
    frequencies = new vector < double > [numberFrequenciesCategories] ( 7 );

    initMatrixIndex();

    //call the basic initialisation method
    RnaModel::initialisation( sequenceTable, modelId );
    
    //if a sequence table is provided initialise rate ratios and frequencies
    //accordingly
    if ( sequenceTable ) {
        vector < double > freq16;
        array2D < double > rate16;

        //initialise the vector of frequencies from the empirical frequencies
        //of the 16 possible symbols
        freq16 = retrieveEmpiricalFrequencies( sequenceTable, modelId );
        for ( int i = 0; i < 7 ; i++ ) {
            frequencies[0] [i] = freq16[i];
        }
        //all mismatch are in the same category
        for ( int i = 7; i < 16 ; i++ ) {
            frequencies[0] [6] += freq16[i];
        }
        //if the discrete gamma model uses more than one set of frequencies
        //initialise the other ones to the same values
        for ( int i = 1; i < numberFrequenciesCategories; ++i ) {
            frequencies[i] = frequencies[0];
        }

        
        //estimation of all the rate ratios (and proportion of invariant sites
        //if relevant from the empirical sequences
        if ( invariantCategory ) {
            rate16 = retrieveEmpiricalRates( sequenceTable, modelId, & proportionInvariantSites );
        }
        else {
            rate16 = retrieveEmpiricalRates( sequenceTable, modelId );
        }
        //fill the rate vector according to the estimation        
        //reference ratio
        double ref = ( rate16( 0, 2 ) + rate16( 3, 5 ) ) / 2;
        // 0 - single transition rate ratio
        double singleRef = ( rate16( 0, 1 ) + rate16( 1, 2 ) + rate16( 3, 4 ) +
        rate16( 4, 5 ) );
        rateRatios[0][0]= singleRef / (4*ref);
        // 1 - double transversion rate ratio
        double doubleRef = 0.0;
        for ( int i = 0; i < 3; ++i ) {
            for ( int j = 3; j < 6; ++j ) {
                doubleRef += rate16( i, j );
            }
        }
        rateRatios[0] [1] = doubleRef/(9*ref);
        // 2 - mismatch rate ratio
        double mismatchRef = 0.0;
        for ( int i = 0; i < 6; ++i ) {
            for ( int j = 6; j < 16; ++j ) {
                mismatchRef += rate16( i, j );
            }
        }
        rateRatios[0][2] = mismatchRef/(60*ref);
        //if the discrete gamma model uses more than one set of rate ratios
        //initialise the other ones to the same values
        for ( int i = 1; i < numberRatesMatrices; i++ ) {
            rateRatios[i] = rateRatios[0];
        }
    }
    //init the equivalency table (correspondance symbol/states)
    initEquivalencyTable();
    //first initialisation of the mutation rate matrix
    //(factor(s) used with the subtitution matrix to reach the specified
    //average substitution rate) and eigen system
    setAverageMutationRate( averageMutationRate );
}


// Destructor
RNA7D::~RNA7D() {
}


// The models unique name
string RNA7D::getName( void ) const {
    string name;
    char num[3];

    name = string( "RNA7D" );
    if ( discreteGamma ) {
        sprintf( num, "%i", numberGammaCategories );
        name = name + string( " + dG" ) + string( num );
    }
    if ( invariantCategory == 1 ) {
        name = name + string( " + I" );
    }
    return ( name );
}



void RNA7D::initMatrixIndex() {
    matrixIndex.resize( 7, 7 );  
    //the reference ratio is given the negative index of -2
    matrixIndex( 0, 2 ) = matrixIndex(3, 5) = -2;
    // 0 - single transition rate ratio
    matrixIndex(0, 1) = matrixIndex(1, 2) =
        matrixIndex(3, 4) = matrixIndex(4, 5) = 0;
    // 1 - double transversion rate ratio
    for ( int i = 0; i < 3; ++i ) {
        for ( int j = 3; j < 6; ++j ) {
            matrixIndex(i, j) = 1;
        }
    }
    // 2 - mismatch rate ratio
    for ( int i = 0; i < 6; ++i ) {
        matrixIndex(i, 6) = 2;
    }
    //copy on the other triangle of the matrix
    for ( int i = 1; i < 7; ++i ) {
        for ( int j = 0; j < i; ++j ) {
            matrixIndex(i, j) = matrixIndex(j, i);
        }
    }
    //initialise the diagonal
    for ( int i = 0 ; i < 7 ; ++i ) {
        matrixIndex(i, i) = -3;
    }
}
