#include "Models/RnaModel.h"

#include <iostream>
#include <iomanip>

#include "SequenceTable.h"

#include "Models/Perturbator.h"
#include "Util/ModelFactory.h"

#include "PatternDesign/Singleton.h"
#include "Util/statlib.h"
#include "Util/array2D.h"
#include "Util/randombox.h"
#include "Util/ParametersSet.h"


using namespace std;

RnaModel::RnaModel( const string & registrationName ) :
        MatrixModel( registrationName ){
}

RnaModel::RnaModel( ParametersSet & parameters, int matrixSize ) :
        MatrixModel( parameters ) {
  
    this->matrixSize = matrixSize;
    
    if ( discreteGamma ){
        eigenMatrix = new array2D < double >
        [numberGammaCategories] ( matrixSize, matrixSize );
        ieigenMatrix = new array2D < double >
        [numberGammaCategories] ( matrixSize, matrixSize );
        eigenValues = new vector < double > [numberGammaCategories] ( matrixSize );
        rateMatrix = new array2D < double >
        [numberGammaCategories] ( matrixSize, matrixSize );
    }
    else{
        eigenMatrix = new array2D < double >( matrixSize, matrixSize );
        ieigenMatrix = new array2D < double >( matrixSize, matrixSize );
        eigenValues = new vector < double >( matrixSize );
        rateMatrix = new array2D < double >( matrixSize, matrixSize );
    }
}

RnaModel::~RnaModel() {
    if ( discreteGamma ){
        if ( eigenMatrix != NULL ){
            delete[] eigenMatrix;
            eigenMatrix = NULL;
        }
        if ( ieigenMatrix != NULL ){
            delete[] ieigenMatrix;
            ieigenMatrix = NULL;
        }
        if ( eigenValues != NULL ){
            delete[] eigenValues;
            eigenValues = NULL;
        }
        if ( rateMatrix != NULL ){
            delete[] rateMatrix;
            rateMatrix = NULL;
        }
    }
    else{
        if ( eigenMatrix != NULL ){
            delete eigenMatrix;
            eigenMatrix = NULL;
        }
        if ( ieigenMatrix != NULL ){
            delete ieigenMatrix;
            ieigenMatrix = NULL;
        }
        if ( eigenValues != NULL ){
            delete eigenValues;
            eigenValues = NULL;
        }
        if ( rateMatrix != NULL ){
            delete rateMatrix;
            rateMatrix = NULL;
        }
    }
}


array2D< double > RnaModel::retrieveEmpiricalRates( SequenceTable * sequenceTable,
int modelId, double * propInvariant ) {


    array2D < double > rate16( 16, 16 );
    
    const array2D< string > & sequences = sequenceTable->getSequences( modelId );
    const vector< pair< string, unsigned int > > & invariantBases = sequenceTable->getInvariantBases( modelId );
    int s1, s2;

    for ( int i = 0 ; i < 16; ++i ) {
        for ( int j = 0 ; j < 16; ++j ) {
            rate16( i, j ) = 0.0;
        }
    }


    for ( unsigned int seq = 0;
          seq < sequenceTable->getNumberSpecies(); ++seq ) {
        for ( unsigned int seq2 = seq + 1;
              seq2 < sequenceTable->getNumberSpecies(); ++seq2 ) {
            for ( unsigned int site = 0;
                  site < sequences.numberColumns();
                  ++site ) {
                s1 = getSymbolNumber( sequences( seq, site ) );
                s2 = getSymbolNumber( sequences( seq2, site ) );
                if ( (s1 == -1) || (s2 == -1) ){
                     cerr << "Unrecognized symbol \"";
                     if (s1 == -1){
                         cerr << sequences( seq, site );
                     }
                     else{
                         cerr << sequences( seq2, site );
                     }
                     cerr << "\" during initialisation of model " << modelId << endl;
                     exit(1);
                }       
                if ( ( s1 < 16 ) && ( s2 < 16 ) && ( s1 != s2 ) ){
                    ++rate16( s1, s2 );
                }
            }
        }
    }

    double reference = ( rate16( 0, 2 ) + rate16( 2, 0 ) );
    if ( reference == 0.0 ) {
        cerr << "Error while retrieving empirical rates. "
        << "Not enough paired site nucleotid ?" << endl;
        exit(1);
    }
    for ( int i = 0 ; i < 16; i ++ ) {
        for ( int j = i + 1; j < 16; j++ ) {
            rate16( i, j ) = ( rate16( i, j ) + rate16( j, i ) ) /
            ( reference );
            rate16( j, i ) = rate16( i, j );
        }
    }

    if ( propInvariant != NULL ) {
        int numberInvariant = 0;
        for ( vector< pair< string, unsigned int > >::const_iterator iter = invariantBases.begin();
              iter != invariantBases.end(); ++iter ){
            numberInvariant += (*iter).second;
        }
        * propInvariant = (double)numberInvariant/
                          (double)(numberInvariant + sequences.numberColumns());
    }
    return rate16;
}


vector < double > RnaModel::retrieveEmpiricalFrequencies( SequenceTable *
sequenceTable, int modelId ) {

    vector < double > freq16( 16 );

    // Find empirical frequencies
    for ( int i = 0; i < 16; ++i ) {
        freq16[i] = 0.0;
    }

    const array2D< string > & sequences = sequenceTable->getSequences( modelId );
    const vector< pair< string, unsigned int > > & invariantBases = sequenceTable->getInvariantBases( modelId );
    int s;

    for ( unsigned int sequence = 0;
          sequence < sequenceTable->getNumberSpecies(); ++sequence ) {
        for ( unsigned int site = 0;
              site < sequenceTable->getSequencesLength(modelId); ++site ) {
            s = getSymbolNumber( sequences( sequence, site ) );
            for ( int nucleotide = 0; nucleotide < 16; ++nucleotide ) {
                if ( s == nucleotide ) {
                    freq16[nucleotide] += 1.0;
                }
            }
        }
    }
    for ( vector< pair< string, unsigned int > >::const_iterator iter = invariantBases.begin();
          iter != invariantBases.end(); ++iter ){
        s = getSymbolNumber( (*iter).first );
        for ( int nucleotide = 0; nucleotide < 16; ++nucleotide ) {
            if ( s == nucleotide ) {
                freq16[nucleotide] += 1.0;
            }
        }        
    }

    double sum = 0.0;
    for ( int nucleotide = 0; nucleotide < 16; nucleotide++ ) {
        sum +=  freq16[nucleotide];
    }
    for ( int nucleotide = 0; nucleotide < 16; nucleotide++ ) {
        freq16[nucleotide] /= sum;
    }

    return freq16;
}



int RnaModel::getSymbolNumber( const string & pairedBase, int symbolCategory ) const {
    assert( symbolCategory == 0 );
    switch ( pairedBase[0] ) {
        case 'a' :
        case 'A' :
            switch ( pairedBase[1] ) {
                case 'a':
                case 'A':
                    return 6;
                case 'g':
                case 'G':
                    return 7;
                case 'c':
                case 'C':
                    return 8;
                case 'u':
                case 'U':
                case 't':
                case 'T':
                    return 0;
                case 'x':
                case 'X':
                case 'N':
                case 'n':
                case '?' :
                    return 16;
                case 'r':
                case 'R':
                    return 17;
                case 'y':
                case 'Y':
                    return 18;
                case '-':
                    return 56;
                default:
                    return 64;
            }
        case 'c' :
        case 'C' :
            switch ( pairedBase[1] ) {
                case 'a':
                case 'A':
                    return 11;
                case 'g':
                case 'G':
                    return 5;
                case 'c':
                case 'C':
                    return 12;
                case 'u':
                case 'U':
                case 't':
                case 'T':
                    return 13;
                case 'x':
                case 'X':
                case 'N':
                case 'n':
                case '?' :
                    return 19;
                case 'r':
                case 'R':
                    return 20;
                case 'y':
                case 'Y':
                    return 21;
                case '-':
                    return 58;
                default:
                    return 64;
            }
        case 'G' :
        case 'g' :
            switch ( pairedBase[1] ) {
                case 'a':
                case 'A':
                    return 9;
                case 'g':
                case 'G':
                    return 10;
                case 'c':
                case 'C':
                    return 2;
                case 'u':
                case 'U':
                case 't':
                case 'T':
                    return 1;
                case 'x':
                case 'X':
                case 'N':
                case 'n':
                case '?' :
                    return 22;
                case 'r':
                case 'R':
                    return 23;
                case 'y':
                case 'Y':
                    return 24;
                case '-':
                    return 57;
                default:
                    return 64;
            }
        case 'u' :
        case 'U' :
        case 't' :
        case 'T' :
            switch ( pairedBase[1] ) {
                case 'a':
                case 'A':
                    return 3;
                case 'g':
                case 'G':
                    return 4;
                case 'c':
                case 'C':
                    return 14;
                case 'u':
                case 'U':
                case 't':
                case 'T':
                    return 15;
                case 'x':
                case 'X':
                case 'N':
                case 'n':
                case '?' :
                    return 25;
                case 'r':
                case 'R':
                    return 26;
                case 'y':
                case 'Y':
                    return 27;
                case '-':
                    return 59;
                default:
                    return 64;
            }
        case 'x' :
        case 'X' :
        case 'n' :
        case 'N' :
        case '?' :
            switch ( pairedBase[1] ) {
                case 'a':
                case 'A':
                    return 28;
                case 'g':
                case 'G':
                    return 34;
                case 'c':
                case 'C':
                    return 31;
                case 'u':
                case 'U':
                case 't':
                case 'T':
                    return 37;
                case 'x':
                case 'X':
                case 'n':
                case 'N':
                case '?' :
                    return 40;
                case 'r':
                case 'R':
                    return 41;
                case 'y':
                case 'Y':
                    return 42;
                case '-':
                    return 60;
                default:
                    return 64;
            }
        case 'r' :
        case 'R' :
            switch ( pairedBase[1] ) {
                case 'a':
                case 'A':
                    return 29;
                case 'g':
                case 'G':
                    return 35;
                case 'c':
                case 'C':
                    return 32;
                case 'u':
                case 'U':
                case 't':
                case 'T':
                    return 38;
                case 'x':
                case 'X':
                case 'n':
                case 'N':
                case '?' :
                    return 46;
                case 'r':
                case 'R':
                    return 47;
                case 'y':
                case 'Y':
                    return 48;
                case '-':
                    return 61;
                default:
                    return 64;
            }
        case 'y' :
        case 'Y' :
            switch ( pairedBase[1] ) {
                case 'a':
                case 'A':
                    return 30;
                case 'c':
                case 'C':
                    return 33;
                case 'g':
                case 'G':
                    return 36;
                case 'u':
                case 'U':
                case 't':
                case 'T':
                    return 39;
                case 'x':
                case 'X':
                case 'n':
                case 'N':
                case '?' :
                    return 43;
                case 'r':
                case 'R':
                    return 44;
                case 'y':
                case 'Y':
                    return 45;
                case '-':
                    return 62;
                default:
                    return 64;
            }
        case '-' :
            switch ( pairedBase[1] ) {
                case 'a':
                case 'A':
                    return 49;
                case 'g':
                case 'G':
                    return 50;
                case 'c':
                case 'C':
                    return 51;
                case 'u':
                case 'U':
                case 't':
                case 'T':
                    return 52;
                case 'x':
                case 'X':
                case 'n':
                case 'N':
                case '?' :
                    return 53;
                case 'r':
                case 'R':
                    return 54;
                case 'y':
                case 'Y':
                    return 55;
                case '-':
                    return 63;
                default:
                    return 64;
            }
        default :
            return 64;
    }
}

string RnaModel::getSymbol( int symbolNumber, int symbolCategory ) const {
    assert( symbolCategory == 0 );
    switch ( symbolNumber ) {
        // Watson-Crick base pairs
        case 0 :
            return ( string( "AU" ) );
        case 1 :
            return ( string( "GU" ) );
        case 2 :
            return ( string( "GC" ) );
        case 3 :
            return ( string( "UA" ) );
        case 4 :
            return ( string( "UG" ) );
        case 5 :
            return ( string( "CG" ) );
            // Mismatches
        case 6 :
            return ( string( "AA" ) );
        case 7 :
            return ( string( "AG" ) );
        case 8 :
            return ( string( "AC" ) );
        case 9 :
            return ( string( "GA" ) );
        case 10 :
            return ( string( "GG" ) );
        case 11 :
            return ( string( "CA" ) );
        case 12 :
            return ( string( "CC" ) );
        case 13 :
            return ( string( "CU" ) );
        case 14 :
            return ( string( "UC" ) );
        case 15 :
            return ( string( "UU" ) );

            // Ambiguous base-pairs
        case 16 :
            return ( string( "AX" ) );
        case 17 :
            return ( string( "AR" ) );
        case 18 :
            return ( string( "AY" ) );
        case 19 :
            return ( string( "CX" ) );
        case 20 :
            return ( string( "CR" ) );
        case 21 :
            return ( string( "CY" ) );
        case 22 :
            return ( string( "GX" ) );
        case 23 :
            return ( string( "GR" ) );
        case 24 :
            return ( string( "GY" ) );
        case 25 :
            return ( string( "UX" ) );
        case 26 :
            return ( string( "UR" ) );
        case 27 :
            return ( string( "UY" ) );
        case 28 :
            return ( string( "XA" ) );
        case 29 :
            return ( string( "RA" ) );
        case 30 :
            return ( string( "YA" ) );
        case 31 :
            return ( string( "XC" ) );
        case 32 :
            return ( string( "RC" ) );
        case 33 :
            return ( string( "YC" ) );
        case 34 :
            return ( string( "XG" ) );
        case 35 :
            return ( string( "RG" ) );
        case 36 :
            return ( string( "YG" ) );
        case 37 :
            return ( string( "XU" ) );
        case 38 :
            return ( string( "RU" ) );
        case 39 :
            return ( string( "YU" ) );
        case 40 :
            return ( string( "XX" ) );
        case 41 :
            return ( string( "XR" ) );
        case 42 :
            return ( string( "XY" ) );
        case 43 :
            return ( string( "YX" ) );
        case 44 :
            return ( string( "YR" ) );
        case 45 :
            return ( string( "YY" ) );
        case 46 :
            return ( string( "RX" ) );
        case 47 :
            return ( string( "RR" ) );
        case 48 :
            return ( string( "RY" ) );

            // "gap" base-pair
        case 49 :
            return ( string( "-A" ) );
        case 50 :
            return ( string( "-G" ) );
        case 51 :
            return ( string( "-C" ) );
        case 52 :
            return ( string( "-U" ) );
        case 53 :
            return ( string( "-X" ) );
        case 54 :
            return ( string( "-R" ) );
        case 55 :
            return ( string( "-Y" ) );
        case 56 :
            return ( string( "A-" ) );
        case 57 :
            return ( string( "G-" ) );
        case 58 :
            return ( string( "C-" ) );
        case 59 :
            return ( string( "U-" ) );
        case 60 :
            return ( string( "X-" ) );
        case 61 :
            return ( string( "R-" ) );
        case 62 :
            return ( string( "Y-" ) );
        case 63 :
            return ( string( "--" ) );
        default :
            return ( string( "ZZ" ) ); // unknown base-pair
    }
}

double RnaModel::probability( int obase, int sbase, double time, int category,
  int symbolCategory )  const {
    assert(symbolCategory==0);


    if ( ( invariantCategory == 1 ) && ( category == 0 ) ) {
        if ( obase == sbase ) {
            return 1.0;
        }
        else {
            return 0.0;
        }
    }

    int cat = category - invariantCategory;
    double probability = 0.0;
    for ( int k = 0; k < matrixSize; ++k )
        probability += ( ieigenMatrix[cat] ( obase, k ) *
        eigenMatrix[cat] ( k, sbase ) * exp( eigenValues[cat] [k] * time ) );

    // Just make sure
    //assert ( ( probability >= -0.000001 ) && ( probability <= 1.000001 ) );
    probability = MIN( probability, 1.0 );
    probability = MAX( probability, 0.0 );

    return ( probability );
}


double RnaModel::diffProbability( int obase, int sbase, double time, int category,
  int symbolCategory )  const {
    assert(symbolCategory==0);

    if ( ( invariantCategory == 1 ) && ( category == 0 ) ) {
        return 0.0;
    }
    int cat = category - invariantCategory;
    double diffProbability = 0.0;
    for ( int k = 0; k < matrixSize; ++k ) {
        diffProbability += ( eigenValues[cat] [k] *
        ieigenMatrix[cat] ( obase, k ) * eigenMatrix[cat] ( k, sbase ) *
        exp( eigenValues[cat] [k] * time ) );
    }
    assert( !isnan( diffProbability ) );
    return ( diffProbability );
}



double RnaModel::secondDiffProbability( int obase, int sbase,
     double time, int category, int symbolCategory )  const {
    assert(symbolCategory==0);
    
    if ( ( invariantCategory == 1 ) && ( category == 0 ) ){
        return 0.0;
    }
    int cat = category - invariantCategory;
    double diff2Probability = 0.0;
    for ( int k = 0; k < matrixSize; ++k ) {
        diff2Probability += ( pow( eigenValues[cat] [k], 2.0 ) *
            ieigenMatrix[cat] ( obase, k ) * eigenMatrix[cat] ( k, sbase ) *
            exp( eigenValues[cat] [k] * time ) );
    }
    assert( !isnan( diff2Probability ) );
    return ( diff2Probability );
}

// print the parameters in readable format
void RnaModel::printParameters( ostream & outputStream ) const {
  
    //print common parameters
    MatrixModel::printParameters( outputStream );

    outputStream.setf(ios::fixed);
    outputStream << setprecision(5);
    for ( int freqCat = 0; freqCat < numberFrequenciesCategories; ++freqCat ) {
        if (numberFrequenciesCategories>=2){
            outputStream << "set of frequencies " << freqCat + 1 << ":" << endl;
        }
        else{
            outputStream << "frequencies:" << endl;
        }
        for ( int i = 0; i < getNumberStates(); ++i ) {
            outputStream << "f[" << getState( i ) << "] = " <<
            frequencies[freqCat] [i] << "  ";
        }
        outputStream << endl;
    }
    outputStream << endl;

    for ( int rateMat = 0 ; rateMat < numberRatesMatrices; ++rateMat ) {
        if (numberRatesMatrices>=2){
            outputStream << "set of rate ratios " << rateMat + 1 << ":" << endl;
        }
        else{
            outputStream << "rate ratios:" << endl;
        }
        outputStream << "    ";
        for ( int i = 0; i < getNumberStates(); ++i ) {
            outputStream << "     " << getState( i ) << "     ";
        }
        outputStream << endl;
        for ( int i = 0; i < getNumberStates(); ++i ) {
            outputStream << ' ' << getState( i ) << ' ';
            for ( int j = 0; j < getNumberStates(); ++j ) {
                int index = matrixIndex( i, j);
                switch ( index ) {
                    //if diagonal
                    case -3 : outputStream << "   -------  ";
                              break;
                    //if reference
                    case -2 : outputStream << "   1.00000  ";
                              break;
                    //if null
                    case -1 : outputStream << "   XXXXXXX  ";
                              break;
                    //otherwise
                    default : outputStream << "  " << setw( 8 ) <<
                                  rateRatios [rateMat] [index] << "  ";
                              break;
                }
            }
            outputStream << endl; //end line
        }
        outputStream << endl; //end rate ratios matrix
    }
    
    if ( (numberRatesMatrices>=2) || (numberFrequenciesCategories>=2) ){
        for ( int gammaCat = 0; gammaCat < MAX(numberGammaCategories,1); ++ gammaCat ) {
            if ( numberGammaCategories>1 ){
                outputStream << "transition rate matrix for the gamma category "
	                         << gammaCat + 1;
                if (numberFrequenciesCategories>=2){
                    outputStream << ", frequencies set " <<
                         frequencyCategory[gammaCat + invariantCategory] + 1;
                }
                if (numberRatesMatrices>=2){
                    outputStream << ", rate ratios set "
                                 << matrixCategory[gammaCat] + 1;
                }
                outputStream << ":" << endl;
            }
            else{
                outputStream << "transition rate matrix";	
                if (numberFrequenciesCategories>=2){
                    outputStream << ", frequencies set " <<
                         frequencyCategory[invariantCategory] + 1;
                }
                outputStream << ":" << endl;
            }
            outputStream << "    ";
            for ( int i = 0; i < getNumberStates(); ++i ) {
                outputStream << "      " << getState( i ) << "     ";
            }
            outputStream << endl;
            for ( int i = 0; i < getNumberStates(); ++i ) {
                outputStream << ' ' << getState( i ) << ' ';
                for ( int j = 0; j < getNumberStates(); ++j ) {
                    outputStream << "  " << setw( 9 )
                             << rateMatrix[gammaCat] (i, j) << "  ";
                }
                outputStream << endl; //end line
            }
            outputStream << endl; //end transition rate matrix
        }
        outputStream << endl;
    }
    else{
        outputStream << "transition rate matrix (scaled so that "
                     << "the average substitution rate is 1.0):" << endl;
        double sum = 0.0;
        for ( int i = 0; i < getNumberStates() - 1; ++i ) {
            for ( int j = i + 1; j < getNumberStates(); ++j ) {
                int index = matrixIndex(i, j);
                switch ( index ) {
                    //if reference
                    case -2 : sum += ( 2.0 * frequencies[0] [i] *
                                       frequencies[0] [j] );
                              break;
                    //if null
                    case -1 : case -3 : break;
                    //otherwise (>=0)
                    default : sum += ( 2.0 * frequencies[0] [i] *
                                       frequencies[0] [j] *
                                       rateRatios[0] [index] );
                }
            }
        }
        outputStream << "    ";
        for ( int i = 0; i < getNumberStates(); ++i ) {
            outputStream << "      " << getState( i ) << "    ";
        }
        outputStream << endl;
        for ( int i = 0; i < getNumberStates(); ++i ) {
            outputStream << ' ' << getState( i ) << ' ';
            for ( int j = 0; j < getNumberStates(); ++j ) {
                int index = matrixIndex(i, j);
                switch (index){
                    //if diagonal
                    case -3 : outputStream << "  --------  ";
                              break;
                    //if reference
                    case -2 : outputStream << " " << setw( 9 )
                                           << frequencies[0] [j] / sum << "  ";
                              break;
                    //if null
                    case -1 : outputStream << " " << setw( 9 )
                                           << 0.0  << "  ";
                    //otherwise (>=0)
                    default : outputStream << " " << setw( 9 )
                                           << rateRatios[0] [index] * frequencies[0] [j] / sum << "  ";

                }
            }
            outputStream << endl; //end line
        }
        outputStream << endl; //end transition rate matrix
    }
    outputStream << endl;
    outputStream.unsetf(ios::fixed);
}

void RnaModel::setAverageMutationRate( double newAverageRate, int symbolCategory ){
    assert( symbolCategory == 0 );
    assert( newAverageRate > 0.0 );
    averageMutationRate = newAverageRate;
    double sum = 0.0, mr;
    int i, j;

    mr = averageMutationRate / ( 1.0 - proportionInvariantSites );

    if ( discreteGamma ) {
        double beta = alpha / mr;
        statlib::gamma_percentile_means( mutationRate, alpha, beta,
        numberGammaCategories );
        for ( int rateCategory = 0 ; rateCategory <
        numberGammaCategories; ++rateCategory ) {
            int freqCat = frequencyCategory[rateCategory + invariantCategory];
            int matrixCat = matrixCategory[rateCategory];
            sum = 0.0;
            for ( i = 0; i < getNumberStates() -1; ++i ) {
                for ( j = i + 1; j < getNumberStates(); ++j ) {
                    int index = matrixIndex(i, j);
                    switch ( index ) {
                        //if reference
                        case -2 : sum += ( 2.0 * frequencies[freqCat] [i] *
                                      frequencies[freqCat] [j] );
                                  break;
                        //if null
                        case -1 : break;
                        //otherwise (>=0)
                        default : sum += ( 2.0 * frequencies[freqCat] [i] *
                                      frequencies[freqCat] [j] *
                                      rateRatios[matrixCat] [index] );
                    }
                }
            }
            mutationRate[rateCategory] /= sum;
        }
    }
    else {
        sum = 0.0;
        for ( i = 0; i < getNumberStates() - 1; ++i ) {
            for ( j = i + 1; j < getNumberStates(); ++j ) {
                int index = matrixIndex(i, j);
                switch ( index ) {
                    //if reference
                    case -2 : sum += ( 2.0 * frequencies[0 + invariantCategory] [i] *
                                       frequencies[0 + invariantCategory] [j] );
                              break;
                    //if null
                    case -1 : break;
                    //otherwise (>=0)
                    default : sum += ( 2.0 * frequencies[0 + invariantCategory] [i] *
                                       frequencies[0 + invariantCategory] [j] *
                                       rateRatios[0] [index] );
                }
            }
        }
        mutationRate[0] = mr / sum;
    }
    setEigenMatrix();
}
