#include <algorithm>

/*
#include <dlfcn.h>
*/

#include <iostream>
#include <iomanip>

#include <ctype.h>
#include <functional>


#include "PatternDesign/Singleton.h"
#include "Util/randombox.h"

#include "Util/FileParser.h"
#include "Util/ParametersSet.h"

#include "Models/Model.h"
#include "Util/ModelFactory.h"

#include "Tree/InferenceTree.h"

#include "Util/Optimise.h"

class SequenceTable;

using namespace std;

int main( int argc, char* argv[]){

    ofstream outputTreeFile;
    ofstream outputModelFile;
    ofstream outputFile;
  
    if ( argc != 2 ) {
        cerr << "usage : " << argv[0] << " control_file" << endl;
        exit( 1 );
    }
    // Create a FileParser from the control-file
    FileParser * fileParser = new FileParser( argv[1] );

    // Retrieve the control-file parameters
    ParametersSet parameters = fileParser->retrieveParametersSet();

    delete fileParser;
    
    string outgroup = parameters( "DATAFILE" ).stringParameter( "Outgroup" );
    
    /** ************************************************************************
     * Data File Initialization
     ************************************************************************ */

    // Read in molecular data and remove unknown characters from the data matrix
    string dataFile = parameters( "DATAFILE" ).stringParameter( "Data file" );
    SequenceTable * table = new SequenceTable( parameters( "DATAFILE" ) );

    /** ************************************************************************
     * Model Initialization
     ************************************************************************ */
    Singleton < ModelFactory > & modelFactory = Singleton < ModelFactory >::instance();
    Model* model = modelFactory.create( parameters( "MODEL" ).stringParameter( "Model" ),
         parameters( "MODEL" ) );
    if ( parameters.find("Starting model parameters file") != parameters.end() ){
        model->initialisation( NULL );    
        // Create a FileParser from the model parameters-file
        FileParser * modelFileParser = new FileParser( parameters.stringParameter("Starting model parameters file") );
        // Retrieve the model parameters
        ParametersSet modelParameters = modelFileParser->retrieveParametersSet();
        //initialise the model with them
        model->setModelParameters(modelParameters);
        delete modelFileParser;
    }
    else{
        model->initialisation( table );
    }
    
    /** ************************************************************************
     * Tree Initialization
     ************************************************************************ */
    string userTreeFileName = parameters.stringParameter( "Tree file" );
    ifstream userTreeFile( userTreeFileName.c_str() );
    string stringTree;
    FileParser::readTree(userTreeFile, stringTree);
    if ( stringTree.length() == 0 ){
        cerr << "Error while reading the tree in the file : "
             << userTreeFileName << endl;
        exit(1);
    }
    cout << stringTree << endl;
    InferenceTree* tree = new InferenceTree( stringTree );
    if ( tree->isRooted() ) {
        tree->unroot();
    }
    tree->standardizeTree( outgroup );    
    tree->loadDataAndModel(table, model);

       
    /** ************************************************************************
     * Output File Initialization
     ************************************************************************ */
    string outputFileName = parameters.stringParameter( "Output file" );
    FileParser::confirmOpenFile( outputTreeFile,
                                 outputFileName + string(".tree") );
    FileParser::confirmOpenFile( outputModelFile,
                                 outputFileName + string(".model") );
    FileParser::confirmOpenFile( outputFile,
                                 outputFileName + string(".output") );
    
                            
    Singleton<randombox>& randBox = Singleton<randombox>::instance();
    randBox.setSeed( parameters.intParameter( "Random seed" ) );
        
    vector<double> branchLengths;
    tree->getBranchVector( branchLengths );
    for ( unsigned int j = 0; j < branchLengths.size(); ++j ){
        if (branchLengths[j]<=0.0){
            branchLengths[j] = 0.7 * randBox.ran();
        }
    }
    tree->setBranchVector( branchLengths );
    
        
    cout.setf(ios::fixed);
    cout << setprecision(4);
    cout << "initial likelihood = " << tree->loglikelihood() << endl;
    
    
    Optimise::optimiseQuasiNewton( *tree, *model, branchLengths,
                                   true, 1e-6, true );
    cout << endl;

    
    //save the model into its file
    outputModelFile.setf(ios::fixed);
    model->getModelParameters().saveToFile( outputModelFile );
    outputModelFile.close();
    //save the tree into its file
    outputTreeFile.setf(ios::fixed);
    outputTreeFile << tree->toString(true) << ";" << endl;
    outputTreeFile.close();
    
    model->printParameters( outputFile );
    model->printParameters( cout );  
      
    cout.setf(ios::fixed);
    outputFile.setf(ios::fixed);
    outputFile << tree->toString(true) << ";" << endl;
    cout << tree->toString(true) << ";" << endl;
    //4 digits precision for the likelihood
    outputFile << setprecision(4);
    cout << setprecision(4);
    cout << "maxlikelihood = " << tree->loglikelihood() << endl;
    outputFile << "maxlikelihood = " << tree->loglikelihood() << endl;
    outputFile.close();    
    delete table;
    delete model;
    delete tree;
}  
