#include "Sequence/SequenceTable.h"

#include "Models/Model.h"
#include "Util/ParametersSet.h"

#include "FileParser.h"

#include <algorithm>
#include <functional>
#include <stack>
#include <iostream>

using namespace std;

// The following function returns <true> if all the characters in the string

void SequenceTable::processStructure() {

    //create a stack to store the positions of '(' in the pairs list
    stack<int> st;

    //read the RNA mask 'structure' and build the 'pairs' list.
    for (unsigned int position = 0; position != structure.size(); ++position ) {
        if ( structure[position] == '(' ) {
            st.push(pairs.size());
	    pairs.push_back( pair < int, int > ( position, 0 ) );
        }
        //if ')' pop the corresponding '(' position in pairs from the stack and
	// complete the pair in the list
        if ( structure[position] == ')' ) {
            if ( st.empty() ) {
                cerr << "structure error at position " << ( position + 1 ) << endl;
                exit( 1 );
            }
            pairs[st.top()].second = position;
            st.pop();
        }
    }
    if ( !st.empty() ) {
        cerr << "Unbalanced pairing mask" << endl;
        while ( !st.empty() ) {
            cerr << "No matching pair for the open bracket at position " <<
            pairs[st.top()].first << endl;
            st.pop();
        }
        exit( 1 );
    }
}

SequenceTable::SequenceTable() {
    assert(0);
}

SequenceTable::SequenceTable( ParametersSet& sequenceParameters ) {

    ifstream ff;
    string filename = sequenceParameters.stringParameter( "Data file" );

    // Attempt to open the file
    ff.open( filename.c_str() );
    if ( !ff ) {
        cerr << "Unable to open file " << filename << " for reading" << endl;
        exit( 1 );
    }
    init( ff, sequenceParameters );
}

void SequenceTable::modelAttribution( bool userModelAttribution,
                                      stringstream& input ){
    //fill the user's model attribution vector if provided
    if ( userModelAttribution ){
        for ( unsigned int i = 0; i < initialModelAttrib.size(); ++i ) {
            input >> ws;
            input >> initialModelAttrib[i];
            if (initialModelAttrib[i] < 1) {
                cerr << "Error for the class of the element "
                     << i << " : " << initialModelAttrib[i] << endl;
                exit(1);
            }
        }
        //check against the structure if provided
        if ( structure.size() != 0 ){
            for ( vector< pair<int,int>  >::iterator iter = pairs.begin();
                  iter != pairs.end(); ++iter ){
                if ( initialModelAttrib[(*iter).first] != initialModelAttrib[(*iter).second] ){
                    cerr << "class error for the pair (" << (*iter).first
                         << "," << (*iter).second << ")" << endl;
                    cerr << "model 1 = " << initialModelAttrib[(*iter).first]
                         << ", model 2 = "
                         << initialModelAttrib[(*iter).second] << endl;
                }
            }
        }
    }
    //no attribution provided, automatic selection
    else{
        if ( type == "MIXED" ){
            //unpaired site for the first model
            for ( unsigned int i = 0; i < initialModelAttrib.size(); ++i){
                initialModelAttrib[i] = 1;
            }
            //paired site for the second model
            for ( vector< pair<int,int>  >::iterator i = pairs.begin();
                  i != pairs.end(); ++i ){
                initialModelAttrib[(*i).first] = 2;
                initialModelAttrib[(*i).second] = 2;
            }
        }
        //only one type of DNA
        else{
            for ( unsigned int i = 0; i < initialModelAttrib.size(); ++i){
                initialModelAttrib[i] = 1;
            }
        }
    }
}


void SequenceTable::init( ifstream & inputFile, ParametersSet& sequenceParameters ) {

    int numberSpecies = 0;
    int initialLength = 0;
    bool interleaved = sequenceParameters.boolParameter("Interleaved data file");

    stringstream input;
    FileParser::skipComments( inputFile, input );

    // Read in header information
    // number of species, length of sequences, type ("RNA" or "DNA")
    input >> ws;
    input >> numberSpecies;
    input >> ws;
    input >> initialLength;
    input >> ws;
    input >> type;

    // Read in structural information i.e. pairing mask for RNA base
    // pairs if data is in RNA format
    if ( ( type == "RNA" ) || (type == "MIXED" ) ){
        char ch;
        for ( int i = 0; i < initialLength; ++i ) {
            input >> ws;
            input >> ch;
            if ( (ch!='(') && (ch!=')') && (ch!='.') && (ch!='-') ){
                cerr << "Unrecognized symbol '" << structure[i]
                << "' in the structure at position " << i << endl;
                exit( 1 );
            }
            structure.push_back(ch);
        }
        // Parse and verify the structure, find the 'nbPairedSites' value
        // and initialise the 'rnaPairs' vector
        processStructure();
    }
    else {
        assert( type == "DNA" );
    }

    initialSequences.resize( numberSpecies, initialLength );
    initialModelAttrib.resize( initialLength );

    
    //fill the initialSequences structure with the sequences in the datafile
    if ( interleaved ) {
        constructInterleaved( input );
    }
    else {
        constructNonInterleaved( input );
    }

    bool userModelAttribution =
       ( ( sequenceParameters.find("Heterogeneous data models") !=
           sequenceParameters.end() ) &&
         sequenceParameters.boolParameter("Heterogeneous data models") );

    modelAttribution(userModelAttribution, input );

    numberModels = *(max_element(initialModelAttrib.begin(),initialModelAttrib.end()));

    constructRawSequences();

    constructSequences();

}



// Constructs a tree given the data in the file is interleaved
void SequenceTable::constructInterleaved( stringstream & input ) {
    int block;
    char next;

    int numberSpecies = initialSequences.numberRows();
    int sequencesLength = initialSequences.numberColumns();
    
    int * point = new int[numberSpecies];

    // read the stream content for all the species (name then first sequence part)
    block = 0;
    for ( int i = 0; i < numberSpecies; ++i ) {
        //Skip white spaces
        input >> ws;
        string speciesName;
        // Read in species "name"
        input >> speciesName;
        species.push_back(speciesName);
        point[i] = 0;
        next = ' ';
        // Read in species sequence and store its length in point[numspecies]
        while ( ( next != '\n' ) && ( next != '\r' ) ) {
            input.get( next );
            if ( !isspace( next ) ) {
                initialSequences(i, point[i]) = next;
                ++( point[i] );
            }
        }
        // Check to see if the interleaved sequence is even i.e. all sequences
        // are the same length
        if ( point[i] != point[0] ) {
            cerr << "The sequence " << i + 1 <<
            " at the beginning of the file " << "is uneven" << endl;
            exit( 1 );
        }
    }

    //Proceed through the blocks of interleaved sequences
    while ( ( point[0] < sequencesLength ) && !input.eof() ) {
        ++block;
        for ( int i = 0; i < numberSpecies; i++ ) {
            //Skip white spaces
            input >> ws;
            next = ' ';
            // Read in species sequence and store its length in point[numspecies]
            while ( ( next != '\n' ) && ( next != '\r' ) ) {
                if( input.eof() ){
                    cerr << "Unexpected end of sequence file at block " << block << endl;
                    exit(1);
                }
                input.get( next );
                if ( !isspace( next ) ) {
                    initialSequences(i, point[i]) = next;
                    ++( point[i] );
                }
            }
            // Check to see if the interleaved sequence is even i.e. all sequences
            // are the same length
            if ( point[i] != point[0] ) {
                cerr << "The interleaved sequence block " << block <<
                " is uneven" << endl;
                exit( 1 );
            }
        }
    }

    //free the point structure
    delete[] point;
}

// Constructs a tree given the data in the file is non-interleaved
void SequenceTable::constructNonInterleaved( stringstream & input ) {

    int numberSpecies = initialSequences.numberRows();
    int sequencesLength = initialSequences.numberColumns();

    // Read in species and sequences
    for ( int i = 0; i < numberSpecies; ++i ) {
        // Skip white spaces
        input >> ws;
        string speciesName;
        // Read in species "name"
        input >> speciesName;
        species.push_back(speciesName);
        // Read in sequences
        for ( int j = 0; j < sequencesLength; ++j ) {
            input >> ws; // Skip white spaces
            if( input.eof() ){
                cerr << "Unexpected end of sequence file when parsing species "
                     << speciesName << endl;
                exit(1);
            }
            input >> initialSequences(i, j);
        }
    }
}

void SequenceTable::constructRawSequences(){
    rawSequences.resize( initialSequences.numberRows(),
                         initialSequences.numberColumns() - pairs.size() );
    //we will apply the RNA mask to initialModelAttrib too
    modelAttrib.resize(initialSequences.numberColumns() - pairs.size());

    initial2Raw.resize( initialSequences.numberColumns() );
    raw2Initial.resize( initialSequences.numberColumns() - pairs.size() );

    // if a structure is not provided symbols and characters are equal
    if ( structure.size() == 0 ){
        for (unsigned int j = 0; j < initialSequences.numberColumns(); ++j){
            initial2Raw[j] = j;
            raw2Initial[j].resize(1);
            raw2Initial[j][0] = j;
            for (unsigned int i = 0; i < initialSequences.numberRows(); ++i){
                rawSequences(i,j) = initialSequences(i,j);
            }
            modelAttrib[j] = initialModelAttrib[j]-1;
        }
    }
    //otherwise apply the RNA mask
    else{
        vector< pair<int,int> >::iterator iter = pairs.begin();
        stack<unsigned int> treated;
        int currentCol = 0;
        for (unsigned int j = 0; j < initialSequences.numberColumns(); ++j){
            //if the nucleotid must be treated
            if ( (treated.empty()) || ( j != treated.top() ) ){
                if ((*iter).first != (int)j){
                    initial2Raw[j] = currentCol;
                    raw2Initial[currentCol].resize(1);
                    raw2Initial[currentCol][0] = j;
                    for (unsigned int i = 0; i < initialSequences.numberRows(); ++i){
                        rawSequences(i,currentCol) = initialSequences(i,j);
                    }
                    modelAttrib[currentCol] = initialModelAttrib[j]-1;
                }
                else{
                    for (unsigned int i = 0; i < initialSequences.numberRows(); ++i){
                        rawSequences(i,currentCol).resize(2);
                        rawSequences(i,currentCol)[0] = initialSequences(i,j);
                        rawSequences(i,currentCol)[1] = initialSequences(i,(*iter).second);
                    }
                    initial2Raw[j] = currentCol;                    
                    initial2Raw[(*iter).second] = currentCol;
                    raw2Initial[currentCol].resize(2);
                    raw2Initial[currentCol][0] = j;
                    raw2Initial[currentCol][1] = (*iter).second;
                    modelAttrib[currentCol] = initialModelAttrib[j]-1;
                   //double-check structure vs. model attribution
                   //already done before
                    assert(modelAttrib[currentCol] ==
                              initialModelAttrib[(*iter).second]-1);
                    treated.push((*iter).second);
                    ++iter;
                }
                ++currentCol;
            }
            //if already treated, remove it from the stack
            else{
                treated.pop();
            }
        }
        assert( treated.empty() );
    }
}

void SequenceTable::constructSequences(){


    sequences.resize( numberModels );
    invariantBases.resize( numberModels );

    bool invar;
    unsigned int speciesId;

    raw2Final.resize(rawSequences.numberColumns());
    final2Raw.resize(numberModels);
    invariant2Raw.resize(numberModels);
    for ( unsigned int i = 0; i < rawSequences.numberColumns() ; ++i ){
        invar = true;
        speciesId = 0;
        while( invar && ( speciesId<species.size() ) ){
            invar = ( rawSequences(speciesId, i) == rawSequences(0, i) );
            ++speciesId;
        }
        if (!invar){
            raw2Final[i] = pair<unsigned int, int>( modelAttrib[i],
                                        (int)final2Raw[modelAttrib[i]].size() );
            final2Raw[modelAttrib[i]].push_back(i);
        }
        //new invariant symbol
        else{
            unsigned int j = 0;
            bool found = false;
            while ( ( j < invariantBases[modelAttrib[i]].size() ) && !found ){
                if ( invariantBases[modelAttrib[i]][j].first == rawSequences( 0, i ) ){
                    ++invariantBases[modelAttrib[i]][j].second;
                    invariant2Raw[modelAttrib[i]][j].push_back(i);
                    raw2Final[i] = pair<unsigned int, int>(modelAttrib[i],-(j+1));
                    found = true;
                }
                ++j;
            }
            if (!found){
                pair<string,int> newInvariant( rawSequences( 0, i ), 1);
                invariantBases[modelAttrib[i]].push_back(newInvariant);
                vector< unsigned int > vec(1);
                vec[0] = i;
                invariant2Raw[modelAttrib[i]].push_back( vec );                
                raw2Final[i] = pair<unsigned int, int>(modelAttrib[i],
                                  -(invariantBases[modelAttrib[i]].size()));
            }
        }
    }
    for ( unsigned int modelId = 0; modelId < numberModels; ++modelId ){
        sequences[modelId].resize(species.size(),final2Raw[modelId].size());
        for ( unsigned int i = 0; i < final2Raw[modelId].size() ; ++i ){
            for ( unsigned int j = 0; j < species.size(); ++j ){
                sequences[modelId](j, i) = rawSequences( j, final2Raw[modelId][i] );
            }
        }
    }
}

SequenceTable::~SequenceTable() {
}

int SequenceTable::findsequence( string label ) {

    string temp1, temp2;

    for_each<string::iterator, int(*)(int)> ( label.begin(), label.end(), tolower );

    for ( unsigned int i = 0; i < species.size(); ++i ) {
        temp2 = species[i];
        for_each<string::iterator, int(*)(int)>
            ( temp2.begin(), temp2.end(), tolower );
        if ( temp2 == label ) return i;
    }
    return ( -1 );
}

void SequenceTable::print( ostream & output ) {

    output << "sequence type : " << type << endl;
    output << "number of species : " << species.size() << endl;
    output << "number of model used : " << sequences.size() << endl;
    for ( unsigned int i = 0; i < species.size(); ++i ) {
        output << "species "<< i + 1 << " : " << species[i] << endl;
        for ( unsigned int j = 0; j < numberModels; ++j ) {
            output << "model "<< j+1 << ", " << sequences[j].numberColumns()
                   << " non-invariant symbols" << endl;
            for ( unsigned int k = 0; k < sequences[j].numberColumns(); ++k ) {
                output << sequences[j](i,k);
            }
            output << endl;
        }
        output << endl;
    }
    output << "invariants " << endl;
    for ( unsigned int j = 0; j < numberModels; ++j ) {
        output << "model "<< j+1 << " : ";
        for ( vector< pair<string, unsigned int> > ::iterator iter =
              invariantBases[j].begin(); iter != invariantBases[j].end();
              ++iter) {
            output << (*iter).first << "-" << (*iter).second << "  ";
        }
        output << endl;
    }
}

vector < vector< unsigned int > > SequenceTable::retrieveInitialNucleotides( int modelId,
               int symbolIndex ){
    vector < vector< unsigned int > > ret;
    if (symbolIndex >= 0){
        assert((unsigned int)symbolIndex < final2Raw[modelId].size());
        unsigned int rawIndex = final2Raw[modelId][symbolIndex];
        ret.push_back( raw2Initial[rawIndex] );
    }
    else {
        symbolIndex = -1-symbolIndex;
        assert((unsigned int)symbolIndex < invariantBases[modelId].size());
        for ( vector<unsigned int>::iterator invSymbolIterator =
                invariant2Raw[modelId][symbolIndex].begin();
              invSymbolIterator != invariant2Raw[modelId][symbolIndex].end();
              ++invSymbolIterator ){
            ret.push_back( raw2Initial[*invSymbolIterator] );
        }
    }        
    return ret;
}

void SequenceTable::printStats( ostream & output, vector<int> species, vector<int> sites ){

    assert (0) ;
    output << "PrintStats" << endl;
    species = species;
    sites=sites;
    
    /*
    map< string, int > res;
    int tot = 0;
    
    for ( unsigned int j = 0 ; j < sites.size(); ++j ){
        if ( sites[j] < ( sequencesLength - numberTypeInvariantBases ) ){
            for ( unsigned int i = 0 ; i < species.size(); ++i ){
	        map< string, int >::iterator iter = 
		    res.find(sequences[species[i]][sites[j]]);
		if (iter == res.end() ){
           res[sequences[species[i]][sites[j]]] = 1;
		}
		else{
		    (*iter).second += 1;
		}
                tot += 1;
            }
        }
        else{
            string search(sequences[species[0]][sites[j]]);
            bool found = false;
            int k = 0;
            while (!found){
                found = ( search == invariantBases[k] );
                assert (k < numberTypeInvariantBases);
                ++k;
            }
            --k;
	    map< string, int >::iterator iter = 
                res.find(sequences[species[0]][sites[j]]);
	    if (iter == res.end() ){
                res[sequences[species[0]][sites[j]]] = numberInvariantBases[k]*species.size();
            }
            else{
              (*iter).second += (numberInvariantBases[k]*species.size());
            }
            tot += (numberInvariantBases[k]*species.size());
        }
    }
    
    output << " nbspecies : " << species.size() << ", number sites : " <<
    tot/species.size() << endl;
    // convenient for now
    map< string, int >::iterator iterA = res.find("A");
    map< string, int >::iterator iterC = res.find("C");
    map< string, int >::iterator iterG = res.find("G");
    map< string, int >::iterator iterT = res.find("T");
    map< string, int >::iterator iterAU = res.find("AT");
    map< string, int >::iterator iterGC = res.find("GC");
    map< string, int >::iterator iterUA = res.find("TA");
    map< string, int >::iterator iterCG = res.find("CG");
    map< string, int >::iterator iterGU = res.find("GT");
    map< string, int >::iterator iterUG = res.find("TG");
    
    int MMnb = 0;
    for ( map< string, int >::iterator i = res.begin(); i != res.end(); ++i){
        if ( ( (*i).first.length() == 2 ) &&
	     ( i != iterCG ) && ( i != iterGC ) &&
	     ( i != iterAU ) && ( i != iterUA ) &&
	     ( i != iterUG ) && ( i != iterGU ) ){
	     MMnb += (*i).second; 
	}
    }
        output << "A" << " : " << (double)(( *iterA ).second) /
	     (double)(( *iterA ).second+( *iterC ).second+( *iterG ).second+(
	     *iterT ).second);
        output << ", C" << " : " << (double)(( *iterC ).second) /
	     (double)(( *iterA ).second+( *iterC ).second+( *iterG ).second+(
	     *iterT ).second);
        output << ", G" << " : " << (double)(( *iterG ).second) /
	     (double)(( *iterA ).second+( *iterC ).second+( *iterG ).second+(
	     *iterT ).second);
        output << ", T" << " : " << (double)(( *iterT ).second) /
	     (double)(( *iterA ).second+( *iterC ).second+( *iterG ).second+(
	     *iterT ).second);
	
	output << endl;

        output << "AU" << " : " << (double)(( *iterAU ).second) /
	     (double)(( *iterAU ).second+( *iterUA ).second+
	              ( *iterGU ).second+( *iterUG ).second+
	              ( *iterGC ).second+( *iterCG ).second+
		      MMnb);
        output << ", GU" << " : " << (double)(( *iterGU ).second) /
	     (double)(( *iterAU ).second+( *iterUA ).second+
	              ( *iterGU ).second+( *iterUG ).second+
	              ( *iterGC ).second+( *iterCG ).second+
		      MMnb);
        output << ", GC" << " : " << (double)(( *iterGC ).second) /
	     (double)(( *iterAU ).second+( *iterUA ).second+
	              ( *iterGU ).second+( *iterUG ).second+
	              ( *iterGC ).second+( *iterCG ).second+
		      MMnb);
        output << ", UA" << " : " << (double)(( *iterUA ).second) /
	     (double)(( *iterAU ).second+( *iterUA ).second+
	              ( *iterGU ).second+( *iterUG ).second+
	              ( *iterGC ).second+( *iterCG ).second+
		      MMnb);
        output << ", UG" << " : " << (double)(( *iterUG ).second) /
	     (double)(( *iterAU ).second+( *iterUA ).second+
	              ( *iterGU ).second+( *iterUG ).second+
	              ( *iterGC ).second+( *iterCG ).second+
		      MMnb);
        output << ", CG" << " : " << (double)(( *iterCG ).second) /
	     (double)(( *iterAU ).second+( *iterUA ).second+
	              ( *iterGU ).second+( *iterUG ).second+
	              ( *iterGC ).second+( *iterCG ).second+
		      MMnb);
        output << ", MM" << " : " << (double)(MMnb) /
	     (double)(( *iterAU ).second+( *iterUA ).second+
	              ( *iterGU ).second+( *iterUG ).second+
	              ( *iterGC ).second+( *iterCG ).second+
		      MMnb);
        output << endl << endl;
*/
}
