#include "Simulate.h"

#include <math.h>
#include <assert.h>

#include <iostream>
#include <iomanip>
#include <time.h>

#include <fstream>
#include <string>

// Molecular sequence table
#include "Sequence/SequenceTable.h"

// Phylogenetic Tree
#include "Tree/SimulationTree.h"
#include "Tree/InferenceTree.h"

// Statistics library
#include "Util/statlib.h"

// Util
#include "Util/array2D.h"
#include "Util/FileParser.h"
#include "Util/ParametersSet.h"

// model prototype
#include "Models/Model.h"
#include "Util/ModelFactory.h"


using namespace std;

// Creates a random tree and sets its outgroup
SimulationTree * Simulate::randomTree( unsigned int numberSpecies,
                                          double maxBranchLength ) {

    Singleton<randombox>& randBox = Singleton<randombox>::instance();
    
    //create the tree
    SimulationTree * tree = new SimulationTree();
    
    vector < double > branchLength;

    vector < int > speciesId( numberSpecies );

    string leafLabel[numberSpecies];

    char tmpString[10];    
    // Create all species names
    for ( unsigned int i = 0; i < numberSpecies; ++i ) {
        sprintf(tmpString,"Species%d", i );
        leafLabel[i] = tmpString;
        speciesId[i] = i;
    }

    // Create a tree by addition of species to randomly selected branches
    speciesId = statlib::random_permutation( speciesId );
    for ( unsigned int i = 0; i < 3; ++i ){
       tree->addLeaf( leafLabel[speciesId[i]], i );
    }
    for ( unsigned int i = 3; i < numberSpecies; ++i ){
        tree->addLeaf( leafLabel[speciesId[i]],
        ( int )( floor( randBox.ran() * tree->getNumberBranches() ) ) );
    }

    tree->getBranchVector( branchLength );
    for ( unsigned int j = 0; j < branchLength.size(); ++j ){
       branchLength[j] = maxBranchLength * randBox.ran();
    }

    tree->setBranchVector( branchLength );
    tree->standardizeTree( "Species1" );
    tree->setBranchLength(0,maxBranchLength);
    return tree;
}


int Simulate::run( int argc, char * argv[] ) {
    if (argc != 2){
        cerr << "usage: " <<  argv[0] << " simulate_control_file" << endl;
	exit(1);
    }
    
    // Retrieve the control-file parameters
    FileParser * fileParser = new FileParser( argv[1] );
    ParametersSet parameters = fileParser->retrieveParametersSet();
    delete fileParser;
    
    // substitution model creation
    Singleton < ModelFactory > & modelFactory = Singleton < ModelFactory >::instance();
    Model* model = modelFactory.create(
                      parameters( "MODEL" ).stringParameter( "Model" ),
                      parameters( "MODEL" ) );
    model->initialisation();
    
    //retrieve the number of models ( 1 or more if the mixed model is used )
    int numberModels = 1;
    if ( parameters( "MODEL" ).stringParameter( "Model" ) == "MIXED" ){
        numberModels = parameters( "MODEL" ).intParameter("Number of models");
    }

    /** *************************************************************************
     * simulate first use : retrieve the name of the model parameters with the
     * given construction parameter
     ************************************************************************** */
    if ( parameters.boolParameter("Retrieve the name of the model's parameters") ){
        ofstream modelParametersInstance;
        string fileName =
            parameters.stringParameter("Model parameters file");
        FileParser::confirmOpenFile( modelParametersInstance, fileName );
        (model->getModelParameters()).saveToFile(modelParametersInstance);
        modelParametersInstance.close();
        cout << "An instance of a model parameters file has been written,"
             << " file name : " << fileName << endl;
        exit(1);
    }

    /** *************************************************************************
     * simulate second use : simulate sequences
     ************************************************************************** */

    // random seed
    clock_t seed = parameters.intParameter( "Random seed" );
    
    Singleton < randombox > & rand = Singleton < randombox >::instance();
    rand.setSeed( seed );
    
    
    // initialise the model according to the model parameters file
    string modelParametersFileName =
        parameters.stringParameter("Model parameters file");
    fileParser = new FileParser( modelParametersFileName );
    ParametersSet modelParameters = fileParser->retrieveParametersSet();
    delete fileParser;
    
    model->setModelParameters( modelParameters );


    // Initialise the tree
    SimulationTree* tree;
    string treeFileName = parameters.stringParameter("Tree file");
    if ( parameters.boolParameter("Random tree") ){
        int numberSpecies = parameters.intParameter("Number of species");
        ofstream treeOutput;
        tree = randomTree( numberSpecies,
                       parameters.doubleParameter("Maximum branch length") );
        FileParser::confirmOpenFile( treeOutput, treeFileName );
        treeOutput << tree->toString( false ) << ";" << endl;
    }
    else{
        ifstream treeInput;
        treeInput.open( treeFileName.c_str() );
        if ( !treeInput ) {
            cerr << "The tree file " << treeFileName << " does not exist" << endl;
            exit( 1 );
        }
        string stree;
        char next;
        treeInput.get( next );
        cout << "Read tree from file" << endl;
        while ( next != ';' ){
            stree = stree + next;
            assert( !treeInput.eof() );
            treeInput.get( next );
        }
        treeInput.close();
        tree = new SimulationTree( stree );

        if ( tree->unroot() ) {
            cout << "The user-defined tree was not rooted" << endl;
            cout << "It has been converted into an unrooted tree" << endl;
        }
    }
    
    
    vector<int> length(numberModels);
    string structure[numberModels];
    
    char tmpString[60];
    
    for( int i = 0; i < numberModels; ++i ){
        sprintf(tmpString,"Number of symbols from class %d", i+1);
        length[i] = parameters.intParameter(tmpString);
        sprintf(tmpString,"Structure for the elements of class %d", i+1);
        if ( parameters.find(tmpString) != parameters.end() ){
            structure[i] = parameters.stringParameter(tmpString);
        }
        else{
            structure[i] = "";
        }
    }


    string sequencesFileName = parameters.stringParameter("Output file");
    ofstream sequencesOutput;
    FileParser::confirmOpenFile( sequencesOutput, sequencesFileName );


    tree->loadModel( model );
    
    tree->simulate( length );

    sequencesOutput << "#Model parameters file : " << modelParametersFileName << endl;
    sequencesOutput << "#Tree : " << endl;
    sequencesOutput << "#" << tree->toString( false ) << endl;
    
    
    // introduction to the file to be rewritten
    sequencesOutput << tree->getNumberLeaves() << " "
                    << parameters.intParameter(
   		                "Total length of the raw sequences") << " "
                    << parameters.stringParameter("Data file type") << endl;	    
    sequencesOutput << "# structure" << endl;
    for( int i = 0; i < numberModels; ++i ){
        for( int j = 0; j < length[i]; ++j ){
            sequencesOutput << structure[i];
        }
    }
    sequencesOutput << endl;
    tree->printSequenceTable( sequencesOutput );

    for( int i = 0; i < numberModels; ++i ){
        for( int j = 0; j < length[i]; ++j ){
            sequencesOutput << i+1 << " ";
            for( unsigned int k = 1; k < structure[i].length(); ++k){
                sequencesOutput << i+1 << " ";
            }
        }
        sequencesOutput << endl;
    }            
    cout << endl << "Sequences written in " << sequencesFileName << endl;
    
    InferenceTree inferenceTree( tree->toString(false) );
    
    ParametersSet sequencesFileParameter("DATAFILE", false);
    sequencesFileParameter["Data file"] = sequencesFileName;
    sequencesFileParameter["Interleaved data file"] = "no";
    sequencesFileParameter["Heterogeneous data models"] = "yes";
    SequenceTable* seqTable = new SequenceTable( sequencesFileParameter );
    
    inferenceTree.loadDataAndModel( seqTable, model );
    
    cout << tree->toString( false ) << endl;
    assert( inferenceTree.toString( false ) == tree->toString( false ) );
    inferenceTree.invalidateNodes();
    cout << "LnLikelihood = " << inferenceTree.loglikelihood() << endl;
    sequencesOutput << "#LnLikelihood = " << inferenceTree.loglikelihood() << endl;
    sequencesOutput.close();

    return (0);
}

int main( int argc, char * argv[] ) {
    Simulate sim;
    return sim.run( argc, argv );
}




