#include "Optimise.h"

#include <iostream>

#include "Tree/InferenceTree.h"

#include "Models/Model.h"

#define ITMAX 1000
#define ZEPS 1.0e-10
#define MAX(a,b) ((a) > (b) ? (a) : (b))
#define SQR(a) (a*a)
#define EPS 1.0e-8
#define TOLX 1.0e-8
#define ALF 1.0e-8
#define STPMAX 1.0
#define SIGN(n) (n>=0.0 ? 1.0 : -1.0)

// Find the vector of partial gradients of the model parameters
vector < double > Optimise::modelPartialGradients( InferenceTree & ptree,
vector < double > & branchesLength, Model & smodel,
vector < double > & modelParameters, double step ) {
    vector < double > changeVector ;
    vector < double > gradientVector( modelParameters.size() );
    vector < double > oldModelParameters;
    vector < double > branchesLengthSquared( branchesLength.size() ) ;
    vector < double > saveBranchLengths;
    double oldLikelihood;
    double newLikelihood;

    ptree.getBranchVector( saveBranchLengths );
    // In case there are no model parameters
    if ( modelParameters.size() == 0 )
        return gradientVector ;

    // initialise branchesLengthSquared
    for ( unsigned int i = 0 ; i < branchesLength.size() ; i++ )
        branchesLengthSquared[i] = ( branchesLength[i] * branchesLength[i] ) ;

    // store the old model parameters of the model
    smodel.getAllParameters( oldModelParameters );

    // update the model parameters with the new ones
    smodel.setAllParameters( modelParameters );

    //initialise changeVector with the provided parameters
    changeVector = modelParameters ;

    //invalidate the nodes to compute the new likelyhood of the tree
    ptree.invalidateNodes();


    //compute the new likelyhood of the tree with the provided branches
    //length
    oldLikelihood = ptree.loglikelihood( branchesLengthSquared ) ;
    // for each parameter
    for ( unsigned int i = 0 ; i < modelParameters.size() ; i++ ) {
        // modify the given parameter of the model by step
        changeVector[i] += step ;
        smodel.setAllParameters( changeVector );
        //and compute the new likelyhood
        ptree.invalidateNodes() ;
        newLikelihood = ptree.loglikelihood( branchesLengthSquared ) ;
        // put back the parameter to its old value
        changeVector[i] = modelParameters[i] ;
        gradientVector[i] = ( newLikelihood - oldLikelihood ) / step ;
#ifdef DEBUG
        if ( isnan( gradientVector[i] ) ) {
            smodel.printParameters( cerr ) ;
            cerr << "Model parameter = " << i << endl ;
            cerr << "Value           = " << modelParameters[i] << endl ;
            cerr << "Nlikelihood = " << newLikelihood << endl ;
            cerr << "Olikelihood = " << oldLikelihood << endl ;
            assert( 0 );
        }
#endif
        assert( !isnan( gradientVector[i] ) );
    }

    // restore original parameters
    smodel.setAllParameters( oldModelParameters ) ;

    ptree.getBranchVector( branchesLengthSquared );
    assert ( saveBranchLengths == branchesLengthSquared );
    
    //invalidate the nodes to be sure the current likelyhood will not be used
    ptree.invalidateNodes() ;
    return gradientVector;
}

double Optimise::loglikelihood( InferenceTree & ptree , Model & smodel ,
vector < double > & start, vector < double > & direction, double magnitude ) {

    vector < double > oldModelParameters;

    // Store old model parameters
    smodel.getAllParameters( oldModelParameters );

    vector < double > modelParameters( oldModelParameters.size() ) ;
    vector < double > branchesLength( ptree.getNumberBranches() );
    vector < double > branchesLengthSquared( ptree.getNumberBranches() );

    double ans ;


    // Determine new branch parameters
    for ( unsigned int i = 0 ; i < branchesLength.size() ; i++ ) {
        branchesLength[i] = start[i] + ( magnitude * direction[i] );
    }
    // Determine new model parameters
    for ( unsigned int i = branchesLength.size() ; i < start.size() ; i++ ) {
        modelParameters[i - branchesLength.size()] =
        start[i] + ( magnitude * direction[i] );
    }

    //if the model has some parameters
    if ( modelParameters.size() > 0 ) {
        // Set new model parameters
        smodel.setAllParameters( modelParameters ) ;
        ptree.invalidateNodes() ;
    }

    //compute branchesLengthSquared
    for ( unsigned int i = 0 ; i < branchesLength.size() ; i++ ) {
        branchesLengthSquared[i] = ( branchesLength[i] * branchesLength[i] );
    }

    // Find log-likelihood with new branches and model parameters
    ans = ptree.loglikelihood( branchesLengthSquared );

    // restore model parameters
    smodel.setAllParameters( oldModelParameters );

    //invalidate the nodes to be sure the current likelyhood will not be used
    ptree.invalidateNodes();
    // WARNING INVALIDATE LEAVES TOO ?

    return ( ans ) ;	
}


double Optimise::loglikelihood( InferenceTree & ptree , vector < double > & start,
vector < double > & direction, double magnitude ) {

    vector < double > branchesLength( ptree.getNumberBranches() );
    vector < double > branchesLengthSquared( ptree.getNumberBranches() );

    double newLoglikelyhood;

    // Determine new branch parameters and initialise branchesLengthSquared
    for ( unsigned int i = 0 ; i < branchesLength.size() ; i++ ) {
        branchesLength[i] = start[i] + ( magnitude * direction[i] );
        branchesLengthSquared[i] = branchesLength[i] * branchesLength[i];
    }
    // Find log-likelihood with new branches' length
    newLoglikelyhood = ptree.loglikelihood( branchesLengthSquared ) ;

    return ( newLoglikelyhood ) ;
}



// The approximate line search algorithm to be used with the quasi-newton optimizations
double Optimise::lineSearch( InferenceTree & tree, Model & smodel,
vector < double > & point, vector < double > & direction, vector < double > & gradient,
bool domodel ) {

    double fold , f;
    double slope = 0.0 , temp , test , alamin , alam , alam2 = 0.0 ,
    disc , tmplam = 0.0;
    double rhs1 = 0.0 , rhs2 = 0.0 , f2 = 0.0 , a = 0.0 , b = 0.0 ;

    
    // compute the slope
    slope = 0.0 ;
    for ( unsigned int i = 0 ; i < direction.size() ; i++ )
        slope += ( gradient[i] * direction[i] ) ;


#ifdef DEBUG
    if ( slope >= 0.0 ) {
        cerr << "Slope = " << slope << endl ;
    }
    assert ( slope < 0.0 );
#endif

    // Convergence testing
    // for each direction, compute direction[i]/point[i]
    // alamin = TOLX/max(direction[i]/point[i])
    test = 0.0 ;
    for ( unsigned int i = 0 ; i < point.size() ; i++ ) {
        temp = fabs( direction[i] ) / ( MAX( fabs( point[i] ), 1.0 ) );
        if ( temp > test )
            test = temp ;
    }
    alamin = TOLX / test ;
    alam = 1.0 ;


    // Initial value, compute the likelihood in point
    if ( domodel )
        fold = -loglikelihood( tree , smodel  , point , direction , 0.0 ) ;
    else
        fold = -loglikelihood( tree , point , direction , 0.0 ) ;

    for ( ; ; ) {
        // Try the full newton step first
        // compute the likelihood in point + alam*direction
        if ( domodel )
            f = -loglikelihood( tree , smodel  , point , direction , alam ) ;
        else
            f = -loglikelihood( tree , point , direction , alam ) ;

        // if alam becomes lower than the threshold : converged
        if ( alam < alamin ) {
            return ( 0.0 ) ;
        }
        // otherwise
        else {
            if ( f <= fold + ALF * alam * slope )
                return alam ;
            else {
                //initialisation for the first run through the loop
                if ( alam == 1.0 )
                    tmplam = -slope / ( 2.0 * ( f - fold - slope ) ) ;
                else {
                    rhs1 =   f - fold - ( alam * slope ) ;
                    rhs2 = f2 - fold - ( alam2 * slope );
                    a = ( rhs1 / ( alam * alam ) - rhs2 / ( alam2 * alam2 ) ) /
                    ( alam - alam2 ) ;
                    b = ( -alam2 * rhs1 / ( alam * alam ) + alam * rhs2 /
                    ( alam2 * alam2 ) ) / ( alam - alam2 ) ;
                    if ( a == 0.0 )
                        tmplam = -slope / ( 2.0 * b ) ;
                    else {
                        disc = b * b - 3.0 * a * slope ;
                        if ( disc < 0.0 )
                            tmplam = 0.5 * alam ;
                        else {
                            if ( b <= 0.0 )
                                tmplam = ( -b + sqrt( disc ) ) / ( 3.0 * a );
                            else
                                tmplam = -slope / ( b + sqrt( disc ) ) ;
                        }
                    }
                    if ( tmplam > ( 0.5 * alam ) )
                        tmplam = 0.5 * alam ;
                }
            }
            alam2 = alam ;
            f2 = f ;
            alam = MAX( tmplam, ( 0.1 * alam ) ) ;
        }
    } // End of optimization loop
} // End lineSearch

vector < double > Optimise::diffLoglikelihoodSqrt( InferenceTree & ptree ,
vector < double > & branchesLengthSqrt ) {

    vector < double > diff;

    vector < double > sqr( branchesLengthSqrt.size() ) ;

    for ( unsigned int i = 0 ; i < sqr.size() ; i++ ) {
        sqr[i] = ( branchesLengthSqrt[i] * branchesLengthSqrt[i] );
    }
    diff = ptree.diffloglikelihood( sqr ) ;
    for ( unsigned int i = 0 ; i < sqr.size() ; i++ ) {
        diff[i] *= (2.0 * branchesLengthSqrt[i]) ;
#ifdef DEBUG
        assert( !isnan( diff[i] ) ) ;
        assert( !isinf( diff[i] ) ) ;
#endif
    }
    return diff ;
}


// The quasi-newton (BGFS update) optimization function
vector < double > Optimise::optimiseQuasiNewton( InferenceTree & ptree ,
Model & smodel , vector < double > & bl , bool domodel , double tol ,
bool display ) {

    vector < double >  mvector ;
    smodel.getAllParameters( mvector );
    
  
    int branch_length = bl.size() , model_length = mvector.size() ;
    int np = ( domodel ? ( branch_length + model_length ) : branch_length ), iter = 0 ;
    array2D < double > H( np, np ) , update1( np, np ) , update2( np, np ) ,
    update3( np, np );
    vector < double > delta( np ) , omega( np ) , u( np ) ;


    vector < double > pvector( branch_length ) , spvector( branch_length ) ,
    tgvector( np ) , otgvector( np ) , mgrad_vector( model_length ) ,
    now( np ) , tempv( np );
    vector < double > start( np ) , gvector( branch_length ) , gradient( np ) ,
    direction( np ) ;


    double alpha , nlikelihood , test , tempx , den , stpmax , sum;

    double olikelihood = ptree.loglikelihood( bl ) ;

    double fac, fae , sumdg , sumxi , slope;


    nlikelihood = olikelihood ;


    for ( int i = 0 ; i < branch_length ; ++i ){
        pvector[i] = sqrt( bl[i] ) ;
    }

    // Initialise H to the identity matrix
    for ( int i = 0 ; i < np ; ++i ) {
        H( i, i ) = 1.0 ;
        for ( int j = ( i + 1 ) ; j < np ; ++j ){
            H( i, j ) = H( j, i ) = 0.0 ;
        }
    }


    // Find the vector of partial gradients w.r.t to SQRT(the tree branch lengths)
    gvector   = diffLoglikelihoodSqrt( ptree , pvector );

    // Find the vector of partial gradients w.r.t to the model parameters
    if ( ( domodel ) && ( model_length > 0 ) ){
        mgrad_vector = modelPartialGradients( ptree , pvector , smodel ,
        mvector, 1e-6 ) ;
    }


    // Combine the two gradient vectors if necessary
    for ( int i = 0 ; i < branch_length ; ++i ) {
        tgvector[i] = gvector[i];
        assert( !isnan( tgvector[i] ) ) ;
    }
    
    if ( ( domodel ) && ( model_length > 0 ) ) {
        for ( int i = branch_length; i < np ; ++i ) {
            tgvector[i] = mgrad_vector[i - branch_length] ;
            assert( !isnan( mgrad_vector[i - branch_length] ) ) ;
        }
    }

    // Calculate the maximum step size
    sum = 0.0 ;
    for ( int i = 0 ; i < branch_length ; ++i ){
        sum += ( pvector[i] * pvector[i] ) ;
    }
    
    if ( domodel ) {
        for ( int i = 0 ; i < model_length ; ++i ){
            sum += ( mvector[i] * mvector[i] ) ;
        }
    }
    stpmax = STPMAX * MAX( sqrt( sum ) , ( double )np ) ;


    do {
        iter++ ;
        olikelihood = nlikelihood ;
        otgvector = tgvector ; // Save the gradient vector

        // Determine search direction
        for ( int i = 0 ; i < np ; ++i ) {
            direction[i] = 0.0 ;
            for ( int j = 0 ; j < np ; ++j ) {
                assert( !isnan( H( i, j ) ) ) ;
                assert( !isnan( tgvector[j] ) ) ;
                assert( !isinf( H( i, j ) ) ) ;
                assert( !isinf( tgvector[j] ) ) ;
                direction[i] += ( H( i, j ) * tgvector[j] ) ; 		
            }
            if ( isnan( direction[i] ) ) {
                cerr << "Iteration   " << iter << endl ;
                cerr << "i = " << i << endl ;
            }
            assert( !isnan( direction[i] ) );
        }

        // Combine the two parameter vectors to get the starting "point"
        for ( int i = 0 ; i < branch_length ; ++i ){
            start[i] = pvector[i];
        }
        if ( ( domodel ) && ( model_length > 0 ) ) {
            for ( int i = branch_length; i < np ; ++i ) {
                start[i] = mvector[i - branch_length] ;
                assert( !isnan( start[i] ) ) ;
            }
        }

        // Scale the direction vector if necessary
        sum = 0.0 ;
        for ( int i = 0 ; i < np ; ++i ){
            sum += ( direction[i] * direction[i] ) ;
        }
        sum = sqrt( sum ) ;

        if ( sum > stpmax ) {
            for ( int i = 0 ; i < np ; ++i )
                direction[i] *= ( stpmax / sum ) ;
        }

        // Find the gradient at the starting "point"
        slope = 0.0 ;
        for ( int i = 0 ; i < np ; ++i ) {
            gradient[i] = -1.0 * tgvector[i] ; // negate the elements tgvector
            slope += ( gradient[i] * direction[i] ) ; // dot product
        }

        // Perform line search
        alpha = lineSearch( ptree , smodel , start , direction , gradient , domodel ) ;

        for ( int i = 0 ; i < branch_length ; ++i ) { // Calculate new branch parameter vector
            delta[i]   = ( alpha * direction[i] ) ;
            pvector[i] += delta[i] ;
        }

        if ( ( domodel ) && ( model_length > 0 ) ) {
            for ( int i = 0 ; i < model_length ; ++i ) { // Calculate new
                // model parameter vector
                delta[i + branch_length] =
                ( alpha * direction[i + branch_length] ) ;
                mvector[i] += delta[i + branch_length] ;
            }
        }

        for ( int i = 0 ; i < branch_length ; ++i ) {
            spvector[i] = ( pvector[i] * pvector[i] ) ;
        }

        // Set new values for the tree
        ptree.setBranchVector( spvector ) ;
        // Set new parameters for the model
        if ( domodel ) {
            smodel.setAllParameters( mvector );
        }
        // Calculate new loglikelihood
        nlikelihood = ptree.loglikelihood( spvector ) ;

        //cout << iter << ". " << ptree.toString(true) << " Ln Likelihood
        // = " << nlikelihood << endl ;
        //#ifdef DEBUG
        // Display resultant tree
        //cout << "Alpha = " << alpha << endl ;
        //cout << iter << ". " << ptree.toString(true) << " Ln Likelihood
        // = " << nlikelihood << endl ;
        if ( display )
            cout << "." << flush;
        //#endif
        // Progress Indicator
        //    cout << "." << flush ;

        // Find the vector of partial gradients w.r.t the tree branch parameters
        // cout << "Gradient (Branch) " << endl ;
        gvector      = diffLoglikelihoodSqrt( ptree , pvector ) ;

        // Find the vector of partial gradients w.r.t to the model parameters
        if ( domodel )
            mgrad_vector = modelPartialGradients( ptree , pvector ,
            smodel , mvector, 1e-6 ) ;
        // Combine the two gradient vectors if necessary
        for ( int i = 0 ; i < branch_length ; ++i ){
            tgvector[i] = gvector[i];
#ifdef DEBUG
            assert( !isnan( tgvector[i] ) ) ;
            assert( !isinf( tgvector[i] ) ) ;
#endif
        }
        
        if ( domodel ) {
            for ( int i = branch_length; i < np ; ++i ) {
                tgvector[i] = mgrad_vector[i - branch_length] ;
#ifdef DEBUG
                assert( !isnan( tgvector[i] ) ) ;
                assert( !isinf( tgvector[i] ) ) ;
#endif
            }
        }



        // Since tgvector is actually the negative of the gradient
        for ( int i = 0 ; i < np ; ++i ) {
            omega[i] = otgvector[i] - tgvector[i];
        }

        for ( int i = 0 ; i < np ; ++i ) {
            tempv[i] = 0.0 ;
            for ( int j = 0 ; j < np ; ++j ) {
                tempv[i] += ( H( i, j ) * omega[j] ) ;
            }
        }


        fac = fae = sumdg = sumxi = 0.0 ;
        for ( int i = 0 ; i < np ; ++i ) {
            fac   += ( omega[i] * delta[i] ) ;
            fae   += ( omega[i] * tempv[i] )  ;
            sumdg += ( omega[i] * omega[i] ) ;
            sumxi += ( delta[i] * delta[i] ) ;
        }


/************* Perform two types of convergence test  *******************/
        // First test for convergence in value
        for ( unsigned int i = 0 ; i < pvector.size() ; ++i ) {
            now[i] = pvector[i] ;
        }
        
        if ( domodel ) {
            for ( unsigned int i = pvector.size(); (int)i < np ; ++i ) {
                now[i] = mvector[i - pvector.size()] ;
            }
        }

        test = 0.0 ;
        for ( unsigned int i = 0 ; i < now.size() ; i++ ) {
            tempx = fabs( delta[i] ) / MAX( fabs( now[i] ) , 1.0 ) ;
            if ( tempx > test )
                test = tempx ;
        }

        if ( test < TOLX ) {
            return now ;
        }
        // end value convergence test

        // Now test for convergence in gradient
        test = 0.0 ;
        den = MAX( -nlikelihood, 1.0 ) ;
        for ( unsigned int i = 0 ; i < now.size() ; ++i ) {
            tempx = fabs( tgvector[i] ) * MAX( fabs( now[i] ), 1.0 ) / den ;
            if ( tempx > test )
                test = tempx ;
        }
        if ( test < tol )
            return now ;
        // end gradient convergence test




        if ( fac > sqrt( EPS * sumdg * sumxi ) ) {

            for ( int i = 0 ; i < np ; ++i ) {
                u[i] = ( delta[i] / fac ) - ( tempv[i] / fae ) ;
            }

            for ( int i = 0 ; i < np ; ++i ) {
                for ( int j = 0 ; j < np ; ++j ) {
                    update3( i, j ) = u[i] * u[j] * fae ;
                }
            }

            fac = 1.0 / fac ;
            fae = 1.0 / fae ;
            for ( int i = 0 ; i < np ; ++i ) {
                for ( int j = 0 ; j < np ; ++j ) {
                    update1( i, j ) = delta[i] * delta[j] * fac  ;
                }
            }

            for ( int i = 0 ; i < np ; ++i ) {
                for ( int j = 0 ; j < np ; ++j ) {
                    update2( i, j ) = tempv[i] * tempv[j] * fae ;
                }
            }
            

            // bfgs update
            for ( int i = 0 ; i < np ; i++ ) {
                for ( int j = 0 ; j < np ; j++ ) {
                    H( i, j ) = H( i, j ) + update1( i, j ) -
                    update2( i, j ) + update3( i, j );
                }
            }



#ifdef DEBUG
            for ( int i = 0 ; i < np ; i++ ) {
                for ( int j = 0 ; j < np ; j++ ) {
                    assert( !isnan( H( i, j ) ) );
                }
            }
#endif
        }
        #ifdef DEBUG
        for ( int i = 0 ; i < np ; i++ ) {
            assert( !isnan( tgvector[i] ) ) ;
            assert( !isinf( tgvector[i] ) ) ;
        }
        #endif

    }
    while ( iter < ITMAX ) ;
    return ( now ) ;
}




