#include "TopologyOptimise.h"

#include "PatternDesign/Singleton.h"
#include "Util/randombox.h"

#include "Tree/TreesSet.h"
#include "Tree/TreeModelValue.h"

#include "Tree/Cluster.h"
#include "Tree/ClustersSet.h"

#include <iostream>
#include <iomanip>


void TopologyOptimise::exhaustiveSearch( SequenceTable * sequenceTable,
                                         Model * model, bool optimiseModel,
                                         TreesSet* treesSet, ostream & out,
                                         ClustersSet& clusters ) {

    unsigned int numberTrees;
    int present;

    int next[clusters.size()] ;
         
    InferenceTree currentTree;
    InferenceTree bestTree;

    // vector of parameters
    vector < double > initialModelParameters;
    vector < double > currentModelParameters;
    vector < double > bestModelParameters;

    // branch vector
    vector < double > branchLengths;

    string stringTree;

    bool first = true ;
     
    double currentLikelihood = 0.0;
    double bestLikelihood = 0.0 ;


    numberTrees = 1;
    for ( unsigned int i = 4 ; i <= clusters.size() ; ++i ){
        numberTrees = numberTrees*( 2 * i - 5 );
    }

    // print to standard output
    cout.setf(ios::fixed);
    out.setf(ios::fixed);    
    cout << endl ;
    cout << "Processing " << numberTrees <<  " phylogenetic trees "
         << endl << endl ;
    out << endl ;
    out << "Processing " << numberTrees << " phylogenetic trees "
        << endl << endl ;

    // Load Data and Model
    currentTree.loadDataAndModel( sequenceTable, model ) ;    

    for ( ClustersSet::iterator i = clusters.begin();
                                   i != clusters.end(); ++i ){
        (*i).commonAncestor = Node::makenode( (*i).name, &currentTree );
    }

    //Create Initial tree
    currentTree.addClusteredNode( clusters[0].commonAncestor , 0 ) ;
    currentTree.addClusteredNode( clusters[1].commonAncestor , 1 ) ;
    currentTree.addClusteredNode( clusters[2].commonAncestor , 2 ) ;

    int topologyNumber = 0;
    present = 0;
    next[0] = 0;
    model->getAllParameters( initialModelParameters ) ;

    while ( ( ( next[0] < currentTree.getNumberBranches() ) ||
              ( present != 0 ) ) &&
            ( present >= 0 ) ){
        // Generate next candidate tree : from the initial
        // tree of 3 clusters add other clusters
        while ( present < ( (int)clusters.size() - 3 ) ) {
            // add the cluster to the tree (refused if inside another cluster)
            while ( !currentTree.addClusteredNode( clusters[present + 3].commonAncestor,
                                           next[present] ) ) {
                 ++(next[present]);
             }
             ++(next[present]);
             ++present;
             next[present] = 0 ;
         }

        if (optimiseModel){
            // no need to invalidate the nodes of the tree,
            // because we change the lengths too
            model->setAllParameters( initialModelParameters ) ;
        }
        //no need to invalidate, because we change the lengths
        currentTree.getBranchVector( branchLengths ) ;
        for ( unsigned int i = 0 ; i < branchLengths.size(); ++i ){
            branchLengths[i] = 0.1;	
        }
        currentTree.setBranchVector( branchLengths ) ;


        // Optimise new tree
        cout << "Optimising Topology " << ( topologyNumber + 1 ) << endl ;
        cout << "Topology = " << currentTree.toStringTopology( false ) << endl ;
        cout << "Progress = " << flush ;
        Optimise::optimiseQuasiNewton( currentTree , *model , branchLengths ,
                                       optimiseModel, 1e-6, true ) ;


        currentLikelihood = currentTree.loglikelihood();
        // If the optimised tree is better than the best tree so far
        // update the best tree and best loglikelihood
        if ( (first) || ( currentLikelihood > bestLikelihood ) ) {
            bestLikelihood = currentLikelihood;
            bestTree = currentTree ;
            model->getAllParameters( bestModelParameters ) ;
            first = false;
        }
        model->getAllParameters( currentModelParameters );
        stringTree = currentTree.toString( false ) ;
        currentTree.getBranchVector( branchLengths ) ;
        treesSet->insert( stringTree, branchLengths,
                         currentModelParameters, currentLikelihood ) ;

        // Print out the optimised tree and its loglikelihood
        cout << setprecision(4);
        out << setprecision(4);
        cout << "Maximum Lnlikelihood = " << currentLikelihood << endl;
        out  << ( topologyNumber + 1 ) << ". "
             << currentTree.toString( false )
             << " Lnlikelihood = " << currentLikelihood << endl;
        ++topologyNumber ;

        // Backtrack
        do {
            present-- ;
            currentTree.removeNode( clusters[present + 3].commonAncestor->toStringTopology() ) ;
        }
        while ( ( present > 0 ) &&
                ( next[present] >= currentTree.getNumberBranches() ) ) ;
    }
    cout.unsetf(ios::fixed);
    out.unsetf(ios::fixed);             
    model->setAllParameters( bestModelParameters ) ;
} // end search constrained


void TopologyOptimise::exhaustiveSearch( SequenceTable * sequenceTable,
                                         Model * model, bool optimiseModel,
                                         TreesSet* treesSet, ostream & out ) {

   int numberTrees;
   int present;
   int numberTaxa = sequenceTable->getNumberSpecies();
   int next[numberTaxa - 2] ;
         
   InferenceTree currentTree;
   InferenceTree bestTree;

     // vector of parameters
    vector < double > initialModelParameters;
    vector < double > currentModelParameters;
    vector < double > bestModelParameters;

    // branch vector
    vector < double > branchLengths;

    string stringTree;

    bool first = true ;
    double currentLikelihood = 0.0 , bestLikelihood = 0.0 ;


    // Load Data and Model
    currentTree.loadDataAndModel( sequenceTable, model ) ;

    // Create Initial Tree by adding the first 3 leaves to the empty tree
    currentTree.addLeaf( sequenceTable->species[0], 0 ) ;
    currentTree.addLeaf( sequenceTable->species[1], 1 ) ;
    currentTree.addLeaf( sequenceTable->species[2], 2 ) ;

    // calculate the number of possible tree topologies
    numberTrees = 1;
    for ( int i = 4 ; i <= numberTaxa ; ++i ){
        numberTrees = numberTrees*( 2 * i - 5 );
    }

     
    cout.setf(ios::fixed);
    out.setf(ios::fixed);    
    cout << "Processing " << numberTrees <<  " phylogenetic trees " << endl;
    out  << "Processing " << numberTrees <<  " phylogenetic trees " << endl;


    int topologyNumber = 0;
    present = 0 ;
    next[0] = 0 ;
    model->getAllParameters( initialModelParameters ) ;

    while ( ( ( next[0] < currentTree.getNumberBranches() ) ||
              ( present != 0 ) ) &&
            ( present >= 0 ) ){
        // Generate next candidate tree
        while ( present < ( numberTaxa - 3 ) ) {
            currentTree.addLeaf( sequenceTable->species[present + 3], next[present] ) ;
            ++next[present];
            ++present;
            next[present] = 0 ;
        }

        currentTree.getBranchVector( branchLengths ) ;
        for ( unsigned int i = 0 ; i < branchLengths.size(); ++i ){
            branchLengths[i] = 0.1;	
        }

        currentTree.setBranchVector( branchLengths ) ;
        model->setAllParameters( initialModelParameters ) ;

        currentTree.invalidateNodes() ;

        // Optimise new tree
        cout << "Optimising Topology " << ( topologyNumber + 1 ) << endl;
        cout << "Topology = " << currentTree.toStringTopology( false ) << endl;
        cout << "Progress = " << flush ;
        Optimise::optimiseQuasiNewton( currentTree, *model, branchLengths,
                                       optimiseModel, 1e-6, true ) ;

        // If the optimised tree is better than the best tree so far
        // update the best tree and best loglikelihood
        currentLikelihood = currentTree.loglikelihood();
        if ( ( first ) || ( currentLikelihood > bestLikelihood ) ){
            bestLikelihood = currentLikelihood;
            bestTree = currentTree;
            model->getAllParameters( bestModelParameters ) ;
            first = false;
        }
        model->getAllParameters( currentModelParameters ) ;
        stringTree = currentTree.toString( false );
        currentTree.getBranchVector( branchLengths ) ;
        treesSet->insert( stringTree, branchLengths,
                         currentModelParameters , currentLikelihood ) ;


        // Print out the optimised tree and its loglikelihood
        cout << endl; 
        cout << setprecision(4);
        out << setprecision(4);
        cout << currentTree.toString(false) << endl;
        cout << "Lnlikelihood = " << currentLikelihood << endl;
        out  << ( topologyNumber + 1 ) << ". "  <<
                currentTree.toStringTopology( false ) <<
            " Lnlikelihood = " << currentLikelihood << endl;
        ++topologyNumber;

        // Backtrack before the next tree
        do {
            --present;
            currentTree.removeLeaf( sequenceTable->species[present + 3] );
        }
        while ( ( present > 0 ) &&
                ( next[present] >= currentTree.getNumberBranches() ) );
    }


    model->setAllParameters( bestModelParameters );
    cout.unsetf(ios::fixed);
    out.unsetf(ios::fixed);    
}


double TopologyOptimise::buildTreeStepwise( SequenceTable * sequenceTable,
        Model * model, bool optimiseModel,
        TreesSet* treesSet, ostream & out ) {
    
    vector< double > modelParameters;
    vector< double > branchesLength;
    vector< double > bestModelParameters;
    vector< double > bestBranchesLength;
    InferenceTree currentTree;
    unsigned int numberTaxa = sequenceTable->getNumberSpecies();
    vector < int > order( numberTaxa ) , order2( numberTaxa );

    // Print to Screen
    cout.setf(ios::fixed);
    out.setf(ios::fixed);    
    cout << "Number of taxa  = " << numberTaxa << endl ;
    cout << "Sequence length = " << sequenceTable->getInitialLength() << endl ;
    cout << "Model Name      = " << model->getName() << endl ;

    // Print to output file
    out << "Number of taxa  = " << numberTaxa << endl ;
    out << "Sequence length = " << sequenceTable->getInitialLength() << endl ;
    out << "Model Name      = " << model->getName() << endl ;

    for ( unsigned int i = 0 ; i < numberTaxa ; ++i ) {
        order2[i] = i ;
    }

    order = statlib::random_permutation( order2 ) ;

    // print order addition order
    cout << "Order :  " ;
    out  << "Order :  " ;
    for ( unsigned int i = 0 ; i < numberTaxa ; ++i ) {
        if (i!=0){
            cout << ", ";
            out << ", ";
        }
        cout << sequenceTable->species[order[i]];
        out << sequenceTable->species[order[i]];
    }
    cout << endl;
    out  << endl;

    // Load Data and Model
    currentTree.loadDataAndModel( sequenceTable, model ) ;
    
    currentTree.addLeaf( sequenceTable->species[order[0]] , 0 ) ;
    currentTree.addLeaf( sequenceTable->species[order[1]] , 1 ) ;
    currentTree.addLeaf( sequenceTable->species[order[2]] , 2 ) ;
    cout << "Added : " << sequenceTable->species[order[0]] << endl ;
    cout << "Added : " << sequenceTable->species[order[1]] << endl ;
    cout << "Added : " << sequenceTable->species[order[2]] << endl ;
    out << "Added : " << sequenceTable->species[order[0]] << endl ;
    out << "Added : " << sequenceTable->species[order[1]] << endl ;
    out << "Added : " << sequenceTable->species[order[2]] << endl ;


    currentTree.getBranchVector( branchesLength ) ;
    // Set the branches to the same value
    for ( unsigned int branchId = 0 ; branchId < branchesLength.size();
          ++branchId ){
        branchesLength[branchId] = 0.1;
    }
    currentTree.setBranchVector( branchesLength );
    // Optimise the tree branches and (depending on choice) model parameters
    Optimise::optimiseQuasiNewton( currentTree, *model, branchesLength,
                                                optimiseModel, 1e-6, false ) ;
    model->getAllParameters( modelParameters );

    currentTree.getBranchVector( branchesLength ) ;


    for ( unsigned int i = 3 ; i < numberTaxa ; ++i ) {
      
        unsigned int bestPosition = 0;
        unsigned int numberPositions = currentTree.getNumberBranches() ;
        double currentLikelihood;
        double bestLikelihood = 1.0;
        
        vector<double> actualBranchesLength;
        
        // set starting model parameters (the best parameters for the tree
        // without the current number of taxa)
        if ( optimiseModel ){
             model->setAllParameters( modelParameters );
             //no need to invalidates the node because the lengths will change
        }
        // reset to starting branch lengths (the best lengths for the tree without the new taxa)
        currentTree.setBranchVector( branchesLength );
        cout << setprecision(4);
        out << setprecision(4);
        cout << "Tree : " << currentTree.toString( false ) << endl;
        cout << " LnLikelihood = " << currentTree.loglikelihood() << endl << endl ;
        out << "Tree : " << currentTree.toString( false ) << endl;
        out << " LnLikelihood = " << currentTree.loglikelihood() << endl << endl;
        
        // Add the new taxa to all the possible branches and find the optimal
        // branch to attach the taxa to
        for ( unsigned int j = 0 ; j < numberPositions ; ++j ) {
            //add the new leaf
            currentTree.addLeaf( sequenceTable->species[order[i]] , j ) ;

            currentTree.getBranchVector( actualBranchesLength );
            // Optimise the tree branches and (depending on choice) model parameters
            Optimise::optimiseQuasiNewton( currentTree, *model, actualBranchesLength,
                                                        optimiseModel, 1e-6, false ) ;
            cout << "." << flush;
            currentLikelihood = currentTree.loglikelihood() ;

            //if it is the first try or if the likelihood is better update the best values
            if ( ( currentLikelihood > bestLikelihood ) || ( j == 0 ) ){
                bestPosition = j;
                bestLikelihood = currentLikelihood;
                currentTree.getBranchVector( bestBranchesLength );
                if ( optimiseModel ){
                    model->getAllParameters( bestModelParameters );
                }
            }
            //backtrack to previous tree (without this taxa)
            currentTree.removeLeaf( sequenceTable->species[order[i]] );
            // reset to starting model parameters (the best parameters for the tree without the new taxa)
            if ( optimiseModel ){
                 model->setAllParameters( modelParameters );
                 //no need to invalidates the node because the lengths will change
            }
            // reset to starting branch lengths (the best lengths for the tree without the new taxa)
            currentTree.setBranchVector( branchesLength );
        }
        //once the best position is found update the initial values for the next taxa
        currentTree.addLeaf( sequenceTable->species[order[i]], bestPosition );
        if(optimiseModel){
            modelParameters = bestModelParameters;
        }
        branchesLength = bestBranchesLength;

        cout << endl << "Added : " << sequenceTable->species[order[i]] << endl ;
        out << endl << "Added : " << sequenceTable->species[order[i]] << endl ;
    }

    if ( optimiseModel ){
        model->setAllParameters( modelParameters ) ;
    }
    currentTree.setBranchVector( branchesLength );
    cout << setprecision(4);
    out << setprecision(4);
    cout << "Best Tree : " << currentTree.toString( false )  << endl;
    cout << "LnLikelihood = " << currentTree.loglikelihood() << endl << endl;
    out << "Best Tree : " << currentTree.toString( false )  << endl;
    out << "LnLikelihood = " << currentTree.loglikelihood() << endl << endl;

    if ( optimiseModel ){
        model->printParameters( cout ) ;
        model->printParameters( out ) ;
    }
    treesSet->insert( currentTree.toString( false ), branchesLength,
                      modelParameters , currentTree.loglikelihood() ) ;
    cout.unsetf(ios::fixed);
    out.unsetf(ios::fixed);    
    return ( currentTree.loglikelihood() ) ;
}


double TopologyOptimise::buildTreeStepwise( SequenceTable * sequenceTable,
        Model * model, bool optimiseModel,
        TreesSet* treesSet, ostream & out, ClustersSet& clusters ) {

      
    vector< double > modelParameters;
    vector< double > branchesLength;
    vector< double > bestModelParameters;
    vector< double > bestBranchesLength;

    InferenceTree currentTree;
    unsigned int numberClusters = clusters.size();
    vector < int > order( numberClusters ) , order2( numberClusters );

    // Print to Screen
    cout.setf(ios::fixed);
    out.setf(ios::fixed);    
    cout << "Number of clusters  = " << numberClusters   << endl ;
    cout << "Sequence length = " << sequenceTable->getInitialLength() << endl ;
    cout << "Model Name      = " << model->getName() << endl ;

    // Print to output file
    if ( & cout != & out ) {
        out << "Number of clusters  = " << numberClusters   << endl ;
        out << "Sequence length = " << sequenceTable->getInitialLength() << endl ;
        out << "Model Name      = " << model->getName() << endl ;
    }

    for ( unsigned int i = 0 ; i < numberClusters ; ++i ) {
        order2[i] = i ;
    }

    order = statlib::random_permutation( order2 ) ;

    // print order addition order
    cout << "Order :  " ;
    out  << "Order :  " ;
    for ( unsigned int i = 0 ; i < numberClusters ; ++i ) {
        if (i!=0){
            cout << ", ";
            out << ", ";
        }
        cout << clusters[order[i]].name;
        out << clusters[order[i]].name;
    }
    cout << endl;
    out  << endl;

    // Load Data and Model
    currentTree.loadDataAndModel( sequenceTable, model ) ;
    
    for ( ClustersSet::iterator i = clusters.begin();
                                   i != clusters.end(); ++i ){
        (*i).commonAncestor = Node::makenode( (*i).name, &currentTree );
    }
    
    //Create Initial tree
    currentTree.addClusteredNode( clusters[0].commonAncestor , 0 ) ;
    currentTree.addClusteredNode( clusters[1].commonAncestor , 1 ) ;
    currentTree.addClusteredNode( clusters[2].commonAncestor , 2 ) ;
    cout << "Added : " << clusters[order[0]].name << endl ;
    cout << "Added : " << clusters[order[1]].name << endl ;
    cout << "Added : " << clusters[order[2]].name << endl ;
    out << "Added : " << clusters[order[0]].name << endl ;
    out << "Added : " << clusters[order[1]].name << endl ;
    out << "Added : " << clusters[order[2]].name << endl ;


    currentTree.getBranchVector( branchesLength ) ;
    // Set the branches to the same value
    for ( unsigned int branchId = 0 ; branchId < branchesLength.size();
          ++branchId ){
        branchesLength[branchId] = 0.1;
    }
    currentTree.setBranchVector( branchesLength );
    
    // Optimise the tree branches and (depending on choice) model parameters
    Optimise::optimiseQuasiNewton( currentTree, *model, branchesLength,
                                                optimiseModel, 1e-6, false );
    model->getAllParameters( modelParameters );
    //load the initial value for the branch length from the optimised tree
    //with three clusters
    currentTree.getBranchVector( branchesLength );


    for ( unsigned int i = 3 ; i < numberClusters ; ++i ) {
      
        unsigned int bestPosition = 0;
        unsigned int numberPossiblePositions = currentTree.getNumberBranches() ;
        double currentLikelihood;
        double bestLikelihood = 1.0;
        
        vector< double > actualBranchesLength;
        
        // set starting model parameters (the best parameters for the tree
        // without the current number of taxa)
        if ( optimiseModel ){
             model->setAllParameters( modelParameters );
             //no need to invalidates the node because the lengths will change
        }
        // reset to starting branch lengths (the best lengths for the tree without the new taxa)
        currentTree.setBranchVector( branchesLength );
        cout << setprecision(4);
        out << setprecision(4);
        cout << "Tree : " << currentTree.toString( false ) << endl;
        cout << " LnLikelihood = " << currentTree.loglikelihood() << endl << endl ;
        out << "Tree : " << currentTree.toString( false ) << endl;
        out << " LnLikelihood = " << currentTree.loglikelihood() << endl << endl;
        
        //boolean used to initialise the best parameters/likelihood
        //with the first values found during the loop
        bool first = true;
        
        // Add the new clusters to all the possible branches and find the optimal
        // branch to attach the clusters to
        for ( unsigned int j = 0 ; j < numberPossiblePositions ; ++j ) {            
          
            // try to add the new cluster and react if it works (if the branch
            // is not inside another cluster)            
            if ( currentTree.addClusteredNode( clusters[order[i]].commonAncestor , j ) ){

                //initialise the branches of the cluster (it might change
                //the value of null internal branch but it does not matter)
                currentTree.getBranchVector( actualBranchesLength );
                for ( unsigned int branchId = 0 ;
                      branchId < actualBranchesLength.size(); ++branchId ){
                    if ( actualBranchesLength[branchId] <= 0.0 ){
                        actualBranchesLength[branchId] = 0.1;
                    }
                }
                currentTree.setBranchVector( actualBranchesLength );
               
                // Optimise the tree branches and (depending on choice) model parameters
                Optimise::optimiseQuasiNewton( currentTree, *model, actualBranchesLength,
                                                        optimiseModel, 1e-6, false );
                cout << "." << flush;
                currentLikelihood = currentTree.loglikelihood() ;

                //if it is the first try or if the likelihood is better update the best values
                if ( ( currentLikelihood > bestLikelihood ) || ( first ) ){
                    first = false;
                    bestPosition = j;
                    bestLikelihood = currentLikelihood;
                    currentTree.getBranchVector( bestBranchesLength );
                    if ( optimiseModel ){
                        model->getAllParameters( bestModelParameters );
                    }
                }
                //backtrack to previous tree (without this taxa)
                currentTree.removeNode( clusters[order[i]].commonAncestor->toStringTopology() );
                // reset to starting model parameters (the best parameters for the tree without the new taxa)
                if ( optimiseModel ){
                     model->setAllParameters( modelParameters );
                }
                // reset to starting branch lengths (the best lengths for the tree without the new taxa)
                currentTree.setBranchVector( branchesLength );
            }
        }
        //once the best position is found update the initial values for the next taxa
        assert( currentTree.addClusteredNode( clusters[order[i]].commonAncestor , bestPosition ) );
        if(optimiseModel){
            modelParameters = bestModelParameters;
        }
        branchesLength = bestBranchesLength;

        cout << endl << "Added : " << clusters[order[i]].name << endl ;
        out << "Added : " << clusters[order[i]].name << endl ;
    }

    if ( optimiseModel ){
        model->setAllParameters( modelParameters ) ;
    }

    currentTree.setBranchVector( branchesLength );
    cout << setprecision(4);
    out << setprecision(4);
    cout << "Best Tree : " << currentTree.toString( false )  << endl;
    cout << "LnLikelihood = " << currentTree.loglikelihood() << endl << endl;
    out << "Best Tree : " << currentTree.toString( false )  << endl;
    out << "LnLikelihood = " << currentTree.loglikelihood() << endl << endl;

    if ( optimiseModel ){
        model->printParameters( cout );
        model->printParameters( out );
    }
    treesSet->insert( currentTree.toString( false ), branchesLength,
                      modelParameters , currentTree.loglikelihood() ) ;
    cout.unsetf(ios::fixed);
    out.unsetf(ios::fixed);    
    return ( currentTree.loglikelihood() ) ;
}  
  
  
void TopologyOptimise::branchAndBoundSearch( SequenceTable * sequenceTable, Model * model,
    bool optimiseModel, TreesSet* treesSet, ostream & out ){

    unsigned int numberTaxa = sequenceTable->getNumberSpecies();
    int present;
    unsigned int next[numberTaxa - 2] ;

    InferenceTree currentTree;

    // vector of parameters
    vector < double > initialModelParameters;
    vector < double > currentModelParameters;
    
    double currentLikelihood;
    double bestLikelihood;

    // branch vector
    vector < double > branchLengths;
    vector<double> actualBranchLengths;
                
    string stringTree;
    
    // Load Data and Model
    currentTree.loadDataAndModel( sequenceTable, model ) ;
    model->getAllParameters( initialModelParameters );

    // Create Initial Tree by adding the first 3 leaves to the empty tree
    currentTree.addLeaf( sequenceTable->species[0], 0 ) ;
    currentTree.addLeaf( sequenceTable->species[1], 1 ) ;
    currentTree.addLeaf( sequenceTable->species[2], 2 ) ;


    // Find the initial upper bound by a heuristic method
    TreesSet heuristicTreesSet(1);
    bestLikelihood = TopologyOptimise::buildTreeStepwise( sequenceTable, model,
                                      optimiseModel, &heuristicTreesSet, out );
    cout.setf(ios::fixed);
    out.setf(ios::fixed);    
    cout << setprecision(4);
    out << setprecision(4);
    cout << endl << "Initial Upper Bound = " << bestLikelihood << endl  ;
    out << endl << "Initial Upper Bound = " << bestLikelihood << endl ;

    present = 0;
    next[0] = 0;

    while ( ( ( next[0] < (unsigned int)currentTree.getNumberBranches() ) ||
                (present != 0 ) ) &&
            (  present >= 0 ) ) {
        if (optimiseModel){
            model->setAllParameters( initialModelParameters );
            //no need to invalidate the likelihood computation at each leaf
            //it will be done when the branch lengths are changed below
        }
        currentTree.getBranchVector( branchLengths ) ;
        for ( unsigned int i = 0 ; i < branchLengths.size() ; ++i ) {
            branchLengths[i] = 0.1;
        }
        currentTree.setBranchVector( branchLengths ) ;
        // Generate next candidate tree (with backtracking if bound is exceeded)
        while ( present < ( (int)numberTaxa - 3 ) ) {
            currentTree.addLeaf( sequenceTable->species[present + 3] , next[present] );
            next[present] ++;
            present++;
            next[present] = 0;
            // Check to see if the upper bound has been exceeded with the "subtree"
            // and if so backtrack immediatly
            if ( present < ( (int)numberTaxa - 3 ) ) {
                currentLikelihood = currentTree.loglikelihood();
                //if the current tree does not pass the bound without optimisation (the usual case)
                currentTree.getBranchVector(actualBranchLengths);
                if ( currentLikelihood < bestLikelihood ) {
                    Optimise::optimiseQuasiNewton( currentTree, *model,
                                  actualBranchLengths , optimiseModel , 1e-6, false);
                    currentLikelihood = currentTree.loglikelihood() ;
                }
                //if the current tree does not pass the bound even with
                //optimisation then backtrack there is no need to add more
                //species because the final tree would not be the best one
                if ( currentLikelihood < bestLikelihood ) {
                    cout << "subtree over the bound : " << currentTree.toStringTopology(false) << endl;
                    out << "subtree over the bound : " << currentTree.toStringTopology(false) << endl;
                    do {
                        present-- ;
                        currentTree.removeLeaf( sequenceTable->species[present + 3] ) ;
                    }
                    while ( ( present > 0 ) &&
                    ( next[present] >= (unsigned int)currentTree.getNumberBranches() ) ) ;
                } // End backtrack	
            } // End upper bound checking
        } // End tree generation
        
        // Print out the tree to optimise
        cout << "Topology = " << currentTree.toStringTopology( false ) << endl ;
        cout << "Progress = " << flush ;
        currentTree.getBranchVector(actualBranchLengths);
        Optimise::optimiseQuasiNewton( currentTree, *model, actualBranchLengths,
                                       optimiseModel , 1e-6, true );
        cout << endl;
        
        currentLikelihood = currentTree.loglikelihood();
        //a better tree is found, update the parameters of the best tree and the bound
        if ( currentLikelihood >= bestLikelihood ) {
            bestLikelihood = currentLikelihood ;
        }

        currentTree.getBranchVector( branchLengths ) ;
        stringTree = currentTree.toString( false ) ;
        cout << "Tree = " << stringTree  << endl ;
        out << "Tree = " << stringTree  << endl ;
        if( optimiseModel ){
            model->getAllParameters( currentModelParameters ) ;
            treesSet->insert( stringTree , branchLengths , currentModelParameters , currentLikelihood );
        }
        else{
            treesSet->insert( stringTree , branchLengths , initialModelParameters , currentLikelihood );
        }

        //print out the loglikelihood of the optimised tree
        cout << setprecision(4);
        out << setprecision(4);
        cout << "ML = " << currentLikelihood << endl;
        out << "ML = " << currentLikelihood << endl;

        // Backtrack
        do {
            --present;
            currentTree.removeLeaf( sequenceTable->species[present + 3] ) ;
        }
        while ( ( present > 0 ) &&
        ( next[present] >= (unsigned int)currentTree.getNumberBranches() ) ) ;
    }
    cout.unsetf(ios::fixed);
    out.unsetf(ios::fixed);    
}

void TopologyOptimise::branchAndBoundSearch( SequenceTable * sequenceTable,
        Model * model, bool optimiseModel,
        TreesSet* treesSet, ostream & out, ClustersSet& clusters ) {

    unsigned int numberClusters = clusters.size();
    int present;
    unsigned int next[numberClusters - 2] ;

    InferenceTree currentTree;

    // vector of parameters
    vector < double > initialModelParameters;
    vector < double > currentModelParameters;
    
    double currentLikelihood;
    double bestLikelihood;

    // branch vector
    vector < double > branchLengths;
    vector<double> actualBranchLengths;
                
    string stringTree;
    
    // Load Data and Model
    currentTree.loadDataAndModel( sequenceTable, model ) ;
    model->getAllParameters( initialModelParameters );

    // Find the initial upper bound by a heuristic method
    ClustersSet heuristicClusters = clusters;
    TreesSet heuristicTreesSet(1);
    bestLikelihood = TopologyOptimise::buildTreeStepwise( sequenceTable, model,
                                   optimiseModel, &heuristicTreesSet, out, heuristicClusters );
    cout.setf(ios::fixed);
    out.setf(ios::fixed);    
    cout << setprecision(4);
    out << setprecision(4);
    cout << endl << "Initial Upper Bound = " << bestLikelihood << endl  ;
    out << endl << "Initial Upper Bound = " << bestLikelihood << endl ;
    
    //then start...
    for ( ClustersSet::iterator i = clusters.begin();
                                   i != clusters.end(); ++i ){
        (*i).commonAncestor = Node::makenode( (*i).name, &currentTree );
    }

    //Create Initial tree
    currentTree.addClusteredNode( clusters[0].commonAncestor , 0 ) ;
    currentTree.addClusteredNode( clusters[1].commonAncestor , 1 ) ;
    currentTree.addClusteredNode( clusters[2].commonAncestor , 2 ) ;

    

    present = 0;
    next[0] = 0;

    while ( ( ( next[0] < (unsigned int)currentTree.getNumberBranches() ) ||
                (present != 0 ) ) &&
            (  present >= 0 ) ) {
        if (optimiseModel){
            model->setAllParameters( initialModelParameters );
            //no need to invalidate the likelihood computation at each leaf
            //it will be done when the branch lengths are changed below
        }
        currentTree.getBranchVector( branchLengths ) ;
        for ( unsigned int i = 0 ; i < branchLengths.size() ; ++i ) {
            branchLengths[i] = 0.1;
        }
        currentTree.setBranchVector( branchLengths ) ;
        // Generate next candidate tree (with backtracking if bound is exceeded)
        while ( present < ( (int)numberClusters - 3 ) ) {
            // add the cluster to the tree (refused if inside another cluster)
            while ( !currentTree.addClusteredNode( clusters[present + 3].commonAncestor,
                                          next[present] ) ) {
                ++(next[present]);
            }
            ++(next[present]);
            ++present;
            next[present] = 0 ;
            //initialise the branches of the cluster (it might change
            //the value of null internal branch but it does not matter)
            currentTree.getBranchVector( actualBranchLengths );
            for ( unsigned int branchId = 0 ;
                  branchId < actualBranchLengths.size(); ++branchId ){
                if ( actualBranchLengths[branchId] <= 0.0 ){
                    actualBranchLengths[branchId] = 0.1;
                }
            }
            currentTree.setBranchVector( actualBranchLengths );
            // Check to see if the upper bound has been exceeded with the "subtree"
            // and if so backtrack immediatly
            if ( present < ( (int)numberClusters - 3 ) ) {
                currentLikelihood = currentTree.loglikelihood();
                //if the current tree does not pass the bound without optimisation (the usual case)
                if ( currentLikelihood < bestLikelihood ) {
                    Optimise::optimiseQuasiNewton( currentTree, *model,
                                  actualBranchLengths , optimiseModel , 1e-6, false);
                    currentLikelihood = currentTree.loglikelihood() ;
                }
                //if the current tree does not pass the bound even with
                //optimisation then backtrack there is no need to add more
                //species because the final tree would not be the best one
                if ( currentLikelihood < bestLikelihood ) {
                    cout << "subtree over the bound : " << currentTree.toStringTopology(false) << endl;
                    out << "subtree over the bound : " << currentTree.toStringTopology(false) << endl;
                    do {
                        present-- ;
                        currentTree.removeNode( clusters[present + 3].commonAncestor->toStringTopology() ) ;
                    }
                    while ( ( present > 0 ) &&
                    ( next[present] >= (unsigned int)currentTree.getNumberBranches() ) ) ;
                } // End backtrack	
            } // End upper bound checking
        } // End tree generation
        
        // Print out the tree to optimise
        cout << "Topology = " << currentTree.toStringTopology( false ) << endl ;
        cout << "Progress = " << flush ;
        currentTree.getBranchVector(actualBranchLengths);
        for ( unsigned int branchId = 0 ;
            branchId < actualBranchLengths.size(); ++branchId ){
            if ( actualBranchLengths[branchId] <= 0.0 ){
                actualBranchLengths[branchId] = 0.1;
            }
        }
        currentTree.setBranchVector(actualBranchLengths);
        Optimise::optimiseQuasiNewton( currentTree, *model, actualBranchLengths,
                                       optimiseModel , 1e-6, true );
        cout << endl;
        currentLikelihood = currentTree.loglikelihood();
        //a better tree is found, update the parameters of the best tree and the bound
        if ( currentLikelihood >= bestLikelihood ) {
            bestLikelihood = currentLikelihood ;
        }

        currentTree.getBranchVector( branchLengths ) ;
        stringTree = currentTree.toString( false ) ;
        cout << "Tree = " << stringTree  << endl ;
        out << "Tree = " << stringTree  << endl ;
        if( optimiseModel ){
            model->getAllParameters( currentModelParameters ) ;
            treesSet->insert( stringTree , branchLengths , currentModelParameters , currentLikelihood );
        }
        else{
            treesSet->insert( stringTree , branchLengths , initialModelParameters , currentLikelihood );
        }

        //print out the loglikelihood of the optimised tree
        cout << setprecision(4);
        out << setprecision(4);
        cout << "ML = " << currentLikelihood << endl;
        out << "ML = " << currentLikelihood << endl;

        // Backtrack
        do {
            --present;
            currentTree.removeNode( clusters[present + 3].commonAncestor->toStringTopology() ) ;
        }
        while ( ( present > 0 ) &&
        ( next[present] >= (unsigned int)currentTree.getNumberBranches() ) ) ;
    }
    cout.unsetf(ios::fixed);
    out.unsetf(ios::fixed);    
}
